#include "PWilib.h"
#include "PWilib_routines.h"
#include "PWcir.h"
#include "array.h"

extern char*  PWdeviceName[];

/* copied from sis. See list.c */
/* should be kept same with that in list.c */
typedef struct list_elem {  /* One list element  */
    struct list_desc *mainList; /* List descriptor   */
    struct list_elem *prevPtr;  /* Previous element  */
    struct list_elem *nextPtr;  /* Next list element */
    lsGeneric userData;     /* User pointer      */
} lsElem;

typedef struct list_desc {  /* List descriptor record            */
    lsElem *topPtr, *botPtr;    /* Pointer to top and bottom of list */
    int length;         /* Length of list                    */
} lsDesc;

typedef struct gen_desc {   /* Generator Descriptor     */
    lsDesc *mainList;       /* Pointer to list descriptor   */
    lsElem *beforeSpot;     /* Item before the current spot */
    lsElem *afterSpot;      /* Item after the current spot  */
} lsGenInternal;
/*----------------------------------------------------------*/

extern int valid_nin_nout_num(ilib_table_t *table, int num, int flag);

/* get the entry from the ilib table, corresponding to one particular gate */
ilib_entry_t *
ilibEntry(ilib_table_t *table, int index)
{
	return array_fetch(ilib_entry_t *, table->body, index);
}

/* get the entry from ilib table, explicitly using the three indexing fields */
ilib_entry_t *
ilibEntry3(ilib_table_t *table, int type_index, int nin, int nout)
{
	int	table_index;

	table_index = lookup_table_index(type_index, nin, table->rule->fanin_field_bits, nout, 
					table->rule->fanout_field_bits);

	return ilibEntry(table, table_index);
}

/* get the headList for an entry */
lsList
ilibEntryHeadList(ilib_entry_t *entry)
{
	return entry->headList;
}

/* get the pairlist with specified turn-on time
 *
 * NOTE: list_type currently belongs to [0,1]
 */
lsList
ilibPairList(ilib_table_t *table, int index, float turnon_time, int list_type)
{
	lsList 			headlist;
	lsGen  			gen;
	ilib_pair_list_head_t   *head;

	headlist = ilibEntryHeadList(ilibEntry(table, index));
	lsForeachItem(headlist, gen, head)
	{
		if(head->turnonTime == turnon_time)
		break;
	}
	if(head == NULL) return NULL;
	else{
		return head->pairList[list_type];
	    }
}

/* get the value of current from a pairlist at the specified time point */
float
get_current(lsList pairlist, float time)
{
    lsGen       gen;
    ilib_pair_t *pair, *pair_tmp;
    int         i;
    float       time1, time2, current1, current2;

    i=0;
    lsForeachItem( pairlist, gen, pair )
    {
        if( time == pair->time )
        {
        	lsFinish(gen); /* exit for loop earily, lsFinish should be explicitly called */
            return pair->current;
		}
        else if( time < pair->time )
        {
            if( i == 0 ) 
            {
            	lsFinish(gen);
            	return 0; /* before the starting point of waveform */
			}
            else
            {
                pair_tmp = (ilib_pair_t*) (((lsGenInternal *) gen)->beforeSpot->prevPtr->userData);
                time1=pair_tmp->time;
                time2=pair->time;
                current1 = pair_tmp->current;
                current2 = pair->current;
                lsFinish(gen);
                return  piece_linear_point(current1, time1, current2, time2, time);
            }
        }
        i++;
    }

    /* after the ending point of waveform */
    return 0;
}

/* access the ILIB, and update the gate_current field in the instance */
int
accessCurrentILIB(ilib_table_t *table, PW_INST *inst, float time, float turnon_time)
{
	char 					*name;
	int						nin, nout;
	ilib_entry_t 			*entry;
	lsList					headlist;
	ilib_pair_list_head_t 	*head;
	lsGen					gen;
	int						i;


	/* gate type */
	if(strcmp(PWdeviceName[inst->instLibNo - 1], "PO") == 0)
	{
		fprintf(stderr,"PO shouldn't appear here in accessCurrentILIB().\n");
		exit(1);
	}else if(strcmp(PWdeviceName[inst->instLibNo - 1], "PI") == 0)
	{
		i = BUFF_ILIB_INDEX;
	}else
	{

		for(i=0; i<GENERIC_TYPE_NUM_IN_ILIB; i++)
		{
			name = PWdeviceName[inst->instLibNo-1];
			if(strcmp(name, table->generic_gate_types[i])==0)
				break;
		}
		if(i>= GENERIC_TYPE_NUM_IN_ILIB)
		{
			fprintf(stderr,"The circuit has the gate type not supported by the library.\n");
			return 1;
		}
	}

	/* fanin/fanout # */
	if(i == BUFF_ILIB_INDEX)
		nin = 1;	/* in case that it's a PI implemented by buff, get correct nin */
	else nin = valid_nin_nout_num(table, inst->fanin, 0);

	nout = valid_nin_nout_num(table, inst->fanout,1);


	entry = ilibEntry3(table,  i, nin, nout);
	headlist = ilibEntryHeadList(entry);

	/* match turnon time */
	lsForeachItem(headlist, gen, head)
	{
		if(head->turnonTime == turnon_time)
		{
			inst->powerup_current->current[RISE_LIST] = get_current(head->pairList[RISE_LIST], time);
			inst->powerup_current->current[STABLE0_LIST] = get_current(head->pairList[STABLE0_LIST], time);
			if(inst->powerup_current->current[RISE_LIST] >= inst->powerup_current->current[STABLE0_LIST])
				inst->powerup_current->preferred_type = RISE_LIST;
			else	inst->powerup_current->preferred_type = STABLE0_LIST;
		}
	}
	return 0;
}

/* get the valid fanin/fanout number.
 * since limited number of fanin/fanout is supported in the lib,
 * the real fanin/fanout number will be saturated now.

 * flag: 0 -> fanin,      1 -> fanout
 *
 * Later, it will be changed to extrapolation.
 */
int
valid_nin_nout_num(ilib_table_t *table, int num, int flag)
{
	switch(flag)
	{
		case 0:
			return DF_MIN(table->max_fanin_num, num);
		break;
		case 1:
			return DF_MIN(table->max_fanout_num, num);
		break;
		default:
			fprintf(stderr,"Unknown flag in valid_nin_nout_num().\n");
			exit(1);
	}
}
