#include "sis.h"
#include "PWilib.h"
#include "PWilib_routines.h"

/* copied from sis. See list.c */
/* should be kept same with that in list.c */
typedef struct list_elem {  /* One list element  */
    struct list_desc *mainList; /* List descriptor   */
    struct list_elem *prevPtr;  /* Previous element  */
    struct list_elem *nextPtr;  /* Next list element */
    lsGeneric userData;     /* User pointer      */
} lsElem;

typedef struct list_desc {  /* List descriptor record            */
    lsElem *topPtr, *botPtr;    /* Pointer to top and bottom of list */
    int length;         /* Length of list                    */
} lsDesc;

typedef struct gen_desc {   /* Generator Descriptor     */
    lsDesc *mainList;       /* Pointer to list descriptor   */
    lsElem *beforeSpot;     /* Item before the current spot */
    lsElem *afterSpot;      /* Item after the current spot  */
} lsGenInternal;
/*----------------------------------------------------------*/

/* get the time from a pair */
extern lsGeneric time_access_in_pairlist(lsGeneric datafield);

/* use time as the key, and do the comparison */
extern int time_compare_in_pairlist(lsGeneric datafield1, lsGeneric datafield2);

/* get the leftmost & rightmost peak time for a piecewise linear waveform */
/* pairlist is ordered by the time of the pair */
/* and the current for both starting point and ending point should be 0 */
void lrPeakTimeOneWaveform(lsList pairlist, float *left_time, float *right_time);

/* Start a in-order traversal of two ordered list just as one merged ordered list
 * The order is determined the key in each element of the lists 
 */
void
lsTraverseInorder2_start(lsList list1, lsList list2, lsGen* gen1, lsGen* gen2)
{
	*gen1 = lsStart(list1);	/* direct pointer afterSpot to topPtr */
	*gen2 = lsStart(list2);
}

/* Get the data contained in the current element of the list, and move to the next element 
 * NOTE: the data field of the element in the list MUST be a pointer!
 */
lsStatus
lsTraverseInorder2_next(lsGen gen1, lsGen gen2, 
			lsGeneric* data, lsGeneric (*data_access)(), int (*data_compare)() )
{
	lsGeneric data_field1, data_field2; /* pointers */
	lsGeneric dummy;		/* a pointer */
	lsStatus status1, status2;
	int			compare_result;

	
	status1 = lsCurrent(gen1, &data_field1, LS_NH);
	status2 = lsCurrent(gen2, &data_field2, LS_NH);

	if(status1 == LS_OK && status2 == LS_OK)
	{
		compare_result = data_compare(data_field1, data_field2);
		if(compare_result > 0)
		{
			(*data) = data_access(data_field2);
			lsNext(gen2, &dummy, LS_NH);
		}else if(compare_result == 0)
		{
			(*data) = data_access(data_field1);
			lsNext(gen1, &dummy, LS_NH);
			lsNext(gen2, &dummy, LS_NH);
		}else
		{
			(*data) = data_access(data_field1);
			lsNext(gen1, &dummy, LS_NH);
		}
		return LS_OK;
	}else if(status1 == LS_OK && status2 == LS_NOMORE)
	{
		(*data) = data_access(data_field1);
		lsNext(gen1, &dummy, LS_NH);
		return LS_OK;
	}else if(status1 == LS_NOMORE && status2 == LS_OK)
	{
		(*data) = data_access(data_field2);
		lsNext(gen2, &dummy, LS_NH);
		return LS_OK;
	}else /*status1 == LS_NOMORE && status2 == LS_NOMORE*/
	{
		return LS_NOMORE;
	}
}

/* access the current element specified by the generator */
lsStatus
lsCurrent(lsGen generator, lsGeneric *data, lsHandle *itemHandle)
{
	lsGenInternal *realGen;
	realGen  = (lsGenInternal *) generator;

	if(realGen->afterSpot == NIL(lsElem)) 
	{
		*data = (lsGeneric) 0;
		if (itemHandle) *itemHandle = (lsHandle) 0;
		return LS_NOMORE;
	} else {
		*data = realGen->afterSpot->userData;
		if(itemHandle) *itemHandle = (lsHandle) (realGen->afterSpot);
		return LS_OK;
	       }
}

/* get the time from a pair */
lsGeneric
time_access_in_pairlist(lsGeneric datafield)
{
	return (lsGeneric) &(((ilib_pair_t*)datafield)->time);
}

/* use time as the key, and do the comparison */
int
time_compare_in_pairlist(lsGeneric datafield1, lsGeneric datafield2)
{
	int   flag;
	float time1, time2;

	time1 = ((ilib_pair_t*)datafield1)->time;
	time2 = ((ilib_pair_t*)datafield2)->time;
	if( time1 > time2 )
		flag = 1;
	else if( time1 == time2 )
		flag = 0;
	else flag = -1;

	return flag;
}

/* finish the traversal */
void
lsTraverseInorder2_finish(lsGen gen1, lsGen gen2)
{
	lsFinish(gen1);
	lsFinish(gen2);
}


/* get the value from piece-wise linear function */
float
piece_linear_point(float y1,float x1,float y2,float x2,float x)
{
	if( x1 > x2 || x < x1 || x > x2 )
	{
		fprintf(stderr,"Error: in piece_linear_point().\n");
		fprintf(stderr,"sampled point doesn't fall into the range.\n");
		exit(1);
	}
	return (y1 + (x-x1)* ((y2 - y1)/(x2 - x1)));
}

/* get the sum of two current waveform */
int
sum_current_waveform( lsList new_pairlist, lsList src1_pairlist, lsList src2_pairlist )
{
	lsGen	gen1, gen2;
	float	*time_p;
	float  i1, i2;

	if(lsLength(new_pairlist) > 0)
	{
		fprintf(stderr,"Error in add_current_waveform(): the new_pairlist is not null.\n");
		return 1;
	}
	lsTraverseInorder2_start(src1_pairlist, src2_pairlist, &gen1, &gen2);
	while(lsTraverseInorder2_next(gen1, gen2, (lsGeneric*) &time_p, 
								time_access_in_pairlist, 
								time_compare_in_pairlist) == LS_OK)
	{
		i1 = get_current(src1_pairlist, *((float *)(time_p)));
		i2 = get_current(src2_pairlist, *((float *)(time_p)));
		pairNewEnd(new_pairlist, i1+i2, *((float *)(time_p)));
	}
	lsTraverseInorder2_finish(gen1, gen2);

	return 0;
}

/* get the time for the left-most & right-most local peak current */
void
left_right_PeakTime(ilib_table_t *table)
{
	float					left, right;
	int						i,j;
	lsGen					gen1;
	ilib_pair_list_head_t	*head;
	ilib_entry_t 			*entry;

	for(i=0; i<array_n(table->body); i++)
	{
		entry = array_fetch(ilib_entry_t *, table->body, i);
		if(lsLength(entry->headList)>0)
		{
			lsForeachItem(entry->headList, gen1, head)
			{
				if(lsLength(head->pairList[RISE_LIST])>0 && lsLength(head->pairList[STABLE0_LIST])>0)
				{
					for(j=0; j<2; j++)
					{
						lrPeakTimeOneWaveform(head->pairList[j], &left, &right);
						head->leftmostPeakTime[j] = left;
						head->rightmostPeakTime[j] = right;
					}
				}
			}
		}
	}
}

/* pairlist is ordered by the time of the pair */
/* and the current for both starting point and ending point should be 0 */
void
lrPeakTimeOneWaveform(lsList pairlist, float *left_time, float *right_time)
{
	float 	prevTime = -1, prevCurrent = -1;
	int		flag = 0;
	lsGen	gen;
	ilib_pair_t	*pair;

	lsForeachItem(pairlist, gen, pair)
	{
		if(pair->current < prevCurrent)
		{
			flag++;
			if( flag == 1 ) *left_time = prevTime;
			*right_time = prevTime;
		}else{
				prevCurrent = pair->current;
				prevTime = pair->time;
			 }
	}

	/* hack: if incorrect library entry */
	if(flag == 0)
		*left_time = *right_time = prevTime;
}
