#include "ga_alg.h"
#include "PWcir.h"
#include "PWglobal_var.h"
#include "PWroutines.h"

#include "stdlib.h"

static int gene_count; /* number of genes in one chromosome */
static int gene_bits;  /* number of bits for one gene in the chromosome */
static int chromo_wsize;      /* size of the chromosome in 'ints' */
static int pop_size;   /* the size of the population */
static int max_generation;    /* the maximum generations */

/* init_ga_alg().
 *
 * initialize the Genetic Algorithm.
 */
void
init_ga_alg( int gene_count_l, int gene_bits_l, int chromo_wsize_l, int pop_size_l, int max_generation_l )
{
  gene_count = gene_count_l;
  gene_bits  = gene_bits_l;
  chromo_wsize = chromo_wsize_l;
  pop_size     = pop_size_l;
  max_generation = max_generation_l;
}

/* access to the static vars */
int get_gene_count(void) { return gene_count; }
int get_gene_bits(void) { return gene_bits; }
int get_chromo_wsize(void) { return chromo_wsize; }
int get_pop_size(void) { return pop_size; }
int get_max_generation(void) { return max_generation; }


/* ga_alg()
 *
 * a genetic algorithm template 
 */
chromosome_t *
ga_alg( array_t * (*init_pop)(void), 
        double fitness_func(chromosome_t *chr),
        array_t * (*select)(array_t *pop),
        void (*crossover)(array_t *pop), 
        void (*mutation)(array_t *pop), 
        void (*replace)(array_t **old_pop, array_t *new_pop), 
        BOOLEAN (*finish)(int gen_num, int max_gen, int no_improve_num) )
{
  array_t *pop, *new_pop;
  chromosome_t *best=NULL, *current_best=NULL;
  int gen_num=0;
  int no_improve_num=0;

  pop = init_pop();

  while(finish(gen_num,max_generation, no_improve_num) == FALSE)
  {
    /* compute the fitness values and return the best chromo DUPLICATION in current generation*/
    current_best = eval_fitness(pop, fitness_func);
    update_best_chromo(&best,current_best,&no_improve_num);

    /*debugging*/
    fprintf(stdout,"Generation=%d\tfitness=%f\n",gen_num,best->fitness);

    /* selection of a new population for reproduction */
    new_pop = select(pop);
    /* genetic operations */
    crossover(new_pop);
    mutation(new_pop);
    /* replacement */
    replace(&pop, new_pop);
    gen_num++;
  }

  return best;
}

/* update the best chromo so far
   and return the number of consecutive generations with no improvement 
 */
void
update_best_chromo(chromosome_t **best, chromosome_t *current_best, int *no_improve_num)
{
  if( (*best) == NULL )
  { (*best) = current_best; 
    no_improve_num = 0;
  }else{
         if( (*best)->fitness < current_best->fitness )
         {
           FREE((*best)->data);
           FREE((*best));
           (*best) = current_best;
           no_improve_num = 0;
         } else{
                 FREE(current_best->data);
                 FREE(current_best);
                 no_improve_num++;
               }
       }
}

void
clear_visited_flag(array_t *pop)
{
  int i;
  chromosome_t *chr;

  for(i=0;i<array_n(pop);i++)
  {
    chr = array_fetch(chromosome_t*, pop, i);
    chr->visited = FALSE;
  }
}

/* set_chromo_bit()
 *
 * HACK: set one gene bit (binary) in a chromosome
 */
void
set_chromo_bit(chromosome_t *chromo_p, int position, int value)
{
  int offset_in_word;
  int word_index;

  word_index = get_word_index(position);
  offset_in_word = get_word_offset(position);
  if(value == 1)
  {
    chromo_p->data[word_index] |= 1 << offset_in_word;
  }else{
         chromo_p->data[word_index] &= ~(1 << offset_in_word);
       }
}

/* get_word_index()
 * (1)it uses global var 'gene_bits' to handle different gene encoding shceme.
 * (2)a word is an 'int'. 
 * (3)'gene_position' is the number of the gene in the chromosome, starting from '0'.
 */
int
get_word_index(int gene_position)
{
  int gene_count_per_word;

  /* number of genes stored in one word 'int' */
  gene_count_per_word = (sizeof(int)*8) / gene_bits;
  return gene_position/(gene_count_per_word);
}

/* get_word_offset()
 * (1)it uses global var 'gene_bits' to handle different gene encoding shceme.
 * (2)get the offset of the MSB bit of the encoded gene in a word, i.e., the number of bits
 * between it to the LSB of the word (starting from 0).
 */
int
get_word_offset(int gene_position)
{
  int gene_count_per_word;

  gene_count_per_word = (sizeof(int)*8) / gene_bits;
  return ( (gene_position%gene_count_per_word + 1) * gene_bits -1 );
}

/* new_chromosome()
 * alloc a chromosome. Global var 'chromo_wsize' is used.
 */
chromosome_t *
new_chromosome(void)
{
  int i;

  chromosome_t *p;
  p = ALLOC(chromosome_t, 1);
  p->data = ALLOC(int, chromo_wsize);
  for(i=0;i<chromo_wsize;i++)
  { p->data[i] = 0; }

  p->visited = FALSE;
  return p;
}

/* chromo_dup_and_ins_last()
 *
 * duplicate the chromo and insert into the array as the last element.
 * global var 'chromo_wsize' is used
 */
void
chromo_dup_and_ins_last(array_t* chr_array, chromosome_t * chr)
{
  chromosome_t* p;

  p = dup_chromo(chr);
  array_insert_last(chromosome_t*, chr_array, p);
}

/* dup_chromo()
 *
 * duplicate a chromosome.
 * global var 'chromo_wsize' is used
 */
chromosome_t *
dup_chromo(chromosome_t * chr)
{
  int i;
  chromosome_t * p;

  p = ALLOC(chromosome_t, 1);
  p->data = ALLOC(int, chromo_wsize);

  for(i=0;i<chromo_wsize;i++)     p->data[i] = chr->data[i];
  p->fitness = chr->fitness;
  p->visited = chr->visited;

  return p;
}

/* eval_fitness()
 *
 * compute the fitness values and return the best chromosome.
 */
chromosome_t*
eval_fitness( array_t *pop, double (*fitness)(chromosome_t *) )
{
  int i;
  chromosome_t *chr;
  chromosome_t *best;

  for(i=0;i<array_n(pop);i++)
  {
    chr = array_fetch(chromosome_t *, pop, i);
    chr->fitness = fitness(chr);
    if(i==0) best = chr;
    else
         best = (best->fitness >= chr->fitness)?(best):(chr) ;
  }

  best = dup_chromo(best);
  return best;
}


/*--------------------------------------------------------------------------------------------*/


