#include "leakage_ga.h"
#include "ga_alg.h"
#include "DFrandom.h"
#include "PWcir.h"
#include "PWglobal_var.h"
#include "leakage_ilib.h"

#include "stdlib.h"

#define 		RECIPROCAL_SCALE_FACTOR 	1000.0
extern void     simple_sim_operation(void);
extern void     PWimplyStackGarbageClean(void);

/*--------------------------------------------------------------------------------------------*/

/* lkg_ga_estimation().
 *
 * Genetic Algorithm for leakage min/max current estimation.
 * NOTE: performed at the imaxup data structure.
 */
int
lkg_ga_estimation(int in_max_gen, int min_max_flag)
{
  int population_size;
  int max_gen;
  int chromo_word_size;
  int bits_per_pi;
  int num_control_pi;
  chromosome_t * solution;


  num_control_pi = controllable_pi();
  population_size = DF_MAX(num_control_pi, 24);
  if( (population_size % 2) != 0 ) population_size++;

  max_gen = DF_MAX(100,in_max_gen);
  bits_per_pi = 1;
  chromo_word_size = ceil( (double)((num_control_pi*1.0) / ((sizeof(int)*8) / bits_per_pi)) );

  /* initialization */
  init_ga_alg(num_control_pi, bits_per_pi, chromo_word_size , population_size, max_gen);

  /* genetic algorithm */
  if(min_max_flag >0) // max leakage current
  {
  solution = ga_alg(init_pop_vector, fitness_lkg_current, 
                    select_vector1, crossover_vector, mutation_vector, 
                    replace_vector, finish_vector);
  }else{ // min leakage current
		solution = ga_alg(init_pop_vector, fitness_lkg_current_reciprocal,
						  select_vector1, crossover_vector, mutation_vector,
						  replace_vector, finish_vector);
	   }

  GA_report_result_vector(solution, min_max_flag, resultFp);

  return 0; /* success */
}

/* get pi & ppi number, excluding GND/VDD */
int
controllable_pi(void)
{
  int i,j;
  for(i=0,j=0;i<PrimInNum;i++)
  {
    if( strcmp(SourceP[i]->name, "VDD") == 0 ||
        strcmp(SourceP[i]->name, "GND") == 0 ||
        strcmp(SourceP[i]->name, "VSS") == 0 ||
        strcmp(SourceP[i]->name, "VCC") == 0)
    j++;
  }
  return (PrimInNum - j);
}

/* init_pop_vector()
 *
 * randomly generate the initial population of vectors for leakage current estimation
 * one bit encoding for one primary input or pseudo primary input
 */
array_t *
init_pop_vector(void)
{
  int i,j;
  array_t * pop;
  chromosome_t *chromo_p;

  /* seed */
  randomGenInit1();
  pop = array_alloc(chromosome_t *, get_pop_size());

  for(i=0;i<get_pop_size();i++)
  {
   	chromo_p = new_chromosome();
    for(j=0;j<get_gene_count();j++)
    {
      /* randomly generating input value */
      random_set_input_bit(chromo_p,j);
    }
    array_insert(chromosome_t *, pop, i, chromo_p);
  }

  /* debugging */
  //fprintf(stdout,"Initial population:\n");
  //decode_dump_pop_vector(pop, stdout);
  return pop;
}

/* random_set_input_bit()
 *
 * randomly set the gene corresponding to the input bit 
 */
void
random_set_input_bit(chromosome_t *chromo_p, int position)
{
  int value;

  value = rand_logic_value(0.5);
  set_chromo_bit(chromo_p, position, value);
}

/* select_vector1()
 *
 * (1)the selection scheme for leakage current estimation.
 * (2)We use tournament selection without replacement.(Two better groups)
 * (3)Call *eval_fitness() first to get fitness values.
 * (4)new population has new chromosome instances.
 */
array_t *
select_vector1(array_t *pop)
{
  array_t * better1, *better2;


  /* do tournament selection twice with the same random gen seed */
  randomGenInit1();
  better1 = tournaSel(pop);
  better2 = tournaSel(pop);

  array_append(better1, better2);
  array_free(better2);

  /* debugging */
  //fprintf(stdout, "Population after selection:\n");
  //decode_dump_pop_vector(better1, stdout);

  return better1;
}

/* tournaSel()
 * (1)tournament selection without replacement.
 * (2)a new array with new chromo instances is returned.
 * (3)random gnerator seed must be called before calling it!
 */
array_t * tournaSel(array_t *pop)
{
  int n,i,j;
  array_t * pop_left;
  array_t * better;
  chromosome_t *p1, *p2;

  better = array_alloc(chromosome_t *, 0);
  pop_left = array_dup(pop); /* pointer addresses are copied */

  while(array_n(pop_left) > 0)
  {
    n = array_n(pop_left);
    /* two different random numbers from 0 to n-1 */
    i = ((int)(drand48()*n)) % n;
    p1 = array_fetch(chromosome_t*, pop_left, i);
    do{
        j = ((int)(drand48()*n)) % n;
      }while (j==i);
    p2 = array_fetch(chromosome_t*, pop_left, j);

    if(p1->fitness > p2->fitness)
    {
      chromo_dup_and_ins_last(better, p1);
    }else{
           chromo_dup_and_ins_last(better, p2);
         }

    p1->visited = p2->visited = TRUE;

    remove_visited_chromo(&pop_left,i,j);
  }
  return better;
}




/* hack: assuming only two chromo are visited previously */
void
remove_visited_chromo(array_t **pop_left, int chromo1, int chromo2)
{
  array_t *tmp_array;
  int i,k,n;
  chromosome_t *chr;

  n = array_n(*pop_left);
  tmp_array = array_alloc(chromosome_t *, n-2);
  for(k=0,i=0;k<array_n(*pop_left);k++)
  {
    if(k != chromo1 && k != chromo2)
    {
      chr = array_fetch(chromosome_t*, *pop_left, k);
      array_insert(chromosome_t*, tmp_array, i, chr);
      i++;
    }
  }
  array_free(*pop_left);
  (*pop_left) = tmp_array;
}

/* replace_vector()
 *
 * the replacement scheme in leakage current estimation.
 * old_pop will be freed, and completley replaced by the new_pop.
 */
void
replace_vector(array_t **old_pop, array_t *new_pop)
{
  int i;
  chromosome_t *chr;

  for(i=0;i<array_n(*old_pop);i++)
  {
    chr = array_fetch(chromosome_t*, *old_pop, i);
    FREE(chr->data);
    FREE(chr);
  }
  array_free(*old_pop);
  (*old_pop) = new_pop;
}

/* crossover_vector()
 *
 * the crossover scheme for leakage current estimation.
 * one-point crossover with probability of 1.
 */
void
crossover_vector(array_t *pop)
{
  array_t * pop_left;
  int n;
  int i,j;
  chromosome_t *p1, *p2;


  clear_visited_flag(pop);
  pop_left = array_dup(pop); /* pointer addresses are copied */

  randomGenInit1();      /* seed for randomly pick up two parent strings */
  randomGenInit2();      /* seed for randomly generates cross-over point, and mutation prob */

  while(array_n(pop_left) > 0)
  {
    n = array_n(pop_left);

    /* two different random numbers from 0 to n-1 */
    i = ((int)(drand48()*n)) % n;
    p1 = array_fetch(chromosome_t*, pop_left, i);
    do{
       j = ((int)(drand48()*n)) % n;
    } while(j == i);
    p2 = array_fetch(chromosome_t*, pop_left, j);

    one_point_crossover(p1,p2);
    p1->visited = p2->visited = TRUE;

    remove_visited_chromo(&pop_left,i,j);
  }

  /* debugging */
  //fprintf(stdout,"Population after crossover:\n");
  //decode_dump_pop_vector(pop, stdout);
}

/* one_point_crossover()
 * the one-point crossover schem
 */
void
one_point_crossover(chromosome_t *p1, chromosome_t* p2)
{
  int gene_position;
  int word_index, offset_in_word;
  int tmp,i;

  /* randomly get a position */
  gene_position = rand() % get_gene_count();
  word_index = get_word_index(gene_position);
  offset_in_word = get_word_offset(gene_position);

  partial_swap(&(p1->data[word_index]), &(p2->data[word_index]), offset_in_word);
  for(i=word_index-1;i>=0;i--)
  {
    tmp = p1->data[i];
    p1->data[i] = p2->data[i];
    p2->data[i] = tmp;
  }
}

/* partial_swap()
 * swap the part from index '0' to 'offset_in_word'
 */
void
partial_swap(int *word1, int *word2, int offset_in_word)
{
  int i;
  int tmp1,tmp2;

  for(i=offset_in_word;i>=0;i--)
  {
    tmp1 = (*word1) & (1<<i);
    tmp2 = (*word2) & (1<<i);
    (*word1) = ((*word1) & (~(1<<i))) | tmp2;
    (*word2) = ((*word2) & (~(1<<i))) | tmp1;
  }
}

/* mutation_vector()
 *
 * (1)muatation scheme for leakage current estimation.
 * (2)mutation probability is chosen as 1/gene_count.
 * (3)HACK: gene_bits is assumed as 1
 */
void
mutation_vector(array_t *pop)
{
  double prob;
  double randout;
  chromosome_t *chr;
  int i,j;
  int word_index, offset_in_word;

  /* mutation probability */
  prob = 1.0 / get_gene_count();

  /* for each chromosome */
  for(i=0;i<array_n(pop);i++)
  {
    chr = array_fetch(chromosome_t*, pop, i);

    /* for each gene */
    for(j=0;j<get_gene_count();j++)
    {
      randout = (rand()*1.0)/RAND_MAX;
      if( randout < prob )
      { 
        /* flip the bit */
        word_index = get_word_index(j);
        offset_in_word = get_word_offset(j);
        chr->data[word_index] = chr->data[word_index] ^ (1<<offset_in_word);
      }
    }
  }

  /* debugging */
  //fprintf(stdout,"Population after mutation:\n");
  //decode_dump_pop_vector(pop, stdout);
}

/* finish proc for leakage current estimation */
BOOLEAN
finish_vector(int gen_num, int max_gen, int no_improve_num)
{
  if(gen_num > max_gen) return TRUE;
  else return FALSE;
}

/* fitness_lkg_current()
 *
 * get the leakage current for the chromo as its fitness
 */
double
fitness_lkg_current(chromosome_t * chr)
{
  double ckt_lkg;

  /* convert chromo to PI inputs */
  decode_chromo_to_vector(chr);

  /* 1-vec simulation */
  PWimplyStackGarbageClean();
  simple_sim_operation();

  /* get leakage current */
  ckt_lkg = compute_lkg_current();

  return ckt_lkg;
}

/* fitness_lkg_current_reciprocal()
 *
 * get the leakage current for the chromo as its fitness
 */
double
fitness_lkg_current_reciprocal(chromosome_t * chr)
{
  double ckt_lkg;

  /* convert chromo to PI inputs */
  decode_chromo_to_vector(chr);

  /* 1-vec simulation */
  PWimplyStackGarbageClean();
  simple_sim_operation();

  /* get leakage current */
  ckt_lkg = compute_lkg_current();

  return (RECIPROCAL_SCALE_FACTOR / ckt_lkg);
}


/* decode_chromo_to_vector()
 *
 * decoding the chromo and convert to an input vector
 * HACK: assuming it is a binary encoding, one input as one bit
 */
void
decode_chromo_to_vector(chromosome_t * chr)
{
  int i,j;
  int word_index, offset_in_word;
  int test;

  for(i=0,j=0;i<get_gene_count();i++)
  {
    word_index = get_word_index(i);
    offset_in_word = get_word_offset(i);
    test = chr->data[word_index] & (1<<offset_in_word);

    if(test == 0) assign_next_input_bit(&j, VALUE_0);
    else assign_next_input_bit(&j, VALUE_1);

    j++;
  }
}

/* assign_next_input_bit()
 *
 * find the next input which is not VDD/GND, and assign the logic value
 */
void
assign_next_input_bit(int * PI_index, LOGIC_VALUE val)
{
  LOGIC_VALUE* pValue;

  while(strcmp(SourceP[(*PI_index)]->name, "GND") == 0 ||
        strcmp(SourceP[(*PI_index)]->name, "VDD") == 0 ||
        strcmp(SourceP[(*PI_index)]->name, "VSS") == 0 ||
        strcmp(SourceP[(*PI_index)]->name, "VCC") == 0)
  { (*PI_index)++; }

  if( (*PI_index) >= PrimInNum ) { fprintf(stderr,"Error in assign_next_input_bit().\n"); exit(1); }

  pValue=(LOGIC_VALUE *)mem_get_block((SIZE_T) sizeof(LOGIC_VALUE));
  (*pValue)=val;
  STK_PUSH(&(SourceP[(*PI_index)]->implyStkTop),(double)NOT_CARE_KEY, (pointer)pValue);
}

/* compute_lkg_current()
 *
 * get the leakage current for each instance and the total leakage current.
 * Assume one simulation has been performed.
 */
double
compute_lkg_current(void)
{
  int i;
  double current, total_current=0;

  if(leakage_ilib_table == NULL) { fprintf(stderr, " No ilib has been read.\n"); exit(1); }

  for(i=0;i<PWinstNum_and_pseudo;i++)
  {
    current = accessLkgILIB_by_instance(leakage_ilib_table, &(instList[i]));
    if(current < 0) { fprintf(stderr,"Something wrong in accessLkgILIB(). Negtive current value returned.\n");
                      exit(1);
                    }

    total_current += current;
  }
  return total_current;
}

/* decode & dump a popualation */
void
decode_dump_pop_vector(array_t *pop, FILE* fp)
{
  int i;
  chromosome_t *chr;

  fprintf(fp, "Dumping the population.\n");
  for(i=0;i<array_n(pop);i++)
  {
    chr = array_fetch(chromosome_t *, pop, i);
    fprintf(fp,"Dumping chromosome No. %d\n", i);
    decode_dump_chromo_vector(chr,fp);
  }
  fflush(fp);
}

/* assuming binary encoding */
void
decode_dump_chromo_vector(chromosome_t *chr, FILE * fp)
{
  int i,j;
  int word_index, word_offset;

  for(i=0,j=0;i<get_gene_count();i++)
  {
    word_index = get_word_index(i);
    word_offset = get_word_offset(i);
    if( (chr->data[word_index] & (1<<word_offset)) == 0)
    {
        while(strcmp(SourceP[j]->name, "GND") == 0 ||
              strcmp(SourceP[j]->name, "VDD") == 0 ||
              strcmp(SourceP[j]->name, "VSS") == 0 ||
              strcmp(SourceP[j]->name, "VCC") == 0)
        { j++; }
        fprintf(fp,"\tSourceP[%d] (%s) = %d\n",j, SourceP[j]->name, 0);
    }else{
           while(strcmp(SourceP[j]->name, "GND") == 0 ||
                 strcmp(SourceP[j]->name, "VDD") == 0 ||
                 strcmp(SourceP[j]->name, "VSS") == 0 ||
                 strcmp(SourceP[j]->name, "VCC") == 0)
           { j++; }
           fprintf(fp,"\tSourceP[%d] (%s) = %d\n",j, SourceP[j]->name, 1);
         }
    j++;
  }
}

void
GA_report_result_vector(chromosome_t *solution, int min_max_flag, FILE *fp)
{
  fprintf(fp,"\n\n==========================================\n");
  fprintf(fp, "The leakage estimation results by GA(Note PIs are treated as buffers: \n");
  if( min_max_flag > 0 ){
    fprintf(fp,"Maximum leakage current= %f\n", solution->fitness);
  }else{
  		fprintf(fp, "Minimum leakage current= %f\n", RECIPROCAL_SCALE_FACTOR/(solution->fitness));
  	   }
  decode_dump_chromo_vector(solution,fp);
  fprintf(fp,"\n\n");
}
