#include "PWilib.h"
#include "PWilib_routines.h"
#include "PWcir.h"
#include "PWroutines.h"
#include "PWglobal_var.h"
#include "leakage_ilib.h"
#include "DFfile.h"
#include "memuser.h"

#define SIS
#include "sis.h"

#define LEAKAGE_REQUIRED_INFO_FIELD_NUM 2

/* use the same table structure of power-up current library */
ilib_table_t * leakage_ilib_table=NULL;


int
read_leakage_ilib(FILE* leakage_ilibFp, FILE* dbgFp)
{
	/* gate type, input pattern will be concatenated to index the lookup table */
	int generic_type_num=-1, fanin_num=-1, entry_num=-1;
	int flag, counter;
	int fanin_num_fit=-1;
	int fanin_field_bits=-1;
	int nin, type_index;
	char str[MAX_STR_BUF], generic_type_name[MAX_STR_BUF];

	if(leakage_ilibFp == NULL)
		return 1;   /* fail */

	flag = -1;
	counter = 0; /* required fields before reading data */
	do{
		if(fscanf(leakage_ilibFp, "%s", str)<=0)
			break;
		strupper(str);

		if(strcmp(str, "#")==0 || str[0] == '#')
			continue;
		if(strcmp(str, "GENERIC_TYPE_NUM=")==0)
		{
			fscanf(leakage_ilibFp, "%d", &generic_type_num);
			fprintf(stdout,"There are %d generic gate types in the lib file.\n",
							generic_type_num);

			fprintf(stdout, "The total supported # of types is %d.\n",GENERIC_TYPE_NUM_IN_ILIB);
			update_counter(&counter, &flag, LEAKAGE_REQUIRED_INFO_FIELD_NUM);
			entry_num = GENERIC_TYPE_NUM_IN_ILIB * fanin_num_fit;
			alloc_init_leakage_ilib(&leakage_ilib_table, entry_num, fanin_num, fanin_field_bits, flag);
		}else if(strcmp(str, "FANIN_NUM=")==0)
		{
			fscanf(leakage_ilibFp, "%d", &fanin_num);
			fanin_num_fit=fit_radix2_field(fanin_num, &fanin_field_bits);
			update_counter(&counter, &flag, LEAKAGE_REQUIRED_INFO_FIELD_NUM);
			entry_num = GENERIC_TYPE_NUM_IN_ILIB * fanin_num_fit;
			alloc_init_leakage_ilib(&leakage_ilib_table, entry_num, fanin_num, fanin_field_bits,flag);
		}else if(strcmp(str, "GATE=")==0)
		{
			while(strcmp(str, "ENDGATE")!=0)
			{
				/* get generic gate type */
				fscanf(leakage_ilibFp, "%s", str);
				split_gate_name_fields(str,&nin);
				strcpy(generic_type_name, str);
				type_index = generic_type_index(str, generic_type_index_table);
				if(nin<=0 || type_index<0)
				{
					fprintf(stderr, "Wrong with %s%d \n", str, nin);
					return 1;
				}

				toNextLine(leakage_ilibFp);
				if(read_leakage_value(leakage_ilibFp, leakage_ilib_table, 
										lookup_table_index(type_index, nin, fanin_field_bits,1,0)) == 1)
					return 1;
				fscanf(leakage_ilibFp, "%s", str);
			}
		}
	}while(toNextLine(leakage_ilibFp));

	/* debugging */
	dump_lkg_ilib(leakage_ilib_table, dbgFp);
	fflush(dbgFp);

	return 0; /* success */
}

/* allocate & initialize the leakage current lookup table */
void
alloc_init_leakage_ilib(ilib_table_t **table, int total_entry_num, int fanin_num, int fanin_field_bits, int flag)
{
	int i;
	int num;
	array_t *sub_array;

	if(flag > 0)
	{
		(*table) = (ilib_table_t *) mem_get_block ((SIZE_T) sizeof(ilib_table_t));
		(*table)->generic_gate_types = generic_type_index_table;
		(*table)->table_size = total_entry_num;
		(*table)->valid_entry_num = 0;
		(*table)->max_fanin_num = fanin_num;
		(*table)->max_fanout_num = -1; /* not used */

		(*table)->turnon_time_list = NULL; /* not used */
		(*table)->rule = (ilib_index_rule_t *) mem_get_block ((SIZE_T) sizeof(ilib_index_rule_t));
		(*table)->rule->fanin_field_bits = fanin_field_bits;
		(*table)->rule->fanout_field_bits = 0;

		(*table)->body=array_alloc(array_t *, total_entry_num);

		for(i=0; i < total_entry_num; i++)
		{
			num = (int) pow((double)2, (double)fanin_num);
			sub_array = array_alloc(float *, num);
			array_insert(array_t *, (*table)->body, i, sub_array);
		}
	}
}

/* read leakage current value and encoded input pattern in the table array */
int
read_leakage_value(FILE* fin, ilib_table_t *table, int index)
{
	int pattern;
	float current,*current_p;
	char ch;

	array_t *entry;

	ch=fgetc(fin);
	while(ch!='\n')
	{
		do{ ch=fgetc(fin);
		}while(ch=='\r' || ch == '\t' || ch == ' ');

		if(ch == '\n') continue;

		if(ch != '(')
		{
			fprintf(stderr,"Wrong ilib file format around %c.\n", ch);
			return 1;
		}else{
				fscanf(fin, "%d", &pattern);
				fscanf(fin, "%f", &current);
				current_p=(float*)mem_get_block((SIZE_T) sizeof(float));
				(*current_p) = current;
				do{ ch=fgetc(fin);
				}while(ch=='\r' || ch == '\t' || ch == ' ');
				if(ch != ')')
				{
                    fprintf(stderr,"Wrong ilib file format around %c.\n", ch);
                    return 1;
                }

			  }

		/* assign current value */
	    entry = array_fetch(array_t*, table->body, index);
	    array_insert(float*, entry, pattern, current_p);
	}
	table->valid_entry_num++;
	return 0;
}

/* dump leakage ilib to file */
void
dump_lkg_ilib(ilib_table_t *table, FILE *dbgFp)
{
	int i,j,nin,nout;
	array_t * entry;
	float * current_p;
	char type_name[MAX_STR_BUF];

	fprintf(dbgFp,"\n");
              fprintf(dbgFp,"   table size : %d \n", table->table_size);
    fprintf(dbgFp,"valid entries : %d \n", table->valid_entry_num);
    fprintf(dbgFp," max fanin #  : %d \n", table->max_fanin_num);

	for(i=0;i<array_n(table->body);i++)
	{
		entry = array_fetch(array_t *, table->body, i);

		if(array_n(entry)>0)
		{
			fprintf(dbgFp,"          index : %d  \n",i);
			index2gate(table, (unsigned)i, type_name, &nin, &nout);
			fprintf(dbgFp,"          name  : %s \n", type_name);
        	fprintf(dbgFp,"         fanin# : %d \n", nin);

			for(j=0;j<array_n(entry);j++)
			{
				current_p = array_fetch(float *, entry, j);
				fprintf(dbgFp,"    (%d %f) ", j, (*current_p));
			}
			fprintf(dbgFp,"\n");
		}
	}
}

/* accessLkgILIB_by_instance()
 *
 * access to leakge current library using the instance pointer
 */
double
accessLkgILIB_by_instance(ilib_table_t *table, PW_INST *inst)
{
  int pattern;
  int type_index,table_index;
  double current;

  if(inst->fanin > table->max_fanin_num)
  {  fprintf(stderr,"Instance:%s Unsupported fanin number %d.\n", inst->name, inst->fanin);
     fprintf(stderr,"You must map the logic first.\n");
     return -1;
  }
  pattern = encoded_input_pattern(inst);
  switch(inst->instLibNo)
  {
    case NOT_LIB_NO:
      type_index = generic_type_index("NOT", generic_type_index_table);
      break;
    case AND_LIB_NO:
      type_index = generic_type_index("AND", generic_type_index_table);
      break;
    case NAND_LIB_NO:
      type_index = generic_type_index("NAND", generic_type_index_table);
      break;
    case OR_LIB_NO:
      type_index = generic_type_index("OR", generic_type_index_table);
      break;
    case NOR_LIB_NO:
      type_index = generic_type_index("NOR", generic_type_index_table);
      break;
    case PI_LIB_NO: /* PI & PPI are treated as buffers, for spice sim comparsion */
    case BUFF_LIB_NO:
      type_index = generic_type_index("BUFF", generic_type_index_table);
      table_index = lookup_table_index(type_index, 1, table->rule->fanin_field_bits, 1, 0);
      current = lookup_lkg_ilib_table(table, table_index, pattern);
      return current;
      break;
    case PO_LIB_NO:
      return 0; /* PO is not considered for leakage current */
    default:
      fprintf(stderr,"Not supported gate type: %s\n", PWdeviceName[inst->instLibNo]);
      fprintf(stderr,"You must map the logic first.\n");
      return -1;
  }

  /* for real gates  & PIs*/
  table_index = lookup_table_index(type_index, inst->fanin, table->rule->fanin_field_bits, 1, 0);
  current = lookup_lkg_ilib_table(table, table_index, pattern);

  return current;
}

/* lookup_ilib_table().
 *
 * lookup the leakge current ilib table.
 */
double
lookup_lkg_ilib_table(ilib_table_t * table, int table_index, int pattern)
{
  array_t * arr;
  float *p;

  arr = array_fetch(array_t *, table->body, table_index);
  p = array_fetch(float*, arr, pattern);

  return (double) (*p);
}

int
encoded_input_pattern(PW_INST *inst)
{
  int i;
  int pattern=0;

  if(inst->instLibNo == PI_LIB_NO) /* input value is its output value */
  {
    return logic2int(fetchImplyValueInStk(inst));
  }else{
         for(i=0;i<inst->fanin;i++)
         {
           pattern += logic2int(fetchImplyValueInStk(inst->in[i])) << i;
         }
         return pattern;
       }
}
