/*
 * mape_curitr.c
 *
 * iterative improvement algorithm for maximum current estimation. 
 *
 * Changbo Long.  Jan. 14, 2002. 
 */
#include "mape_int.h"

#define MAX_FETCH_TIMES		10
#define MAX_NO_EFFECT_LOOP	500
#define RDMSIM_TIMES		1800

static lsList untran_list = NULL ;
static lsList backtrace_list = NULL ;
static lsList noeffect_list = NULL ;
static lsList change_list = NULL ;
static int stop_times = 0 ;

extern void mape_curitr_dump_input(network_t *) ;
void mape_curitr_clear(network_t *) ;

/*
 * mape_curitr_max_noeffect_loop
 *
 *
 * return the maximum no_effective loop times 
 */
int 
mape_curitr_max_noeffect_loop(network)
network_t *network ;
{
	return MIN(network_num_pi(network) * 13, MAX_NO_EFFECT_LOOP) ;
}

/*
 * mape_curitr_refine_stop
 *
 * indicate if to stop the refine process. 
 */
int 
mape_curitr_refine_stop(network, reset)
network_t *network ; 
int reset ;
{
	if(reset > 0)
		stop_times = 0 ;
	else 
		stop_times ++ ;

	if(stop_times >= mape_curitr_max_noeffect_loop(network))
		return TRUE ;
	else 
		return FALSE ;
}

/*
 * mape_curitr_push_change
 *
 *
 * push node into the change's list. 
 */
void 
mape_curitr_push_change(node)
node_t *node ;
{
	if(node == NULL)
		return ;
	if(!mape_is_given_node(node))
		return ;
	if(change_list == NULL)
		change_list = lsCreate() ;

	if(node_num_fanout(node) > 0 )
		lsNewEnd(change_list, (lsGeneric)node, LS_NH) ;
}

/*
 * mape_curitr_clear_change_list. 
 *
 * clear all the node inside the change's list. 
 */
void 
mape_curitr_clear_change_list(void)
{
	node_t *tmp ;

	if(!change_list)
		return ;

	while(lsLength(change_list))
		lsDelEnd(change_list, (lsGeneric *)&tmp) ;
}

/*
 * mape_curitr_push_noeffect
 *
 * push node into the noeffect list. 
 */
void 
mape_curitr_push_noeffect(node)
node_t *node ;
{
	if(!noeffect_list)
		noeffect_list = lsCreate() ;
	lsNewEnd(noeffect_list, (lsGeneric)node, LS_NH) ;
}

/*
 * mape_curitr_clear_noeffect
 *
 * clear the noeffect list. 
 */
void 
mape_curitr_clear_noeffect(void)
{
	node_t *node ;

	if(!noeffect_list)
		return ;

	while(lsLength(noeffect_list))
		lsDelEnd(noeffect_list, (lsGeneric *)&node) ;
}

/*
 * mape_itr_push_fetch
 *
 *
 * add the node's fetch times. 
 */
void 
mape_curitr_push_fetch(node)
node_t *node ;
{
	if(node == NULL)
		return ;
	if(!(node->mape))
		node->mape = (char *)mape_create_mape_t() ;

	((mape_t *)(node->mape))->visit ++ ;
}

/*
 * mape_curitr_clear_backtrace. 
 *
 * clear backtrace list. 
 */
void mape_curitr_clear_backtrace(inc_fetch)
int inc_fetch ;
{
	node_t *node ;

	while(lsLength(backtrace_list)){
		lsDelEnd(backtrace_list, (lsGeneric *)&node) ;
		if(inc_fetch)
			mape_curitr_push_fetch(node) ;
	}
}


/*
 * mape_curitr_push_backtrace.
 *
 * push node into the backtrace list. 
 */
void
mape_curitr_push_backtrace(node)
node_t *node ;
{
	if(!backtrace_list)
		backtrace_list = lsCreate() ;

	lsNewEnd(backtrace_list, (lsGeneric)node, LS_NH) ;
}

/*
 * mape_curitr_backtrace
 *
 * backtrace to a pi. 
 */
node_t *
mape_curitr_backtrace(node)
node_t *node ;
{
	int i ;
	node_t *p, *next = NULL ;
	int pri = -1, new_pri = -1 ;

	mape_curitr_push_backtrace(node) ;

	if(node_type(node) == PRIMARY_INPUT)
		return node ;

	foreach_fanin(node, i, p){
		new_pri = mape_itr_node_priority(p) ;
		if(!next || new_pri > pri){
			next = p ;
			pri = new_pri ;
		}
	}
	return mape_curitr_backtrace(next) ;
}

/*
 * mape_curitr_fetch_times
 *
 * find node's fetch times. 
 */
int 
mape_curitr_fetch_times(node)
node_t *node ;
{
	if(node == NULL)
		return 0 ;
	if(!(node->mape))
		return 0 ;

	return ((mape_t *)(node->mape))->visit ;
}

/*
 * mape_curitr_node_sort_value. 
 *
 * sort node's priority value. 
 */
int
mape_curitr_node_sort_value(node, time)
node_t *node ;
double time ;
{
	/*return node_num_fanout(node) - mape_curitr_fetch_times(node) - 2*mape_get_node_level(node); */

	if(fabs(mape_cur_top_node_delay(node) - time) >= SLOP_TIME)
		return node_num_fanout(node) - mape_curitr_fetch_times(node) - 2*mape_get_node_level(node) ;
	else 
		return  node_num_fanout(node) - mape_curitr_fetch_times(node) - 2*mape_get_node_level(node) + 10 ;
}

/*
 * mape_curitr_push_untran
 *
 *
 * store the node without transition. 
 */
void 
mape_curitr_push_untran(node)
node_t *node ;
{
	if(node == NULL)
		return ;
	if(node_num_fanout(node) < 1)
		return ;
	if(!mape_is_given_node(node))
		return ;

	if(untran_list == NULL)
		untran_list = lsCreate() ;
	lsNewEnd(untran_list, (lsGeneric)node, LS_NH) ;
}

/*
 * mape_curitr_remove_untran
 *
 * remove a node from the untran list. 
 */
void 
mape_curitr_remove_untran(node)
node_t *node ;
{
	lsGen gen ;
	lsHandle h ;
	node_t *p, *tmp ;

	if(!untran_list)
		return ;

	lsForeachItem(untran_list, gen, p)
		if(p == node)
			lsDelBefore(gen, (lsGeneric *)&tmp) ;
}

/*
 * mape_curitr_update_untran_list
 *
 * update the untransition list. 
 */
void 
mape_curitr_update_untran_list(void)
{
	node_t *node, *tmp ;

	if(!change_list)
		return ;

	if(!untran_list)
		untran_list = lsCreate() ;

	while(lsLength(change_list)){
		lsDelEnd(change_list, (lsGeneric *)&node) ;

		if(mape_curitr_is_trans(node))
			lsNewEnd(untran_list, (lsGeneric)node, LS_NH) ;
		else 
			mape_curitr_remove_untran(node) ;
	}
}

/*
 * mape_curitr_fetch_untran
 *
 *
 * fetch node from the untran_list which has the highest 
 * priority
 */
node_t *
mape_curitr_fetch_untran(f, prev_node, time)
int (*f)() ;
node_t *prev_node ;
double time ;
{
	lsGen gen ;
	int pri = -99, new_pri = -99 ;
	int i ;
	node_t *p, *rtv = NULL ;
	lsHandle ht, h ;

	if(untran_list == NULL)
		return NULL ;

	if(lsLength(untran_list) == 0){
		fprintf(mapeout, "Warning: untran_list is empty\n") ; 
		return NULL ;
	}

	gen = lsStart(untran_list) ;
	while(lsNext(gen, (lsGeneric *)&p, &h) == LS_OK){
		new_pri = (int)((*f)(p, time)) ;
		if(new_pri > pri){
			pri = new_pri ;
			ht = h ;
			rtv = p ;
		}
	}
	lsFinish(gen) ;
	return rtv ;
}

/*
 * mape_curitr_is_trans
 *
 * is the node a transistion node. 
 */
int 
mape_curitr_is_trans(node)
node_t *node ;
{
	lsGen gen ;
	lsList vol_points ;
	mape_vol_point_t *prev = NULL ;
	mape_vol_point_t *curr ;

	vol_points = mape_cur_top_vol_points(node) ;

	if(vol_points == NULL)
		return FALSE ;

	lsForeachItem(vol_points, gen, curr){
		if(prev == NULL){
			prev = curr ;
			continue ;
		}
		if(prev != NULL){
			if(prev->value != curr->value){
				lsFinish(gen) ;
				return TRUE ;
			}else {
				prev = curr ;
			}
		}
	}
	return FALSE ;
}

/*
 * mape_curitr_backtrack_do
 *
 * for mape_imp utility, using for backtrack. 
 */
int 
mape_curitr_backtrack_do(node, arg)
node_t *node ;
char *arg ;
{
	lsList tl ;
	mape_cur_node_value_t *nv ;
	node_t *pi ;
	mape_node_value_t pi_value ;
	mape_delay_value_t *top_value ;
	lsHandle h ;

	pi = ((mape_nv_pair_t *)arg)->node ;
	pi_value = ((mape_nv_pair_t *)arg)->value ;

	nv = mape_cur_top_node_value(node) ;
	if(nv == NULL)
		mape_error("mape_curitr_backtrack_do", __FILE__, __LINE__) ;


	if(nv->pi == pi && nv->pi_value == pi_value){
		//fprintf(mapeout, "release node %s\n", node->name) ;
		tl = mape_cur_pop_vol_points(node) ;
		lsDestroy(tl, free) ;
		return TRUE ;
	}else {
		return FALSE;
	}
}

/*
 * mape_curitr_backtrack_change. 
 *
 * backtrack the change of the PI. 
 */
void 
mape_curitr_backtrack_change(pi)
node_t *pi ;
{
	mape_nv_pair_t pi_pair ;
	mape_cur_node_value_t *top ;
	lsHandle h ;

	top = mape_cur_top_node_value(pi) ;
	if(top == NULL)
		mape_error("mape_curitr_backtrack_change!", __FILE__, __LINE__) ;

	pi_pair.node = pi ;
	pi_pair.value = top->pi_value ;

	mape_imp_reset(mape_curitr_backtrack_do, (char *)(&pi_pair)) ;
	mape_imp_push(pi) ;
	mape_imp_start() ;
}


/*
 * mape_curitr_imp_do
 *
 * implication of the input vector. 
 */
int 
mape_curitr_imp_do(node, arg)
node_t *node ;
char *arg ; /*no use*/
{
	mape_cur_simulate_node_v(node, node, MAPE_0_0) ;

	if(!mape_curitr_is_trans(node))
		mape_curitr_push_untran(node) ;

	return TRUE ;
}

/*
 * mape_curitr_init
 *
 * initialize the maximum estimation
 * with some special input vector. 
 */
void 
mape_curitr_init(network, is_fall)
network_t *network ;
int is_fall ;
{
	lsGen gen ;
	node_t *pi ;

	mape_imp_reset(mape_curitr_imp_do, (char *)NULL) ;
	if(is_fall){
		foreach_primary_input(network, gen, pi){
			mape_cur_push_vol_point(pi, (double)MAX_VOLTAGE, (double)0.0, 1) ;
			mape_cur_push_vol_point(pi, (double)0.0, (double)SLOP_TIME, 1) ;
			mape_cur_set_change_pi(pi, pi, MAPE_FALL) ;
			mape_imp_push(pi) ;
		}
	}else {
		foreach_primary_input(network, gen, pi){
			mape_cur_push_vol_point(pi, (double)0.0, (double)0.0, 1) ;
			mape_cur_push_vol_point(pi, (double)MAX_VOLTAGE, (double)SLOP_TIME, 1) ;
			mape_cur_set_change_pi(pi, pi, MAPE_RISE) ;
			mape_imp_push(pi) ;
		}
	}
	mape_imp_start() ;
}

/*
 * mape_curitr_next_value
 *
 * get pi's next value ;
 */
mape_node_value_t 
mape_curitr_pi_next_value(pi)
node_t *pi ;
{
	lsList tl ;
	mape_vol_point_t *top_value ;
	lsHandle h ;

	tl = mape_cur_top_vol_points(pi) ;
	if(tl == NULL)
		return MAPE_FALL ;

	lsFirstItem(tl, (lsGeneric *)&top_value, &h) ;
	if(top_value->value == 0.0){
		return MAPE_FALL ;
	}else 
		return MAPE_RISE ;
}

/*
 * mape_curitr_is_trans_equ
 *
 * is two transition list equal. 
 */
int 
mape_curitr_vol_points_equ(l1, l2)
lsList l1, l2 ;
{
	lsGen gen1, gen2;
	mape_vol_point_t *dv1, *dv2 ;
	lsHandle h1, h2 ;
	int rtv = TRUE ;

	if(l1 == NULL || l2 == NULL)
		return FALSE ;

	if(lsLength(l1) != lsLength(l2))
		return FALSE ;

	gen1 = lsStart(l1) ;
	gen2 = lsStart(l2) ;

	while(lsNext(gen1, (lsGeneric *)&dv1, &h1) == LS_OK && 
	      lsNext(gen2, (lsGeneric *)&dv2, &h2) == LS_OK ){
	        if(dv1->time != dv2->time || dv1->value != dv2->value || dv1->multi != dv2->multi) {
			rtv = FALSE ;
			break ;
		}
	      }
	lsFinish(gen1) ;
	lsFinish(gen2) ;

	return rtv ;
}

/*
 * mape_curitr_change_do
 *
 *
 * change the pi's value 
 */
int
mape_curitr_change_do(node, arg)
node_t *node ;
char *arg  ;
{
	mape_nv_pair_t *pi_pair = (mape_nv_pair_t *)arg ;
	lsList new_value, old_value;
	int ot, nt ;
	int i ;

	if(node == NULL)
		return FALSE ;
	if(node_type(node) == PRIMARY_INPUT)
		return TRUE ;

	old_value = mape_cur_top_vol_points(node) ;
	ot = mape_curitr_is_trans(node) ;

	mape_cur_new_node_value(node) ;
	mape_cur_simulate_node_v(node, pi_pair->node, pi_pair->value);

	new_value = mape_cur_top_vol_points(node) ;
	nt = mape_curitr_is_trans(node) ;

	if(mape_curitr_vol_points_equ(old_value, new_value)){
		return FALSE ;
	}
	
	if(ot != nt)
		mape_curitr_push_change(node) ;
	return TRUE ;
}

/*
 * mape_curitr_change. 
 *
 * change the pi's value. 
 */
void 
mape_curitr_change_pi(pi)
node_t *pi ;
{
	mape_nv_pair_t pi_pair ;
	mape_node_value_t old_value, new_value ;

	new_value = mape_curitr_pi_next_value(pi) ;

	pi_pair.node = pi ;

	mape_cur_new_node_value(pi) ;
	/*
	mape_cur_push_trans_value(pi, new_value, 0.0, pi, new_value) ;
	*/
	if(new_value == MAPE_FALL){
		mape_cur_push_vol_point(pi, (double)MAX_VOLTAGE, (double)0.0, 1) ;
		mape_cur_push_vol_point(pi, (double)0.0, (double)SLOP_TIME, 1) ;
		mape_cur_set_change_pi(pi, pi, MAPE_FALL) ;
		pi_pair.value = MAPE_FALL ;
	}else {
		mape_cur_push_vol_point(pi, (double)0.0, (double)0.0, 1) ;
		mape_cur_push_vol_point(pi, (double)MAX_VOLTAGE, (double)SLOP_TIME, 1) ;
		mape_cur_set_change_pi(pi, pi, MAPE_RISE) ;
		pi_pair.value = MAPE_RISE ;
	}
	mape_imp_reset(mape_curitr_change_do, (char *)(&pi_pair)) ;
	//fprintf(mapeout, "new node value: %s\n", pi->name) ;
	mape_imp_push(pi) ;
	mape_imp_start() ;
}

/*
 * mape_curitr_refine_rdm
 *
 * random refine for inprovment. 
 */
double
mape_curitr_refine_rdm(network)
network_t *network ;
{
	double ocur, ncur ;
	int i ;
	double time ;
	int times ;
	int noi_times ; 


	/*
	ocur = mape_cur_max_current(network, &time, NULL) ;
	fprintf(mapeout, "Initially set the maximum value to be %.2lf\n",ocur) ; 
	*/

	/*
	times = 600*(int)(sqrt(network_num_internal(network) + network_num_pi(network))) ;
	fprintf(mapeout, "Simulation times: %d\n", times) ;
	*/
	ocur = 0 ;
	times = 0 ;
	noi_times = 0;
	for(;;){
		if(times >= 10000 && noi_times >= 2000)
			break ;
		mape_cur_clear(network) ;
		mape_cur_initialize_random(network) ;
		ncur = mape_cur_max_current(network, &time, NULL) ;
		if(fabs(ncur) > fabs(ocur)){
			ocur = ncur ;
			noi_times = 0 ;
		}else 
			noi_times ++ ;
		times ++ ;
	}
	fprintf(mapeout, "Simulation times: %d\n", times) ;
	return ocur ;
}

/*
 * mape_curitr_refine 
 *
 * refine the estimation result until
 * get the maximum current estimation. 
 */
void 
mape_curitr_refine(network)
network_t *network ;
{
	node_t *node ;
	node_t *pi ;
	node_t *prev_node = NULL;
	double ocur, ncur ;
	double otime, ntime ;
	int i ;

	ocur = mape_cur_max_current(network, &otime, NULL) ;

	for(;;){
		if((node = mape_curitr_fetch_untran(mape_curitr_node_sort_value, prev_node, otime)) == NULL) 
			break;
		pi = mape_curitr_backtrace(node) ;
		/*
		fprintf(mapeout, "backtrace from node %s(%d) to pi %s\n", node->name, node_num_fanout(node), pi->name) ;
		*/
		mape_curitr_change_pi(pi) ;
		ncur = mape_cur_max_current(network, &ntime, NULL) ;
		//fprintf(mapeout, "Maximum current estimation: %.2lf, time = %.2lf\n", ncur, ntime) ;
		if(fabs(ncur) > fabs(ocur)){ /*improved*/
			ocur = ncur ;
			mape_curitr_update_untran_list() ;
			prev_node = NULL ;
			mape_curitr_clear_backtrace(FALSE) ;
			mape_curitr_clear_noeffect() ;
			otime = ntime ;
		}else{ /*not improved*/
			mape_curitr_backtrack_change(pi) ;	
			mape_curitr_clear_change_list() ;
			prev_node = node ;
			mape_curitr_clear_backtrace(TRUE) ;
			mape_curitr_push_noeffect(node) ;
		}
		if(mape_curitr_refine_stop(network))
			break ;
	}
}

/* 
 * mape_currdm
 *
 * randomly get current
 */
double
mape_currdm(network, time, switch_list)
network_t *network ;
double *time ;
lsList switch_list ;
{
	double max_cur ;

	srandom(util_cpu_time()) ;
	mape_curlib_set_various_mode(TRUE) ;

	mape_timer_reset() ;
	mape_timer_start() ;
	max_cur = mape_curitr_refine_rdm(network) ;
	mape_timer_stop() ;
	mape_timer_report("Random simulation time: ") ; 
	mape_curitr_clear(network) ;
	fprintf(mapeout, "Maximum current estimation: %.2lf\n", max_cur) ;
	return max_cur ;
}

/*
 * mape_currdm_grp
 *
 * random simulation get maximum current. 
 */
void 
mape_currdm_grp(network) 
network_t *network ;
{
	lsList groups, group ;
	lsGen gen, gen1, gen2 ;
	node_t *node ;
	int times ;
	int i, j ;
	double x[2048] ;
	double ncur ;
	double time ;

	groups = mape_group_get_groups(network) ;
	srandom(util_cpu_time()) ;
	mape_curlib_set_various_mode(TRUE) ;

	/*
	mape_cur_set_cone_mode(TRUE) ;
	*/

	mape_timer_reset() ;
	mape_timer_start() ;
	memset (x, '0',  sizeof(double)*2048) ;
	times = 500 ;
	for(i=0;i<=times;i++){
		if(i%500 == 0)
			fprintf(mapeout, "simulation times: %d/%d\n", i, times) ;
		mape_cur_clear(network) ;
		mape_cur_initialize_random(network) ;
		j = 0 ;
		lsForeachItem(groups, gen1, group){
			j++ ;
			if(j%GROUP_SAMPLE_RATE != 0)
				continue ;
			mape_clean_given_gate() ;
			lsForeachItem(group, gen2, node)
				mape_push_given_gatep(node) ;
			ncur = mape_cur_max_current(network, &time, NULL) ;
			if(fabs(ncur) > fabs(x[j]))
				x[j] = ncur ;
		}
	}
	mape_timer_stop() ;
	mape_timer_report("random simulation 50000 times") ;
	fprintf(mapeout, "Simulation times: %d\n", times) ;
	i = 0 ;
	lsForeachItem(groups, gen1, group){
		i++ ;
		if(i%GROUP_SAMPLE_RATE != 0)
			continue ;
		lsForeachItem(group, gen2, node)
			fprintf(mapeout, "%s ", node->name) ;
		fprintf(mapeout, "\n") ;
		fprintf(mapeout, "Maximum current estimation: %.2lf\n", x[i]) ;
	}
	

	/*
	i = 0 ;
	lsForeachItem(groups, gen1, group){
		i++ ;
		if(i%5 != 0)
			continue ;
		mape_cur_group_cone(group) ;
		mape_clean_given_gate() ;
		lsForeachItem(group, gen2, node){
			mape_push_given_gatep(node) ;
			fprintf(mapeout, "%s ", node->name) ;
		}
		fprintf(mapeout, "\n") ;
		mape_currdm(network, &time, NULL) ;
	}
	*/
}

/*
 * mape_curitr
 *
 * iteratively get the maximum current.  
 */
double
mape_curitr(network, time, switch_list)
network_t *network ;
double *time ;
lsList switch_list ;
{
	double max_cur = 0.0 ;
	double rise_cur = 0.0 ;
	double fall_cur = 0.0 ;

	mape_curlib_set_unit_mode(TRUE) ;

	mape_cur_initialize(network, FALSE) ;
	rise_cur = mape_cur_max_current(network, time, NULL) ;
	//fprintf(mapeout, "Get initial maximum current : %lf\n", rise_cur) ;
	mape_curitr_clear(network) ;
	mape_cur_initialize(network, TRUE) ;
	fall_cur = mape_cur_max_current(network, time, NULL) ;
	mape_curitr_clear(network) ;
	//fprintf(mapeout, "Get initial maximum current : %lf\n", fall_cur) ;

	if(fabs(rise_cur) > fabs(fall_cur))
		mape_curitr_init(network, FALSE) ;
	else 
		mape_curitr_init(network, TRUE) ;

	mape_curitr_refine(network) ;
	max_cur = mape_cur_max_current(network, time, switch_list) ;

	mape_curitr_clear(network) ;
	fprintf(mapeout, "Maximum current estimation: %.2lf\n", max_cur) ;
	return max_cur ;
}

/*
 * mape_curitr_pi_value
 *
 * return pi's value
 */
mape_node_value_t 
mape_curitr_pi_value(pi)
node_t *pi ;
{
	lsList tl ;
	mape_vol_point_t *top_value ;
	lsHandle h ;
	lsGen gen ;

	tl = mape_cur_top_vol_points(pi) ;
	lsFirstItem(tl, (lsGeneric *)&top_value, &h) ;
	if(top_value->value == 0.0)
		return MAPE_RISE ;
	else 	
		return MAPE_FALL ;
}

/*
 * mape_curitr_dump_input
 *
 *
 * dump input of the network. 
 */
void
mape_curitr_dump_input(network)
network_t *network ;
{
	lsList tl ;
	mape_delay_value_t *top_value ;
	lsHandle h ;
	lsGen gen ;
	node_t *pi ;

	foreach_primary_input(network, gen, pi){
		if(mape_curitr_pi_value(pi) == MAPE_FALL)
			fprintf(mapeout, "F") ;
		else 
			fprintf(mapeout, "R") ;
	}
	fprintf(mapeout, "\n") ;
}

/*
 * mape_curitr_node_clear
 *
 * clear the node
 */
void 
mape_curitr_node_clear(node)
node_t *node ;
{
	lsList cur_values ;
	mape_cur_node_value_t *cur_value ;

	if(!(node->mape))
		return ;
	
	cur_values = ((mape_t *)(node->mape))->cur_values ;

	if(!cur_values)
		return ;

	while(lsLength(cur_values)){
		lsDelEnd(cur_values, (lsGeneric *)&cur_value) ; 
		/*
		if(cur_value->trans_values)
			lsDestroy(cur_value->trans_values, free) ;
		*/
		if(cur_value->vol_points)
			lsDestroy(cur_value->vol_points, free) ;
		if(cur_value->trans_values)
			lsDestroy(cur_value->trans_values, free) ;
		free(cur_value) ;
	}
	lsDestroy(cur_values, NULL) ;
	((mape_t *)(node->mape))->cur_values = NULL ;
	((mape_t *)(node->mape))->visit = 0 ;

	if(NODE_SIM_CURR(node)){
		lsFinish(NODE_SIM_CURR(node)) ;
		NODE_SIM_CURR(node) = NULL ;
	}
}

/*
 * mape_curitr_clear
 *
 * clear all the temperate information. 
 */
void 
mape_curitr_clear(network)
network_t *network ;
{
	lsGen gen ;
	node_t *node , *pi;

	
	foreach_node(network, gen, node)
		mape_curitr_node_clear(node) ;

	stop_times = 0 ;
	if(untran_list){
		lsDestroy(untran_list, NULL) ;
		untran_list = NULL;
	}
	if(backtrace_list){
		lsDestroy(backtrace_list, NULL) ;
		backtrace_list = NULL ;
	}
	if(noeffect_list){
		lsDestroy(noeffect_list, NULL) ;
		noeffect_list = NULL ;
	}
	if(change_list){
		lsDestroy(change_list, NULL) ;
		change_list = NULL ;
	}
}
