/*
 * mape_curlib.c 
 *
 * For current library. 
 * Changbo Long (Dec. 16, 2001). 
 */
#include "mape_int.h"

#define CURLIB_FILENAME "simple.curlib7"
#define CURLIB_MAX_INSTANCE 128
#define CURLIB_MAX_GATE	    56

#define CURLIB_UNIT_DELAY	25.0 	/*unit delay time*/
#define CURLIB_UNIT_CURRENT	-180.0 	/*unit current value */
#define CURLIB_UNIT_POS		12.5	/*peak value position*/
#define CURLIB_UNIT_DISTANCE	50.0	/*idis value*/

lsList cur_lib = NULL;
int unit_mode = FALSE ;
int various_mode = FALSE ;

extern void mape_curlib_build_lib(void) ;
extern void mape_curlib_set_unit_mode(int) ;

/* Spline iterpolation routines, adapted from Numerical Recipes */

/*
 * malloc a vector. 
 */
double *mape_vector1(nl, nh)
int nl, nh ;
{
	double *v ;

	v=(double *)ALLOC(unsigned, (nh-nl+1)*sizeof(double)) ;
	if(!v)
		mape_error("Memory allocation failure !", __FILE__, __LINE__) ;
	return v-nl ;
}

/*
 * free a vector. 
 */
void mape_free_vector1(v, nl, nh)
double *v ;
int nl, nh ;
{
	free((double *) (v+nl)) ;
}

/*
 * mape_curlib_set_unit_mode
 *
 * set the delay to be unit mode. 
 */
void 
mape_curlib_set_unit_mode(v)
int v ;
{
	unit_mode = v ;
}

/*
 * mape_curlib_set_various_mode
 *
 * in various delay mode, the rising and falling 
 * delay is the same for a gate, but they are different 
 * in different gates. 
 */
void 
mape_curlib_set_various_mode(v)
{
	various_mode = v ;
}

/*
 * initialize the coeffient. 
 */
void spline(x, y, n, yp1, ypn, y2a)
double *x, *y ;
int n ;
double yp1, ypn ;
double *y2a ;
{
	int i,k;
	double p,qn,sig,un,*u;

	u=mape_vector1(1,n-1);

	if (yp1 > 0.99e30) 
		y2a[1]=u[1]=0.0;
	else {
		y2a[1] = -0.5;
		u[1]=(3.0/(x[2]-x[1]))*((y[2]-y[1])/(x[2]-x[1])-yp1);
	}

	for(i=2;i<=n-1;i++)
	{
		sig=(x[i]-x[i-1])/(x[i+1]-x[i-1]);
		p=sig*y2a[i-1]+2.0;
		y2a[i]=(sig-1.0)/p;
		u[i]=(y[i+1]-y[i])/(x[i+1]-x[i]) - (y[i]-y[i-1])/(x[i]-x[i-1]);
		u[i]=(6.0*u[i]/(x[i+1]-x[i-1])-sig*u[i-1])/p;
	}

	if (ypn > 0.99e30) 
		qn=un=0.0;
	else {
		qn=0.5;
		un=(3.0/(x[n]-x[n-1]))*(ypn-(y[n]-y[n-1])/(x[n]-x[n-1]));
	}

	y2a[n]=(un-qn*u[n-1])/(qn*y2a[n-1]+1.0);
	for (k=n-1;k>=1;k--) 
		y2a[k]=y2a[k]*y2a[k+1]+u[k];

	mape_free_vector1(u,1,n-1);
}

/*
 * interplate the result. 
 */
double splint(xa, ya, y2a, n, x)
double *xa, *ya, *y2a ;
int n ;
double x ;
{
	int klo, khi, k;
	double h,b,a,y;

	klo=1;
	khi=n;

	while(khi-klo>1) {
		k=(khi+klo) >> 1;

		if(xa[k]>x) 
			khi=k;
		else 
			klo=k;
	}

	h=xa[khi]-xa[klo];
	if(h==0.0) {
		mape_error("Interpolate", __FILE__, __LINE__) ;
	}

	a=(xa[khi]-x)/h;
	b=(x-xa[klo])/h;
	return(a*ya[klo]+b*ya[khi]+((a*a*a-a)*y2a[klo]+(b*b*b-b)*y2a[khi])*(h*h)/6.0);
}

/*
 * void 
 * spline test. 
 */
void 
spline_test(void)
{
	int n=10 ;
	double x[10]={1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0} ;
	double y[10]={0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5} ;
	double y2a[10] ;

	/* hehe, strange usage of the array :) */
	spline(x-1, y-1, n, 1e+30, 1e+30, y2a-1) ;
	fprintf(mapeout, "spline_test: interplate x=5.5, y=%.2lf\n", splint(x-1, y-1, y2a-1, n, 5.5)) ;
}

/*
 * mape_curlib_create_param
 *
 * create the parameter 
 */
mape_curlib_param_t *
mape_curlib_create_param(void)
{
	mape_curlib_param_t *rtv ;

	rtv 		= ALLOC(mape_curlib_param_t, 1) ;
	rtv->x 		= ALLOC(double, CURLIB_MAX_INSTANCE) ;
	rtv->y 		= ALLOC(double, CURLIB_MAX_INSTANCE) ;
	rtv->y2a 	= ALLOC(double, CURLIB_MAX_INSTANCE) ;

	return rtv ;
}

/*
 * mape_curlib_find_gate
 *
 * find the gate in the current library. 
 */
mape_curlib_gate_t *
mape_curlib_find_gate(name)
char *name ;
{
	lsGen gen ;
	mape_curlib_gate_t *rtv = NULL, *gate, *first = NULL ;
	char tmp[128] ;

	if(name == 0)
		mape_error("can't find gate's corresponding curlib\n", __FILE__, __LINE__) ;

	if(!cur_lib)
		mape_curlib_build_lib() ;

	lsForeachItem(cur_lib, gen, gate){
		if(!first)
			first = gate ;
		if(strcmp(name, gate->name) == 0)
			rtv = gate ;
	}
	if(rtv)
		return rtv ;
	else {
		sprintf(tmp, "can't find gate %s's corresponding curlib, return default\n", name) ;
		fprintf(stderr, "%s\n", tmp) ;
		return first ;
		/*
		sprintf(tmp, "can't find gate %s's corresponding curlib\n", name) ;
		fprintf(stderr, "%s\n", tmp) ;
		return NULL ;
		*/
	}
}

/*
 * mape_curlib_create_loadg_t
 *
 *
 * create the mape_curlib_loadg_t structure. 
 */
mape_curlib_loadg_t *
mape_curlib_create_loadg_t(void)
{
	mape_curlib_loadg_t *rtv ;	

	rtv 		= ALLOC(mape_curlib_loadg_t, 1) ; 
	rtv->gate	= ALLOC(char *, CURLIB_MAX_GATE) ;
	memset(rtv->gate, 0, sizeof(char *)*CURLIB_MAX_GATE) ;

	rtv->rdlay	= ALLOC(double, CURLIB_MAX_GATE) ;
	rtv->fdlay	= ALLOC(double, CURLIB_MAX_GATE) ;
	rtv->irmax	= ALLOC(double, CURLIB_MAX_GATE) ;
	rtv->ifmax	= ALLOC(double, CURLIB_MAX_GATE) ;
	rtv->irpos	= ALLOC(double, CURLIB_MAX_GATE) ;
	rtv->ifpos	= ALLOC(double, CURLIB_MAX_GATE) ;
	rtv->irdis	= ALLOC(double, CURLIB_MAX_GATE) ;
	rtv->ifdis	= ALLOC(double, CURLIB_MAX_GATE) ;

	return rtv ;
}

/*
 * mape_curlib_create_gate. 
 *
 * create a curlib gate. 
 */
mape_curlib_gate_t *
mape_curlib_create_gate(void)
{
	mape_curlib_gate_t *rtv ;
	
	rtv 		= ALLOC(mape_curlib_gate_t, 1) ;
	rtv->n 		= 0 ;
	rtv->effin	= 0.0 ;
	rtv->loadg	= mape_curlib_create_loadg_t();
	rtv->rdlay	= mape_curlib_create_param();
	rtv->fdlay	= mape_curlib_create_param();
	rtv->irmax	= mape_curlib_create_param();
	rtv->ifmax	= mape_curlib_create_param();
	rtv->irpos	= mape_curlib_create_param();
	rtv->ifpos	= mape_curlib_create_param();
	rtv->irdis	= mape_curlib_create_param();
	rtv->ifdis	= mape_curlib_create_param();

	return rtv ;
}

/*
 * mape_curlib_push_curlib_list
 *
 * push one gate into the curlib list. 
 */
void 
mape_curlib_push_curlib_list(gate)
mape_curlib_gate_t *gate ;
{
	if(!cur_lib)
		cur_lib = lsCreate() ;

	lsNewEnd(cur_lib, (lsGeneric)gate, LS_NH) ;
}

/*
 * mape_curlib_param_value
 *
 * get the param's value with certain loading capcitance. 
 */
double 
mape_curlib_param_value(param, n, cload)
mape_curlib_param_t *param; 
int n ;
double cload ;
{
	return splint((param->x)-1, (param->y)-1, (param->y2a)-1, n, cload) ;
}

/*
 * mape_curlib_unit_value
 *
 * return the unit value for some type. 
 */
double 
mape_curlib_unit_value(type)
mape_curlib_param_type_t type ;
{
	switch(type){
		case MAPE_CURLIB_RDLAY: 
			return CURLIB_UNIT_DELAY ;
		case MAPE_CURLIB_FDLAY: 
			return CURLIB_UNIT_DELAY ;
		case MAPE_CURLIB_IRMAX: 
			return CURLIB_UNIT_CURRENT ;
		case MAPE_CURLIB_IFMAX: 
			return 0.0 ;
		case MAPE_CURLIB_IRPOS: 
			return CURLIB_UNIT_POS ;
		case MAPE_CURLIB_IFPOS: 
			return CURLIB_UNIT_POS ;
		case MAPE_CURLIB_IRDIS: 
			return CURLIB_UNIT_DISTANCE ;
		case MAPE_CURLIB_IFDIS: 
			return CURLIB_UNIT_DISTANCE ;
		default: 
			return 0.0;
	}
}


/*
 * mape_curlib_loadg_value
 *
 * get the parameter value for certain gate when fanout is 1. 
 */
double 
mape_curlib_loadg_value(gate, type, loadg)
mape_curlib_gate_t *gate ;
mape_curlib_param_type_t type ;
char *loadg ;
{
	int n = 0, tmp ;
	double rtv ;

	if(unit_mode)
		return mape_curlib_unit_value(type) ;

	if(gate == NULL)
		return 0.0 ;

	for(;;){
		if(gate->loadg->gate[n] == NULL)
			return 0.0 ;
		if(!strcmp(gate->loadg->gate[n], loadg)) 
			break ;
		n ++ ;			
	}

	switch(type){
		case MAPE_CURLIB_RDLAY: 
			if(various_mode){
				rtv = (gate->loadg->rdlay[n] + gate->loadg->fdlay[n])/2 ;

				if(CURLIB_APPROX_DELAY_UNIT == 0){
					return rtv ;
				}else{
					tmp = (int)(rtv+CURLIB_APPROX_DELAY_UNIT/2)
						/CURLIB_APPROX_DELAY_UNIT ;
					return (double)(tmp*CURLIB_APPROX_DELAY_UNIT) ;
				}
			}
			return gate->loadg->rdlay[n] ;
		case MAPE_CURLIB_FDLAY: 
			if(various_mode){
				rtv = (gate->loadg->rdlay[n] + gate->loadg->fdlay[n])/2 ;

				if(CURLIB_APPROX_DELAY_UNIT == 0){
					return rtv ;
				}else{
					tmp = (int)(rtv+CURLIB_APPROX_DELAY_UNIT/2)
						/CURLIB_APPROX_DELAY_UNIT ;
					return (double)(tmp*CURLIB_APPROX_DELAY_UNIT) ;
				}
			}
			return gate->loadg->fdlay[n] ;
		case MAPE_CURLIB_IRMAX: 
			return gate->loadg->irmax[n] ;
		case MAPE_CURLIB_IFMAX: 
			return gate->loadg->ifmax[n] ;
		case MAPE_CURLIB_IRPOS: 
			return gate->loadg->irpos[n] ;
		case MAPE_CURLIB_IFPOS: 
			return gate->loadg->ifpos[n] ;
		case MAPE_CURLIB_IRDIS: 
			return gate->loadg->irdis[n] ;
		case MAPE_CURLIB_IFDIS: 
			return gate->loadg->ifdis[n] ;
		default: 
			return 0.0;
	}
}

/*
 * mape_curlib_value
 *
 * get the parameter value for certain gate and it's loading capacitance. 
 * 
 */
double 
mape_curlib_value(gate, type, cload)
mape_curlib_gate_t *gate ;
mape_curlib_param_type_t type ;
double cload ;
{
	double rtv ;
	int tmp ;


	if(gate == NULL)
		gate = mape_curlib_find_gate(DEFAULT_LIB_GATE) ;
		
	if(unit_mode)
		return mape_curlib_unit_value(type) ;

	switch(type){
		case MAPE_CURLIB_RDLAY: 
			if(various_mode){
				rtv = (mape_curlib_param_value(gate->rdlay, gate->n, cload) +
				      mape_curlib_param_value(gate->fdlay, gate->n, cload))/2 ;

				if(CURLIB_APPROX_DELAY_UNIT == 0){
					return rtv ;
				}else{
					tmp = (int)(rtv+CURLIB_APPROX_DELAY_UNIT/2)
						/CURLIB_APPROX_DELAY_UNIT ;
					return (double)(tmp*CURLIB_APPROX_DELAY_UNIT) ;
				}
			}
			return mape_curlib_param_value(gate->rdlay, gate->n, cload);
		case MAPE_CURLIB_FDLAY: 
			if(various_mode){
				rtv = (mape_curlib_param_value(gate->rdlay, gate->n, cload) +
				      mape_curlib_param_value(gate->fdlay, gate->n, cload))/2 ;

				if(CURLIB_APPROX_DELAY_UNIT == 0){
					return rtv ;
				}else{
					tmp = (int)(rtv+CURLIB_APPROX_DELAY_UNIT/2)
						/CURLIB_APPROX_DELAY_UNIT ;
					return (double)(tmp*CURLIB_APPROX_DELAY_UNIT) ;
				}
			}
			return mape_curlib_param_value(gate->fdlay, gate->n, cload);
		case MAPE_CURLIB_IRMAX: 
			if(unit_mode)
				return CURLIB_UNIT_CURRENT ;
			return mape_curlib_param_value(gate->irmax, gate->n, cload);
		case MAPE_CURLIB_IFMAX: 
			if(unit_mode)
				//return CURLIB_UNIT_CURRENT ;
				return 0.0 ;
			return mape_curlib_param_value(gate->ifmax, gate->n, cload);
		case MAPE_CURLIB_IRPOS: 
			if(unit_mode)
				return CURLIB_UNIT_POS ;
			return mape_curlib_param_value(gate->irpos, gate->n, cload);
		case MAPE_CURLIB_IFPOS: 
			if(unit_mode)
				return CURLIB_UNIT_POS ;
			return mape_curlib_param_value(gate->ifpos, gate->n, cload);
		case MAPE_CURLIB_IRDIS: 
			if(unit_mode)
				return CURLIB_UNIT_DISTANCE ;
			return mape_curlib_param_value(gate->irdis, gate->n, cload);
		case MAPE_CURLIB_IFDIS: 
			if(unit_mode)
				return CURLIB_UNIT_DISTANCE ;
			return mape_curlib_param_value(gate->ifdis, gate->n, cload);
		case MAPE_CURLIB_EFFIN:
			if(unit_mode)
				return 0.0 ;
			return gate->effin ;
		default: 
			return 0.0;
	}
}

/*
 * mape_cur_lib_fitting_factor
 *
 * get the fitting factor for each parameter of each gate in the library. 
 */
void 
mape_curlib_fitting_factor(void)
{
	mape_curlib_gate_t *gate ;
	mape_curlib_param_t *param ;
	int n ;
	lsGen gen ;

	if(!cur_lib)
		return ;

	/*assume the space has been allocated for each gate*/
	lsForeachItem(cur_lib, gen, gate){
		n=gate->n ;

		param=gate->rdlay ;
		spline((param->x)-1, (param->y)-1, n, 1e+30, 1e+30, (param->y2a)-1) ;
		param=gate->fdlay ;
		spline((param->x)-1, (param->y)-1, n, 1e+30, 1e+30, (param->y2a)-1) ;
		param=gate->irmax ;
		spline((param->x)-1, (param->y)-1, n, 1e+30, 1e+30, (param->y2a)-1) ;
		param=gate->ifmax ;
		spline((param->x)-1, (param->y)-1, n, 1e+30, 1e+30, (param->y2a)-1) ;
		param=gate->irpos ;
		spline((param->x)-1, (param->y)-1, n, 1e+30, 1e+30, (param->y2a)-1) ;
		param=gate->ifpos ;
		spline((param->x)-1, (param->y)-1, n, 1e+30, 1e+30, (param->y2a)-1) ;
		param=gate->irdis ;
		spline((param->x)-1, (param->y)-1, n, 1e+30, 1e+30, (param->y2a)-1) ;
		param=gate->ifdis ;
		spline((param->x)-1, (param->y)-1, n, 1e+30, 1e+30, (param->y2a)-1) ;
	}
}

/*
 * mape_curlib_token_type
 *
 * return the token's type ;
 */
mape_curlib_param_type_t 
mape_curlib_token_type(token)
char *token ;
{
	if(strcmp(token, "circuit") == 0)
		return MAPE_CURLIB_CIRCUIT ;
	if(strcmp(token, "load") == 0)
		return MAPE_CURLIB_LOAD ;
	if(strcmp(token, "rdlay") == 0)
		return MAPE_CURLIB_RDLAY ;
	if(strcmp(token, "fdlay") == 0)
		return MAPE_CURLIB_FDLAY ;
	if(strcmp(token, "irmax") == 0)
		return MAPE_CURLIB_IRMAX ;
	if(strcmp(token, "ifmax") == 0)
		return MAPE_CURLIB_IFMAX ;
	if(strcmp(token, "irpos") == 0)
		return MAPE_CURLIB_IRPOS ;
	if(strcmp(token, "ifpos") == 0)
		return MAPE_CURLIB_IFPOS ;
	if(strcmp(token, "irdis") == 0)
		return MAPE_CURLIB_IRDIS ;
	if(strcmp(token, "ifdis") == 0)
		return MAPE_CURLIB_IFDIS ;
	if(strcmp(token, "effin") == 0)
		return MAPE_CURLIB_EFFIN ;
	if(strcmp(token, "loadg") == 0) 
		return MAPE_CURLIB_LOADG ;
	else {
		fprintf(stderr, "token=%s\n", token) ;
		mape_error("unrecognized token", __FILE__, __LINE__) ;
	}
}

/*
 * mape_curlib_parse_paramg_line. 
 *
 * parse the line describing the loadg parameter. 
 */
int
mape_curlib_parse_gparam_line(param, inf, n)
double *param ;
char *inf ;
int *n ;
{
	char *word, *next, *rest ; 
	double value ;

	next = inf ;
	while(word = (char *)gettoken(next, &rest)){
		next = rest ;
		value = atof(word) ;
		param[*n] = value ;
		(*n)++; 
	}
	return TRUE ;
}

/*
 * mape_curlib_parse_param_line. 
 *
 * parse the line describing the parameter. 
 */
int
mape_curlib_parse_param_line(param, inf, n)
mape_curlib_param_t *param ;
char *inf ;
int *n ;
{
	char *word, *next, *rest ; 
	double value ;

	next = inf ;
	while(word = (char *)gettoken(next, &rest)){
		next = rest ;
		value = atof(word) ;
		(param->y)[*n] = value ;
		(*n)++; 
	}
	return TRUE ;
}

/*
 * mape_curlib_parse_loadg_line. 
 *
 * parse the line with  loadg begin. 
 */
int
mape_curlib_parse_loadg_line(gate, inf, n)
mape_curlib_gate_t *gate ;
char *inf ;
int *n ;
{
	char *word, *next, *rest, *name ; 
	double load ;

	next = inf ;
	while(word = (char *)gettoken(next, &rest)){
		next = rest ;
		name = strdup(word) ;

		gate->loadg->gate[*n]	= name ;   
		(*n)++; 
	}
	return TRUE ;
}

/*
 * mape_curlib_parse_param_line. 
 *
 * parse the line with  load begin. 
 */
int
mape_curlib_parse_load_line(gate, inf, n)
mape_curlib_gate_t *gate ;
char *inf ;
int *n ;
{
	char *word, *next, *rest ; 
	double load ;

	next = inf ;
	while(word = (char *)gettoken(next, &rest)){
		gate->n ++ ;
		next = rest ;
		load = atof(word) ;

		(gate->rdlay->x)[*n] = load ;
		(gate->fdlay->x)[*n] = load ;
		(gate->irmax->x)[*n] = load ;
		(gate->ifmax->x)[*n] = load ;
		(gate->irpos->x)[*n] = load ;
		(gate->ifpos->x)[*n] = load ;
		(gate->irdis->x)[*n] = load ;
		(gate->ifdis->x)[*n] = load ;

		(*n)++; 
	}
	return TRUE ;
}

/*
 * mape_curlib_parse_file
 *
 * parse the library file and store the parameter values. 
 */
int 
mape_curlib_parse_file(fp)
FILE *fp ;
{
	int line_len ;
	char *line = NULL ;
	char *word, *rest, *tmp ;
	mape_curlib_gate_t *gate = NULL ;
	mape_curlib_param_type_t type ;
	int load_n, rdlay_n, fdlay_n, irmax_n, ifmax_n, irpos_n, ifpos_n, irdis_n, ifdis_n ;
	int loadg_n, rdlayg_n, fdlayg_n, irmaxg_n, ifmaxg_n, irposg_n, ifposg_n, irdisg_n, ifdisg_n ;
	int loadg_param = FALSE ;

	line_len = 0 ;
	while (read_line(fp, &line, &line_len) != EOF) {
		word = (char *)gettoken(line, &rest) ;
		type = mape_curlib_token_type(word) ;
		switch(type){
			case MAPE_CURLIB_CIRCUIT:
				if(gate != NULL)
					mape_curlib_push_curlib_list(gate) ;
				gate = mape_curlib_create_gate() ;
				load_n=0 ;
				loadg_n=0 ;
				rdlay_n=irmax_n=irpos_n=irdis_n=0 ;
				fdlay_n=ifmax_n=ifpos_n=ifdis_n=0 ;
				rdlayg_n=irmaxg_n=irposg_n=irdisg_n=0 ;
				fdlayg_n=ifmaxg_n=ifposg_n=ifdisg_n=0 ;
				gate->name = strdup(rest) ;
				break ;
			case MAPE_CURLIB_LOAD:
				if(!mape_curlib_parse_load_line(gate, rest, &load_n))
					return FALSE;
				loadg_param = FALSE ;
				break ;
			case MAPE_CURLIB_LOADG: 
				if(!mape_curlib_parse_loadg_line(gate, rest, &loadg_n))
					return FALSE ;
				loadg_param = TRUE ;
				break ;
			case MAPE_CURLIB_RDLAY:
				if(loadg_param){
					if(!mape_curlib_parse_gparam_line(gate->loadg->rdlay, rest, &rdlayg_n))
						return FALSE ;

				}else 
					if(!mape_curlib_parse_param_line(gate->rdlay, rest, &rdlay_n))
						return FALSE ;
				break ;
			case MAPE_CURLIB_FDLAY:
				if(loadg_param){
					if(!mape_curlib_parse_gparam_line(gate->loadg->fdlay, rest, &fdlayg_n))
						return FALSE ;

				}else 
					if(!mape_curlib_parse_param_line(gate->fdlay, rest, &fdlay_n))
						return FALSE ;
				break ;
			case MAPE_CURLIB_IRMAX:
				if(loadg_param){
					if(!mape_curlib_parse_gparam_line(gate->loadg->irmax, rest, &irmaxg_n))
						return FALSE ;

				}else 
					if(!mape_curlib_parse_param_line(gate->irmax, rest, &irmax_n))
						return FALSE ;
				break ;
			case MAPE_CURLIB_IFMAX:
				if(loadg_param){
					if(!mape_curlib_parse_gparam_line(gate->loadg->ifmax, rest, &ifmaxg_n))
						return FALSE ;

				}else 
					if(!mape_curlib_parse_param_line(gate->ifmax, rest, &ifmax_n))
						return FALSE ;
				break ;
			case MAPE_CURLIB_IRPOS:
				if(loadg_param){
					if(!mape_curlib_parse_gparam_line(gate->loadg->irpos, rest, &irposg_n))
						return FALSE ;

				}else 
					if(!mape_curlib_parse_param_line(gate->irpos, rest, &irpos_n))
						return FALSE ;
				break ;
			case MAPE_CURLIB_IFPOS:
				if(loadg_param){
					if(!mape_curlib_parse_gparam_line(gate->loadg->ifpos, rest, &ifposg_n))
						return FALSE ;

				}else 
					if(!mape_curlib_parse_param_line(gate->ifpos, rest, &ifpos_n))
						return FALSE ;
				break ;
			case MAPE_CURLIB_IRDIS:
				if(loadg_param){
					if(!mape_curlib_parse_gparam_line(gate->loadg->irdis, rest, &irdisg_n))
						return FALSE ;

				}else 
					if(!mape_curlib_parse_param_line(gate->irdis, rest, &irdis_n))
						return FALSE ;
				break ;
			case MAPE_CURLIB_IFDIS:
				if(loadg_param){
					if(!mape_curlib_parse_gparam_line(gate->loadg->ifdis, rest, &ifdisg_n))
						return FALSE ;

				}else 
					if(!mape_curlib_parse_param_line(gate->ifdis, rest, &ifdis_n))
						return FALSE ;
				break ;
			case MAPE_CURLIB_EFFIN:
				word = (char *)gettoken(rest, &tmp) ;
				gate->effin = atof(word) ;
				break ;
			default: 
				return FALSE ;
		}
	}
	if(gate != NULL)
		mape_curlib_push_curlib_list(gate) ;
	return TRUE ;
}

/*
 * mape_curlib_dump_param
 *
 * dump the gate's parameter. 
 */
void 
mape_curlib_dump_param(param, n)
mape_curlib_param_t *param ;
int n ;
{
	int i ;
	
	for(i=0;i<n;i++){
		if(i%10 == 0 && i!=0)
			fprintf(mapeout, "\n        \t") ;
		fprintf(mapeout, "%.2lf ", (param->x)[i]) ;
	}
	fprintf(mapeout, "\n\n        \t") ;
	for(i=0;i<n;i++){
		if(i%10 == 0 && i!=0)
			fprintf(mapeout, "\n        \t") ;
		fprintf(mapeout, "%.2lf ", (param->y)[i]) ;
	}
	fprintf(mapeout, "\n\n        \t") ;
	for(i=0;i<n;i++){
		if(i%10 == 0 && i!=0)
			fprintf(mapeout, "\n        \t") ;
		fprintf(mapeout, "%.2lf ", (param->y2a)[i]) ;
	}
	fprintf(mapeout, "\n\n") ;
}

/*
 * mape_curlib_dump_loadg
 *
 *
 * dump the loadg information. 
 */
void 
mape_curlib_dump_loadg(loadg)
mape_curlib_loadg_t *loadg ;
{
	int n = 0 ;
	int i = 0 ;

	fprintf(mapeout, "\nLoading Gates:") ;
	for(;;){
		if(loadg->gate[n] == NULL)
			break ;
		fprintf(mapeout, "\t%7s", loadg->gate[n]) ;
		n ++ ;
	}
	fprintf(mapeout, "\nRDLAY:") ;
	for(i=0;i<n;i++){
		if(i == 0)
			fprintf(mapeout, "\t") ;
		fprintf(mapeout, "\t%7.2f", loadg->rdlay[i]) ;
	}
	fprintf(mapeout, "\nFDLAY:") ;
	for(i=0;i<n;i++){
		if(i == 0)
			fprintf(mapeout, "\t") ;
		fprintf(mapeout, "\t%7.2f", loadg->fdlay[i]) ;
	}
	fprintf(mapeout, "\nIRMAX:") ;
	for(i=0;i<n;i++){
		if(i == 0)
			fprintf(mapeout, "\t") ;
		fprintf(mapeout, "\t%7.2f", loadg->irmax[i]) ;
	}
	fprintf(mapeout, "\nIFMAX:") ;
	for(i=0;i<n;i++){
		if(i == 0)
			fprintf(mapeout, "\t") ;
		fprintf(mapeout, "\t%7.2f", loadg->ifmax[i]) ;
	}
	fprintf(mapeout, "\nIRPOS:") ;
	for(i=0;i<n;i++){
		if(i == 0)
			fprintf(mapeout, "\t") ;
		fprintf(mapeout, "\t%7.2f", loadg->irpos[i]) ;
	}
	fprintf(mapeout, "\nIFPOS:") ;
	for(i=0;i<n;i++){
		if(i == 0)
			fprintf(mapeout, "\t") ;
		fprintf(mapeout, "\t%7.2f", loadg->ifpos[i]) ;
	}
	fprintf(mapeout, "\nIRDIS:") ;
	for(i=0;i<n;i++){
		if(i == 0)
			fprintf(mapeout, "\t") ;
		fprintf(mapeout, "\t%7.2f", loadg->irdis[i]) ;
	}
	fprintf(mapeout, "\nIFDIS:") ;
	for(i=0;i<n;i++){
		if(i == 0)
			fprintf(mapeout, "\t") ;
		fprintf(mapeout, "\t%7.2f", loadg->ifdis[i]) ;
	}
}

/*
 * mape_curlib_dump_gate
 *
 * dump the information of the gate. 
 */
void 
mape_curlib_dump_gate(gate)
mape_curlib_gate_t *gate ;
{
	fprintf(mapeout, "\n\n\nCircuit Name:\t%s\n", gate->name) ;
	fprintf(mapeout, "Instance :\t%d\n", gate->n) ;
	fprintf(mapeout, "Effin : \t%.2lf\n", gate->effin) ;

	fprintf(mapeout, "RDLAY : \t") ;
	mape_curlib_dump_param(gate->rdlay, gate->n) ;
	fprintf(mapeout, "FDLAY : \t") ;
	mape_curlib_dump_param(gate->fdlay, gate->n) ;
	fprintf(mapeout, "IRMAX : \t") ;
	mape_curlib_dump_param(gate->irmax, gate->n) ;
	fprintf(mapeout, "IFMAX : \t") ;
	mape_curlib_dump_param(gate->ifmax, gate->n) ;
	fprintf(mapeout, "IRPOS : \t") ;
	mape_curlib_dump_param(gate->irpos, gate->n) ;
	fprintf(mapeout, "IFPOS : \t") ;
	mape_curlib_dump_param(gate->ifpos, gate->n) ;
	fprintf(mapeout, "IRDIS : \t") ;
	mape_curlib_dump_param(gate->irdis, gate->n) ;
	fprintf(mapeout, "IFDIS : \t") ;
	mape_curlib_dump_param(gate->ifdis, gate->n) ;
	mape_curlib_dump_loadg(gate->loadg) ;
}

/*
 * mape_curlib_dump 
 *
 * dump the information that we get
 */
void 
mape_curlib_dump(void)
{
	lsGen gen ;
	mape_curlib_gate_t *gate ;

	lsForeachItem(cur_lib, gen, gate)
		mape_curlib_dump_gate(gate) ;
}

/*
 * mape_cur_build_lib
 *
 * build the current library. 
 */
void 
mape_curlib_build_lib(void)
{
	char *real_path ;
	FILE *flib ;

	if((flib = com_open_file(CURLIB_FILENAME, "r", &real_path, 0)) == NULL)
		mape_error("Can't open the library file\n", __FILE__, __LINE__); 

	if(!mape_curlib_parse_file(flib))
		mape_error("Parsing current library file fail\n", __FILE__, __LINE__); 

	/* get the fitting factor for each parameter of each gate in the library*/
	mape_curlib_fitting_factor() ;
	/* dump the lib information*/
	//mape_curlib_dump() ;
}
