/*
 * mape_group.c 
 *
 *
 * divide the circuit into groups and compare the 
 * PI vectors for each group and for the whole circuit. 
 *
 * Changbo Long (Nov. 26, 2001)  clong@cae.wisc.edu
 */
#include "mape_int.h"

#define ONE_TIME_MAX		5	
#define MIN_GROUP_DIFFERENCE 	2
#define PERTURBATION_DEGREE 	0.5

typedef struct group_static_struct {
	lsList global ; /*statistics for global*/
	lsList group ;  /*statistics for group*/
} group_static_t ;

lsList groups = NULL; 
lsList reports = NULL ;
static void mape_group_report(network_t *, lsList, int, long) ;
static void mape_group_static_reset(void) ;

static char *table_begin_s 	= "\\begin{center}\n\\begin{tabular}" ;
static char *table_end_s	= "\\end{tabular}\n\\end{center}" ;
static char *table_line_end_s 	= "\\\\ \\hline" ;
static int group_num = 10  ;
static int group_node_num ;
static group_static_t *group_static = NULL ; 

/*
 * mape_group_static_push_group
 *
 * push a value into global. 
 */
void 
mape_group_static_push_group(v)
long v ;
{
	long *p = ALLOC(long, 1) ;

	if(!group_static)
		mape_group_static_reset() ;

	*p = v ;
	lsNewEnd(group_static->group, (lsGeneric)p, LS_NH) ;
}

/*
 * mape_group_static_push_global
 *
 * push a value into global. 
 */
void 
mape_group_static_push_global(v)
long v ;
{
	long *p = ALLOC(long, 1) ;
	if(!group_static)
		mape_group_static_reset() ;

	*p = v ;
	lsNewEnd(group_static->global, (lsGeneric)p, LS_NH) ;
}

/*
 * mape_group_create_static
 *
 * create a static structure. 
 */
group_static_t *
mape_group_create_static(void)
{
	group_static_t *rtv ;

	rtv 	= ALLOC(group_static_t, 1) ;

	rtv->global	= lsCreate() ;
	rtv->group	= lsCreate() ;

	return rtv ;
}

/*
 * mape_group_static_reset 
 *
 * reset the statics. 
 */
void
mape_group_static_reset(void)
{
	long *tmp ;

	if(!group_static){
		group_static = mape_group_create_static() ;
		return ;
	}

	while(lsLength(group_static->global)){
		LS_ASSERT(lsDelEnd(group_static->global, (lsGeneric*)&tmp))  ;
		free(tmp) ;
	}
	while(lsLength(group_static->group)){
		LS_ASSERT(lsDelEnd(group_static->group, (lsGeneric*)&tmp)) ;
		free(tmp) ;
	}

}

/*
 * mape_group_set_group_num ;
 *
 * set the group number 
 */
void 
mape_group_set_group_num(n)
int n ;
{
	group_node_num = n ;
	group_num = MAX(network_num_internal(global_network)/n, 1) ;
	/*
	group_num = n ;
	*/
}

/*
 * mape_group_get_group_num ;
 *
 * set the group number 
 */
int
mape_group_get_group_num(viod)
{
	return group_num ;
}

/*
 * mape_group_num_cmp
 *
 * for lsSort, compare the group's num
 */
int 
mape_group_num_cmp(obj1, obj2)
lsGeneric obj1, obj2 ;
{
	int length1 = lsLength((lsList)obj1) ;
	int length2 = lsLength((lsList)obj2) ;

	return (length1 - length2) ;
}

/*
 * mape_node_degree
 *
 * return the degree of the node. 
 */
int 
mape_node_degree(node)
node_t *node ;
{
	return node_num_fanout(node) + node_num_fanin(node) ;
}

/*
 * mape_group_push_reports
 * 
 * push the report string into reports. 
 */
void 
mape_group_push_reports(s)
char *s ;
{
	if(!s)
		return ;
	if(!reports)
		reports = lsCreate() ;
	lsNewEnd(reports, (lsGeneric)s, LS_NH) ;
}

/*
 * mape_group_node_set_grouped
 *
 * set the node to be grouped or not. 
 */
void 
mape_group_node_set_grouped(node, b)
node_t *node ;
int b ;
{
	if(node->mape == NULL) 
		node->mape = (char *)mape_create_mape_t() ;

	((mape_t *)node->mape)->C0 = b ;
}

/*
 * mape_group_dump_groups
 *
 * dump the groups. 
 */
void
mape_group_dump_groups(void)
{
	int i = 0, j = 0 ;
	lsGen gen, gen1 ;
	lsList group ;
	node_t *node ;

	/*
	lsForeachItem(groups, gen, group){
		i ++ ;
		fprintf(mapeout, "Group #%d(%d) \n", i, lsLength(group)) ;
		mape_upb_dump_tie(group) ;
		fprintf(mapeout, "\n") ;
	}
	*/

	lsList last ;
	lsList first ;

	lsSort(groups, mape_group_num_cmp) ;
	lsLastItem(groups, (lsGeneric *)&last, LS_NH) ;
	lsFirstItem(groups, (lsGeneric *)&first, LS_NH) ;

	fprintf(mapeout, "\tGroup Size: %d~%d\n", lsLength(first), lsLength(last)) ; 
}

/*
 * mape_group_node_is_grouped
 *
 * Is the node be grouped?
 */
int
mape_group_node_is_grouped(node)
node_t *node ;
{
	if(node->mape == NULL) 
		return FALSE ;

	if(node_type(node) == PRIMARY_INPUT || node_type(node) == PRIMARY_OUTPUT)
		return TRUE ;

	return 	((mape_t *)node->mape)->C0 ;
}

/*
 * mape_min_node. 
 *
 * find the node with the minimal function
 * value. 
 */
void 
mape_min_node(list, f, m_value, m_h)
lsList list ;
int (*f)() ;
int *m_value ;
lsHandle *m_h ;
{
	lsGen gen ;
	node_t *p ;
	int value ;
	lsHandle h ;

	*m_value = 9999 ;
	*m_h 	 = NULL ;

	gen = lsStart(list) ;
	while(lsNext(gen, (lsGeneric *)&p, &h) == LS_OK){
		value = (*f)(p) ;
		if(value < *m_value){
			*m_value = value ;
			*m_h = h ;
		}
	}
	lsFinish(gen) ;
}

/*
 * mape_node_name_value
 *
 * return the name value for a node. 
 */
int 
mape_node_name_value(n)
node_t *n ;
{
	int i, rtv = 0 ;
	char *s ;

	s = n->short_name ;
	for(i=0;i<strlen(s);i++)
		rtv += s[i]*(i+1)*30 ;

	return rtv ;
}

/*
 * mape_node_name_cmp
 *
 * compare two nodes according to it's short name. 
 */
int 
mape_node_name_cmp(obj1, obj2)
lsGeneric obj1, obj2 ;
{
	node_t *n1, *n2 ;

	n1 = (node_t *)obj1 ;
	n2 = (node_t *)obj2 ;

	return mape_node_name_value(n1) - mape_node_name_value(n2) ;
}

/*
 * mape_group_create_seeds. 
 *
 * create seeds for expansion. 
 */
lsList 
mape_group_create_seeds(network)
network_t *network ;
{
	lsList rtv, tmpList ;
	lsGen  gen ;
	node_t *node, *tmp ;
	int seed_num ;
	int th = 9999; 	/*threthold for accept as a seed*/
	lsHandle h ;
	int f_num ;

	rtv = lsCreate();
	tmpList = lsCreate() ;
	/*seed_num = (lsLength(network->nodes))/GROUP_SIZE + 1 ;*/
	seed_num = mape_group_get_group_num() ;

	foreach_node(network, gen, node){
		if(node_type(node) != PRIMARY_INPUT && node_type(node) != PRIMARY_OUTPUT)
			lsNewEnd(tmpList, (lsGeneric)node, &h) ;
	}

	lsSort(tmpList, mape_node_name_cmp) ;

	lsForeachItem(tmpList, gen, node){
		f_num = mape_node_degree(node) ;
		if(lsLength(rtv) < seed_num){  /*no enough seed*/
			if(f_num < th){
				th = f_num ;
				lsNewEnd(rtv, (lsGeneric)node, &h) ;
			}else 
				lsNewEnd(rtv, (lsGeneric)node, LS_NH) ;
		}else{
			if(f_num > th){
				LS_ASSERT(lsRemoveItem(h, (lsGeneric *)&tmp)) ;
				lsNewEnd(rtv, (lsGeneric)node, LS_NH) ; 
				mape_min_node(rtv, mape_node_degree, &th, &h) ;
			}
		}
	}

	/*
	fprintf(mapeout, "Seeds(%d): \n", lsLength(rtv)) ;
	lsForeachItem(rtv, gen, node)
		fprintf(mapeout, "\t%s:\t%2d+%-2d=%-2d\n", node->name, 
			node_num_fanout(node), node_num_fanin(node), mape_node_degree(node)) ;
	*/
	return rtv ;
}

/*
 * mape_create_groups
 *
 * create the groups from seeds. 
 * just push the initial seeds into 
 * each group. 
 */
void
mape_create_groups(seeds)
lsList seeds ;
{
	lsList p ;
	lsGen gen ;
	node_t *node ;

	groups = lsCreate() ;
	lsForeachItem(seeds, gen, node){
		p = lsCreate() ;	
		lsNewEnd(p, (lsGeneric)node, LS_NH) ;
		mape_group_node_set_grouped(node, TRUE) ;
		lsNewEnd(groups, (lsGeneric)p, LS_NH) ;
	}
}

/*
 * mape_group_node_eat_neighbores
 *
 *
 * node eat all the neighbores. 
 */
int 
mape_group_node_eat_neighbores(node, list, num)
node_t *node ;
lsList list ;
int *num ;
{
	node_t *p ;
	lsGen gen ;
	int rtv = FALSE ;
	int i ;


	/* eat all the ungrouped fanout */
	foreach_fanout(node, gen, p){
		if(node_type(p) == PRIMARY_INPUT || node_type(p) == PRIMARY_OUTPUT)
			continue ;
		if(*num >= ONE_TIME_MAX){
			lsFinish(gen) ;
			return rtv ;
		}	
		if(!mape_group_node_is_grouped(p)){
			lsNewEnd(list, (lsGeneric)p, LS_NH) ;
			mape_group_node_set_grouped(p, TRUE) ;
			if(!rtv)
				rtv = TRUE ;
			(*num) ++ ;
		}
	}

	/* eat all the ungrouped fanin */
	foreach_fanin(node, i, p){
		if(node_type(p) == PRIMARY_INPUT || node_type(p) == PRIMARY_OUTPUT)
			continue ;
		if(*num >= ONE_TIME_MAX)
			return rtv ;
		if(!mape_group_node_is_grouped(p)){
			lsNewEnd(list, (lsGeneric)p, LS_NH) ;
			mape_group_node_set_grouped(p, TRUE) ;
			if(!rtv)
				rtv = TRUE ;
			(*num) ++ ;
		}
	}
	return rtv;
}

/*
 * mape_group_eat_neighbores
 *
 *
 * for each node in the group list eat all it's 
 * neighbors. 
 */
int
mape_group_eat_neighbores(void)
{
	lsGen gen1, gen2 ;
	lsList group ;
	node_t *node ;
	lsList tmp ;
	node_t *p ;
	int rtv = FALSE ; /* didn't get any*/
	int num ;

	tmp = lsCreate() ;
	lsForeachItem(groups, gen1, group){
		num = 0 ;
		lsForeachItem(group, gen2, node)
			if(mape_group_node_eat_neighbores(node, tmp, &num) && !rtv)
				rtv = TRUE;
		while(lsLength(tmp)){
			LS_ASSERT(lsDelEnd(tmp, (lsGeneric *)&p)) ;
			lsNewEnd(group, (lsGeneric)p, LS_NH) ;
		}
	}
	lsDestroy(tmp, NULL) ;
	return rtv ;
}

/*
 * mape_group_expansion
 *
 *
 * expand the seeds to eat up all the nodes in the
 * circuit. 
 */
void 
mape_group_expansion(seeds)
lsList seeds ;
{
	int no_left  ;

	mape_create_groups(seeds) ;

	do{  
		no_left = TRUE ;
		lsSort(groups, mape_group_num_cmp) ;
		if(mape_group_eat_neighbores())
			no_left = FALSE ;
	}while(!no_left) ;
}

/*
 * mape_group_max_difference. 
 *
 * get the maximum difference between groups. 
 */
int 
mape_group_max_difference(void)
{
	lsList first ;
	lsList last ;

	lsFirstItem(groups, (lsGeneric*)&first, LS_NH) ;
	lsLastItem(groups, (lsGeneric*)&last, LS_NH) ;

	return (lsLength(last) - lsLength(first)) ;
}

/*
 * mape_group_which   
 *
 * which group is a node in? 
 */
lsList
mape_group_which(node)
node_t *node ;
{
	lsList p ;
	lsGen gen ;

	lsForeachItem(groups, gen, p){
		if(mape_node_in_list(node, p)){
			lsFinish(gen) ;
			return p ;
		}
	}
	mape_error("can't find node's group!\n", __FILE__, __LINE__) ;
}

/*
 * mape_group_transfer
 *
 *
 * transfer from one group to another 
 */
void 
mape_group_transfer(node, from, to)
node_t *node ;
lsList from, to ;
{
	lsGen gen ;
	node_t *p, *tmp ;

	lsNewEnd(to, (lsGeneric)node, LS_NH) ;
	lsForeachItem(from, gen, p){
		if(p == node){
			LS_ASSERT(lsDelBefore(gen, (lsGeneric *)&tmp)) ;
			lsFinish(gen) ;
			return ;
		}
	}
	mape_error("mape_transfer, can not find node!", __FILE__, __LINE__) ;
}

/*
 * mape_group_neutrialize_small
 *
 * neutrialize small group. 
 */
int 
mape_group_neutrialize_small(first)
lsList first ;
{
	lsList g ;
	node_t *node, *fo, *fi, *tmp ;
	int rtv = FALSE ;
	lsGen gen, gen1, gen2;
	lsHandle h ;
	int i ;

	lsForeachItem(first, gen, node){
		foreach_fanout(node, gen1, fo){
			if(node_type(fo) == PRIMARY_INPUT || node_type(fo) == PRIMARY_OUTPUT)
				continue ;
			g = mape_group_which(fo) ;
			if(g == first)
				continue ;
			if(lsLength(g) - lsLength(first) >= 2){
				mape_group_transfer(fo, g, first) ;
				rtv = TRUE ;
			}
		}
		foreach_fanin(node, i, fi){
			if(node_type(fi) == PRIMARY_INPUT || node_type(fi) == PRIMARY_OUTPUT)
				continue ;
			g = mape_group_which(fi) ;
			if(g == first)
				continue ;
			if(lsLength(g) - lsLength(first) >= 2){
				mape_group_transfer(fi, g, first) ;
				rtv = TRUE ;
			}
		}
	}
	return rtv ;
}

/*
 * mape_group_neutrialize_large
 *
 * neutrialize large group. 
 */
int 
mape_group_neutrialize_large(last)
lsList last ;
{
	lsList g ;
	node_t *node, *fo, *fi, *tmp ;
	int rtv = FALSE ;
	lsGen gen, gen1, gen2;
	lsHandle h ;
	int transfered ;
	int i ;

	gen = lsStart(last) ;
	while(lsNext(gen, (lsGeneric *)&node, &h) == LS_OK){
		transfered = FALSE ;
		foreach_fanout(node, gen1, fo){
			if(node_type(fo) == PRIMARY_INPUT || node_type(fo) == PRIMARY_OUTPUT)
				continue ;
			g = mape_group_which(fo) ;
			if(g == last)
				continue ;
			if(lsLength(last) - lsLength(g) >= 2){
				lsDelBefore(gen, (lsGeneric *)&tmp)  ;
				lsNewEnd(g, (lsGeneric)node, LS_NH) ;
				transfered = TRUE ;
				lsFinish(gen1) ;
				rtv = TRUE ;
				break ;
			}
		}
		if(transfered)
			continue ;
		foreach_fanin(node, i, fi){
			if(node_type(fi) == PRIMARY_INPUT || node_type(fi) == PRIMARY_OUTPUT)
				continue ;
			g = mape_group_which(fi) ;
			if(g == last)
				continue ;
			if(lsLength(last) - lsLength(g) >= 2){
				lsDelBefore(gen, (lsGeneric *)&tmp)  ;
				lsNewEnd(g, (lsGeneric)node, LS_NH) ;
				transfered = TRUE ;
				rtv = TRUE ;
				break ;
			}
		}
	}
	lsFinish(gen) ;
	return rtv ;
}

/*
 * mape_group_neutrialize
 *
 * neutrialize the groups. 
 */
int
mape_group_neutrialize(void)
{
	lsList last ;
	lsList first ;
	lsList g ;
	node_t *node, *fo, *fi, *tmp ;
	int rtv = FALSE ;
	lsGen gen, gen1, gen2;
	lsHandle h ;
	int i ;
	int transfered ;
	int total ;


	lsLastItem(groups, (lsGeneric *)&last, LS_NH) ;
	lsFirstItem(groups, (lsGeneric *)&first, LS_NH) ;

	if(lsLength(last) - lsLength(first) <= 2)
		return rtv ;

	total = lsLength(groups)/4 ;
	//total = total==0?1:total ;
	total = 1 ;

	i = 0 ;
	lsForeachItem(groups, gen, first){
		if(i >= total) {
			lsFinish(gen) ;
			break ;
		}
		if(mape_group_neutrialize_small(first))
			rtv = TRUE ;
		i++ ;
	}

	i = 0 ;
	gen = lsEnd(groups) ;
	while(lsPrev(gen, (lsGeneric *)&last, LS_NH) == LS_OK){
		if(i >= total){
			lsFinish(gen) ;
			break ;
		}
		if(mape_group_neutrialize_large(last))
			rtv = TRUE ;
		i++ ;
	}
	return rtv ;
}

/*
 * mape_group_perturbation
 *
 * perturbation of the group distribution. 
 */
void
mape_group_perturbation(degree)
float degree ;
{
	long r;
	lsGen gen ;
	lsList group, target ;
	node_t *node1, *node2 ;
	array_t *gv ;
	int length, g_length ;
	int i, j = 0 ;

	degree = MIN(1.0, degree) ;
	gv = array_alloc(lsList, 0) ;
	lsForeachItem(groups, gen, group)
		array_insert_last(lsList, gv, group) ;

	length = lsLength(groups) ;
	/*
	srandom(util_cpu_time()) ;
	*/
	lsForeachItem(groups, gen, group){
		g_length = lsLength(group) ;
		for(i=0;i<g_length*degree;i++){
			/*
			r = random() ;
			*/
			target = array_fetch(lsList, gv, i%length) ; 

			LS_ASSERT(lsDelEnd(group, (lsGeneric*)&node1)) ;
			LS_ASSERT(lsDelEnd(target, (lsGeneric*)&node2)) ;

			lsNewBegin(target, (lsGeneric)node1, LS_NH) ;
			lsNewBegin(group, (lsGeneric)node2, LS_NH) ;
		}
	}
	array_free(gv) ;
}

/*
 * mape_group_absolute_neutrilize
 *
 * absoluately neutrilize the group node's number. 
 */
void 
mape_group_absolute_neutrialize(void)
{
	lsList tmp ;
	lsList group ;
	lsGen gen ;
	node_t *node ;

	tmp = lsCreate() ;

	lsForeachItem(groups, gen, group){
		while(lsLength(group) > group_node_num){
			lsDelEnd(group, (lsGeneric *)&node) ;
			lsNewEnd(tmp, (lsGeneric)node, LS_NH) ;
		}
	}

	lsForeachItem(groups, gen, group){
		if(lsLength(tmp) == 0){
			lsFinish(gen) ;
			lsDestroy(tmp, NULL);
			return ;
		}
		while(lsLength(group) < group_node_num){
			lsDelEnd(tmp, (lsGeneric *)&node) ;
			lsNewEnd(group, (lsGeneric)node, LS_NH) ;
		}
	}

	gen = lsStart(groups) ;
	while(lsLength(tmp)){
		if(lsNext(gen, (lsGeneric *)&group, LS_NH) != LS_OK){
			lsFinish(gen) ;
			gen = lsStart(groups) ;
			continue ;
		}
		lsDelEnd(tmp, (lsGeneric *)&node) ;
		lsNewEnd(group, (lsGeneric)node, LS_NH) ;
	}

	lsDestroy(tmp, NULL) ;
}


/*
 * mape_group_redistribute
 *
 *
 * redistribute to make the group equalization. 
 */
void 
mape_group_redistribute(void)
{
	int i = 0;

	/*neutialization*/
	do{
		lsSort(groups, mape_group_num_cmp) ;
	}while(mape_group_neutrialize()) ;
	
	mape_group_absolute_neutrialize() ;

	/*pertubation of the group*/
	mape_group_perturbation(PERTURBATION_DEGREE) ;
}		

/*
 * mape_group_divide 
 *
 * devide the circuit into groups. 
 */
void 
mape_group_divide(network)
network_t *network ;
{
	lsList seeds ;

	/* select seeds from the network */
	seeds = mape_group_create_seeds(network) ;
	fprintf(mapeout, "\n\nGroup Numbers:%d\n", lsLength(seeds)) ;

	/*expand seeds to eat up all their neighborehood*/
	mape_group_expansion(seeds) ;

	/*redistribution of the node to make equlizaiton*/
	mape_group_redistribute() ;

	/*dump groups */
	mape_group_dump_groups() ;

	lsDestroy(seeds, NULL) ;
}

/*
 * mape_group_estimate_group
 *
 * estimate the group. 
 */
void 
mape_group_estimate_group(network, group, index)
network_t *network ;
lsList group ;
int index ;
{
	lsGen gen ;
	node_t *node ;
	long last_time ;

	mape_clean_given_gate() ;
	mape_itr_clean(network);
	lsForeachItem(group, gen, node)
		mape_push_given_gatep(node) ;

	last_time = util_cpu_time() ;
	mape_itr_init(network) ;
	mape_itr_refine(network) ;
	mape_group_report(network, group, index, util_cpu_time() - last_time) ;
}

/*
 * mape_group_estimate_network
 *
 * estimate the network. 
 */
void 
mape_group_estimate_network(network)
network_t *network ;
{
	int i ;
	long last_time ;

	mape_clean_given_gate() ;
	mape_itr_clean(network) ;

	last_time = util_cpu_time() ;
	mape_itr_init(network) ;
	mape_itr_refine(network) ;
	mape_group_report(network, network->nodes, 0, util_cpu_time() - last_time) ;
	//mape_report(network, user_option->verb, util_cpu_time() - last_time) ;
}

/*
 * mape_group_estimate
 *
 *
 * estiamte the maximum power and get the vectors for the 
 * whole circuit and for each group. 
 */
void 
mape_group_estimate(network)
network_t *network ;
{
	lsList group ;
	lsGen gen ;
	int index = 0 ;

	/*
	fprintf(mapeout, "%s", table_begin_s) ;
	fprintf(mapeout, "{|c|c|c|c|c|}\n\\hline\n") ;
	fprintf(mapeout, "Group &\\#Node & NMF & Vector & Time %s\n", table_line_end_s) ;
	*/

	mape_group_estimate_network(network) ;

	lsForeachItem(groups, gen, group){
		index ++ ;
		mape_group_estimate_group(network, group, index) ; 
	}
	
	/*
	fprintf(mapeout, "%s", table_end_s) ;
	*/
}

/*
 * mape_group_pi_string, 
 *
 * put pi's value into s. 
 */
void
mape_group_pi_string(network, s) 
network_t *network ;
char *s ;
{
	lsGen gen ;
	node_t *pi ;
	int i = 0 ;

	*s = '\0' ;
	foreach_primary_input(network, gen, pi)
		sprintf(s, "%s%s", s, mape_print_node_value_c(mape_top(pi))) ;
}

/*
 * mape_group_NMF
 *
 * get the NMF of a list. 
 */
long 
mape_group_NMF(list)
lsList list ;
{
	long rtv = 0 ;
	lsGen gen ;
	node_t *node ;

	lsForeachItem(list, gen, node)
		rtv += LSLENGTH(node->fanout) ;
	return rtv ;
}

/*
 * mape_group_TMF
 *
 * get the TMF of a list. 
 */
long 
mape_group_TMF(list)
lsList list ;
{
	long rtv = 0 ;
	lsGen gen ;
	node_t *node ;

	lsForeachItem(list, gen, node)
		rtv += (TRANSITION_VALUE(mape_top(node))?1:0) * LSLENGTH(node->fanout) ;
	return rtv ;
}

/*
 * mape_group_trans_percent 
 *
 * get the transpercent of the list's nodes. 
 */
float 
mape_group_trans_percent(list)
lsList list ; 
{
	lsGen gen ;
	node_t *node ;
	long tTF = 0, aTF = 0 ;

	lsForeachItem(list, gen, node){
		tTF += LSLENGTH(node->fanout) ;
		aTF += (TRANSITION_VALUE(mape_top(node))?1:0) * LSLENGTH(node->fanout) ;
	}
	return (float)((100.0 * aTF)/(float)(tTF * 1.0)) ;
}

/*
 * mape_group_statistics_string
 *
 *
 * return the statistics information 
 * for the circuit.  this information
 * is for constructing the latex table
 * file. 
 */
char *
mape_group_statistics_string(network, index)
network_t *network ;
int index ;
{
	lsGen gen ;
	char tmp[1024] ;
	lsList p ;
	long tmf ;

	if(index == 0)
		sprintf(tmp, "Whole Circuit") ;
	else
		sprintf(tmp, "Group %d", index) ;
	tmf = mape_group_TMF(network->nodes) ;

	/* TMF for the whole circuit */
	sprintf(tmp, "%s & %ld", tmp, tmf) ; 	

	lsForeachItem(groups, gen, p){
		tmf = mape_group_TMF(p) ;
		sprintf(tmp, "%s & %ld", tmp, tmf) ; 	
	}
	sprintf(tmp, "%s %s", tmp, table_line_end_s) ;
	return strdup(tmp) ;
}

/*
 * mape_group_report
 *
 * report the result of estimation.  
 */
void 
mape_group_report(network, group, index, time) 
network_t *network;
lsList group ; /*node's list */
int index ;
long time ;
{
	/*
	char c_name[128]  ;
	char pi[128] ;
	char *statistics ;

	if(index == 0)
		strcpy(c_name, "Whole Circuit") ;
	else 
		sprintf(c_name, "Group %d", index) ;

	mape_group_pi_string(network, pi) ;

	fprintf(mapeout, "%s & %d & %d & %s & %s %s\n", c_name, lsLength(group), 
		mape_group_NMF(group), pi, util_print_time(time), table_line_end_s) ;

	statistics = mape_group_statistics_string(network, index) ;
	mape_group_push_reports(statistics) ;
	*/
	lsGen gen ;
	lsList p ;

	if(index == 0){
		lsForeachItem(groups, gen, p)
			mape_group_static_push_global(mape_group_TMF(p)) ;
	}else {
		mape_group_static_push_group(mape_group_TMF(group)) ;
	}
}

/*
 * mape_group_report_static
 *
 * report the statistics for every group. 
 */
void 
mape_group_report_static(void)
{
	/*
	char text[1024] ;
	int i ;
	lsGen gen ;
	char *s ;

	fprintf(mapeout, "%s", table_begin_s) ;
	fprintf(mapeout, "{") ;
	for(i=0;i<lsLength(groups) + 2;i++)
		fprintf(mapeout, "|c") ;
	fprintf(mapeout, "|}\n\\hline\n") ;
	fprintf(mapeout, "Vector Goal & Whole circuit ") ; 
	for(i=0;i<lsLength(groups);i++)
		fprintf(mapeout, "&G%d ", i+1) ;  
	fprintf(mapeout, "%s\n", table_line_end_s) ;

	lsForeachItem(reports, gen, s)
		fprintf(mapeout, "%s\n", s) ;
	fprintf(mapeout, "%s", table_end_s) ;
	*/
	long *p_global, *p_group ;
	long global_t = 0 ;
	long group_t = 0  ;
	long global_m = 0 ;
	long group_m = 0 ;
	lsGen gen1, gen2 ;
	lsHandle h ;
	int i = 0 ;


	gen1 = lsStart(group_static->global) ;
	gen2 = lsStart(group_static->group) ;

	for(;;){
		i ++ ;
		if(lsNext(gen1, (lsGeneric *)&p_global, &h) != LS_OK)
			break;		
		if(lsNext(gen2, (lsGeneric *)&p_group, &h) != LS_OK)
			mape_error("static_report!", __FILE__, __LINE__) ;
		global_t += *p_global ;
		group_t += *p_group ;

		if(group_m == 0 || global_m == 0){
			group_m = *p_group ;
			global_m = *p_global ;
		}else if((float)(100.0 * (group_m - global_m))/((float)(global_m)) < 
		   (float)(100.0 * (*p_group - *p_global))/((float)(*p_global)) ){
			group_m = *p_group ;
			global_m = *p_global ;
		}
	}
	lsFinish(gen1) ;
	lsFinish(gen2) ;
	fprintf(mapeout, "\tAverage difference: %.2f%%\n", (100.0 * (group_t - global_t))/((float)(global_t))) ;
	fprintf(mapeout, "\tMaximum difference: %.2f%%\n", (100.0 * (group_m - global_m))/((float)(global_m))) ;
}

/*
 * mape_group
 *
 * devide the circuit into groups and compare 
 * the difference between the vector for whole 
 * circuit maximum power estimation and the vector 
 * for a group.
 */
void 
mape_group(network)
network_t *network ;
{
	int i  ;

	for(i=4;i<=6;i++){
		mape_group_set_group_num(i) ;
		mape_group_static_reset() ;
		/*
		fprintf(mapeout, "Circuit %s:\n\n", network->net_name) ;
		*/

		/* devide the circuit into groups */
		mape_group_divide(network) ;

		/* get the vectors for maximum power estimation*/ 
		mape_group_estimate() ;
		lsDestroy(groups, free) ;

		mape_group_report_static() ;
	}
}

/*
 * mape_group_get_groups
 *
 * get one group from the network 
 * it's a interface to the outside to
 * return a group of nodes.  
 */
lsList 
mape_group_get_groups(network)
network_t *network ;
{
	if(groups)
		return groups ;

	mape_group_set_group_num(100) ;
	mape_group_divide(network) ;
	return groups ;
}
