/*
 * mape_itr.c
 *
 * Using iterative method to obtain the max switch
 * activity. 
 *
 * Changbo Long.  Nov.16, 2001. 
 */
#include "mape_int.h"

#define MAX_FETCH_TIMES 	10
#define MAX_NO_EFFECT_LOOP	500
#define MAX_UNTRAN_CANDIDATE 	200 /*store at more MAX_UNTRAN_STORE node as candidate*/

#define untran_list 		(untran_pool->list) 

typedef struct mape_untran_pool_struct {
	lsList 		list;
	int 		min_p ;	
	lsHandle	min_h ;
	int 		max_p ;
	lsHandle	max_h ; 
	int		(*f)(); 	/*priority function*/
	int		max_entry; 	/*define the maximum entry number*/
}mape_untran_pool_t;

mape_untran_pool_t *untran_pool = NULL; /*efficient untransistion node's pool */
lsList change_list = NULL; /*this list record all the transistion changed node when change the pi's value*/
lsList backtrace_list = NULL ; /* record all the backtrace node. */
lsList noeffect_list = NULL ;
int gain = 0 ;
int stop_times = 0 ;

static mape_node_value_t mape_itr_next_value(node_t *) ;
static mape_node_value_t mape_itr_simulate_node(node_t *node) ;
static void mape_itr_push_fetch(node_t *) ;
extern int mape_itr_node_sort_value(node_t *) ;

/*
 * mape_itr_untran_empty
 *
 * Is the untransis list empty. 
 */
int
mape_itr_untran_empty(void)
{
	if(untran_pool == NULL)
		return TRUE ;
	return (lsLength(untran_list) == 0) ;
}

/*
 * mape_itr_fetch_times
 *
 * find node's fetch times. 
 */
int
mape_itr_fetch_times(node)
node_t *node ;
{
	if(node == NULL)
		return 0;
	if(!(node->mape))
		return 0 ;
	return ((mape_t *)(node->mape))->visit ;
}

/*
 * mape_itr_is_noeffect
 *
 * is the node in the noeffect list. 
 */
int
mape_itr_is_noeffect(node)
node_t *node ;
{
	lsGen gen ;
	node_t *p ;
	int rtv = FALSE ;

	if(!noeffect_list)
		return FALSE;
	lsForeachItem(noeffect_list, gen, p){
		if(node == p)
			rtv = TRUE ;
	}
	return rtv ;
}

/*
 * mape_itr_clear_noeffect
 *
 * clear the noeffect list. 
 */
void 
mape_itr_clear_noeffect(void)
{
	node_t *node ;

	if(!noeffect_list)
		return ;
	while(lsLength(noeffect_list))
		lsDelEnd(noeffect_list, (lsGeneric *)&node) ;
}

/*
 * mape_itr_push_noeffect
 *
 * push node into the noeffect_list. 
 */
void 
mape_itr_push_noeffect(node)
node_t *node ;
{
	if(!noeffect_list)
		noeffect_list = lsCreate() ;
	lsNewEnd(noeffect_list, (lsGeneric)node, LS_NH);
}

/*
 * mape_itr_push_backtrace
 *
 * push the backtrace node into the backtrace list. 
 */
void 
mape_itr_push_backtrace(node)
node_t *node ;
{
	if(!backtrace_list)
		backtrace_list = lsCreate() ;

	lsNewEnd(backtrace_list, (lsGeneric)node, LS_NH);
}

/*
 * mape_itr_clear_backtrace_list.  
 *
 * clear backtrace, if gain>0, do nothing. 
 * if gain <= 0, increase the fetch time 
 * for every node. 
 */
void 
mape_itr_clear_backtrace(inc_fetch)
int inc_fetch ;
{
	node_t *node ;

	while(lsLength(backtrace_list)){
		lsDelEnd(backtrace_list, (lsGeneric *)&node) ;
		if(inc_fetch)
			mape_itr_push_fetch(node) ;
	}
}

/*
 * mape_itr_push_fetch
 *
 * push the node into the fetch's list. 
 */
void 
mape_itr_push_fetch(node)
node_t *node ;
{
	if(node == NULL)
		return ;
	if(!(node->mape))
		node->mape = (char *)mape_create_mape_t() ;

	((mape_t *)(node->mape))->visit ++ ;
}

/*
 * mape_itr_push_change
 *
 * push the transition changed node into the change list. 
 */
void
mape_itr_push_change(node)
node_t *node ;
{
	if(node == NULL)
		return ;
	if(!mape_is_given_node(node))
		return ;
	if(change_list == NULL) 
		change_list = lsCreate() ;

	if(node_num_fanout(node) > 0)
		lsNewEnd(change_list, (lsGeneric)node, LS_NH) ;
}

/*
 * mape_itr_create_untran_pool
 *
 *
 * create the untransistion pool. 
 */
mape_untran_pool_t *
mape_itr_create_untran_pool(f, max_entry)
int (*f)() ;
int max_entry ;
{
	mape_untran_pool_t *rtv ;

	rtv 		= ALLOC(mape_untran_pool_t, 1) ;
	rtv->list 	= lsCreate() ;
	rtv->f		= f ;
	rtv->max_p	= -999 ;
	rtv->max_h	= (lsHandle)0 ;
	rtv->min_p	= 999 ;
	rtv->min_h	= (lsHandle)0 ; ;
	rtv->max_entry	= max_entry ;

	return rtv ;
}

/*
 * mape_itr_untran_pool_refresh
 *
 * refresh the statistics in untran pool 
 */
void 
mape_itr_untran_pool_refresh(void)
{
	lsGen gen ;
	node_t *p ;
	int sort_value ;
	lsHandle h ;

	untran_pool->max_p	= -999 ;
	untran_pool->max_h	= (lsHandle)0 ;
	untran_pool->min_p	= 999 ;
	untran_pool->min_h	= (lsHandle)0 ; ;

	gen = lsStart(untran_list) ;
	while(lsNext(gen, (lsGeneric *)&p, &h) == LS_OK){
		sort_value = (*(untran_pool->f))(p) ;
		if(sort_value > untran_pool->max_p){
			untran_pool->max_p = sort_value ;
			untran_pool->max_h = h ;
		}
		if(sort_value < untran_pool->min_p){
			untran_pool->min_p = sort_value ;
			untran_pool->min_h = h ;
		}
	}
	lsFinish(gen) ;
}

/*
 * mape_itr_push_untransis
 *
 * push untransisted node into untransis node list. 
 */
void
mape_itr_push_untran(node)
node_t *node ;
{
	int sort_value ;
	lsHandle h ;
	node_t *tmp ;

	/* if node == NULL, update the min and max statistics */
	if(node == NULL)
		return ;
	if(node_num_fanout(node) < 1)
		return ;
	if(!mape_is_given_node(node))
		return ;

	if(untran_pool == NULL) 
		untran_pool = mape_itr_create_untran_pool(mape_itr_node_sort_value, MAX_UNTRAN_CANDIDATE) ;

	sort_value = (*(untran_pool->f))(node) ;
	if(lsLength(untran_list) < untran_pool->max_entry){
		lsNewEnd(untran_list, (lsGeneric)node, &h) ;
		if(sort_value > untran_pool->max_p){
			untran_pool->max_p = sort_value ;
			untran_pool->max_h = h ;
		}
		if(sort_value < untran_pool->min_p){
			untran_pool->min_p = sort_value ;
			untran_pool->min_h = h ;
		}
	}else if(sort_value > untran_pool->min_p){ 
		lsNewEnd(untran_list, (lsGeneric)node, LS_NH) ;
		LS_ASSERT(lsRemoveItem(untran_pool->min_h, (lsGeneric*)&tmp)) ;
		mape_itr_untran_pool_refresh() ;
	}
}

/*
 * mape_itr_node_reps. 
 *
 * push_node's reps list. 
 */
void 
mape_itr_node_reps(node)
node_t *node ;
{
	node_t *p ;
	int i ;

	foreach_fanin(node, i, p){
		if(node_type(p) == PRIMARY_INPUT)
			mape_node_push_reps(node, p) ;
		else 
			mape_node_push_reps_list(node, ((mape_t*)(p->mape))->reps) ;
	}
}

/*
 * mape_itr_imp_do
 *
 *
 * for mape_imp utility. carry out the simulation 
 * and store the untransisted node. 
 */
int 
mape_itr_imp_do(node, arg)
node_t *node ;
char *arg ; /*no use */
{
	mape_node_value_t value;

	/* simulate the node's value*/
	value = mape_itr_simulate_node(node) ;

	if(value != MAPE_FALL && value != MAPE_RISE)
		mape_itr_push_untran(node) ;
	
	if(node->type != PRIMARY_INPUT){
		mape_push(node, value, node, value) ;
		/*process the node's reps list*/
		mape_itr_node_reps(node) ;
	}

	return TRUE ;
}

/*
 * mape_itr
 *
 * initializely set all input to be FALL. and carry out 
 * the simulation. 
 */
void 
mape_itr_init(network)
network_t *network ;
{
	lsGen gen ;
	node_t *pi ;

	mape_imp_reset(mape_itr_imp_do, NULL) ; 
	foreach_primary_input(network, gen, pi){
		mape_push(pi, MAPE_FALL, pi, MAPE_FALL) ;
		mape_imp_push(pi) ;
	}
	mape_imp_start() ;
}

/*
 * mape_itr_fetch_untran
 *
 * fetch node from the untran_list which has the 
 * highest priority. 
 */
node_t *
mape_itr_fetch_untran(f, prev_node)
int (*f)() ;
node_t *prev_node ;
{
	lsGen gen ;
	int pri = -99, new_pri = -99;
	int i ;
	node_t *p ;
	node_t *rtv = NULL ;
	lsHandle ht, h ;

	if(mape_itr_untran_empty())
		return NULL ;

	if(lsLength(untran_list) == 1){
		lsFirstItem(untran_list, (lsGeneric *)&rtv, LS_NH) ; 
		return rtv ;
	}

	if(untran_pool->max_h == NULL)
		return NULL ;

	rtv = (node_t *)lsFetchHandle(untran_pool->max_h) ;
	if(!mape_itr_is_noeffect(rtv))
		return rtv ;

	gen = lsStart(untran_list) ;
	while(lsNext(gen, (lsGeneric *)&p, &h) == LS_OK){
		if(mape_itr_is_noeffect(p))
			continue;
		new_pri = (int)((*f)(p)) ;
		if(new_pri > pri){
			pri = new_pri ;
			ht = h ;
			rtv = p ;
		}
	}
	lsFinish(gen) ;
	return rtv ;
}

/*
 * mape_itr_node_sort_value
 *
 * return the node's sort value for taking 
 * the priority.  
 */
int 
mape_itr_node_sort_value(node)
node_t *node ;
{
	return node_num_fanout(node) - mape_itr_fetch_times(node) - 2*mape_get_node_level(node);
}

/*
 * mape_itr_node_priority
 *
 * return node's priority when backtrace.
 */
int 
mape_itr_node_priority(node)
node_t *node ;
{
	int times = 1 ;

	if(!TRANSITION_VALUE(mape_top(node))){
		times = 1*(mape_is_given_node(node)?2:1) ;
	}
	return (node_num_fanout(node)*2-mape_itr_fetch_times(node)) * times ;
}

/*
 * mape_itr_change_acceptable
 *
 * is the change acceptable? 
 */
int 
mape_itr_change_acceptable(void)
{
	/*fprintf(mapeout, "gain=%d, untran=%d, change_list=%d\n", gain, lsLength(untran_list), change_list == NULL? 0:lsLength(change_list)) ;*/
	return (gain > 0) ;
}

/*
 * mape_itr_change_do
 *
 * the change process implication do. 
 */
int
mape_itr_change_do(node, arg)
node_t *node ;
char *arg ;
{
	mape_nv_pair_t *pi_pair = (mape_nv_pair_t *)arg ;
	mape_node_value_t new_value, old_value ;

	if(node == NULL)
		return ;

	old_value = mape_top(node) ;
	new_value = mape_itr_simulate_node(node) ;
	if(node->type != PRIMARY_INPUT && new_value == old_value)
		return FALSE ;

	if(node->type != PRIMARY_INPUT)
		mape_push(node, new_value, pi_pair->node, pi_pair->value) ;

	if(TRANSITION_VALUE(old_value) != TRANSITION_VALUE(new_value) && mape_is_given_node(node)){
		mape_itr_push_change(node);
		gain += node_num_fanout(node) * (TRANSITION_VALUE(new_value)?1:-1) ; 
	}
	return TRUE ;
}

/*
 * mape_itr_backtrack_do
 *
 * for mape_imp utility. using for backtrack. 
 */
int
mape_itr_backtrack_do(node, arg)
node_t *node ;
char *arg ;
{
	node_t *pi = ((mape_nv_pair_t *)arg)->node ;
	mape_node_value_t pi_value = ((mape_nv_pair_t *)arg)->value ;

	if(mape_pop(node, pi, pi_value) == TRUE){
		return TRUE ;
	}else{
		return FALSE ;
	}
}

/*
 * mape_itr_backtrack_change
 * 
 * unchang the pi's value and implication
 * 
 */
void 
mape_itr_backtrack_change(pi)
node_t *pi ;
{
	mape_nv_pair_t pi_pair ;

	pi_pair.node = pi ;
	pi_pair.value = mape_top(pi) ;

	mape_imp_reset(mape_itr_backtrack_do, (char *)(&pi_pair)) ;
	mape_imp_push(pi) ;
	mape_imp_start() ;
}

/*
 * mape_itr_max_no_effect_loop
 *
 * return the maximum no effective loop. 
 */
int
mape_itr_max_noeffect_loop(network)
network_t *network ;
{
	return MIN(network_num_pi(network) * 8, MAX_NO_EFFECT_LOOP) ;
}

/*
 * mape_itr_refine_stop
 *
 * stop the processing of the refine. 
 */
int
mape_itr_refine_stop(network)
network_t *network ;
{
	if(gain > 0)
		stop_times = 0 ;
	else 
		stop_times ++ ;

	if(stop_times >= mape_itr_max_noeffect_loop(network))
		return TRUE ;
	else 
		return FALSE ;
	/*
	mape_report(global_network, FALSE, 0) ;
	*/
}

/*
 * mape_itr_next_value
 *
 * return the next value for the pi. 
 */
mape_node_value_t
mape_itr_next_value(node)
node_t *node ;
{
	mape_node_value_t value = mape_top(node);
	int t = node_num_fanout(node) - mape_itr_fetch_times(node) ;

	switch(value){
		case MAPE_FALL:	
			return MAPE_RISE ;
		case MAPE_RISE:
			if(t <= 2)
				return MAPE_1_1 ;
			else 
				return MAPE_FALL;
		case MAPE_1_1:
			return MAPE_0_0 ;
		case MAPE_0_0:
		default:
			return MAPE_FALL ;
	}
}

/*
 * mape_itr_change_pi
 * 
 * chang the pi's value and implication
 * it. 
 */
void 
mape_itr_change_pi(pi)
node_t *pi ;
{
	mape_node_value_t old_value, new_value ;	
	mape_nv_pair_t pi_pair ;

	gain = 0 ;
	old_value = mape_top(pi) ;
	new_value = mape_itr_next_value(pi) ;
	pi_pair.node = pi ;
	pi_pair.value = new_value ;

	/*
	fprintf(mapeout, "change pi %s from %s ", pi->name, mape_print_node_value(mape_top(pi))) ;
	fprintf(mapeout, "to %s \n", mape_print_node_value(new_value)) ;
	*/

	mape_push(pi, new_value, pi, new_value) ;
	if(TRANSITION_VALUE(old_value) != TRANSITION_VALUE(new_value) && mape_is_given_node(pi)){
		mape_itr_push_change(pi);
		//gain += node_num_fanout(pi) * (TRANSITION_VALUE(new_value)?1:-1) ; 
	}
	mape_imp_reset(mape_itr_change_do, (char *)(&pi_pair)) ;
	mape_imp_push(pi) ;
	mape_imp_start() ;
}


/*
 * mape_itr_backtrace. 
 *
 * backtrace to pi. 
 */
node_t *
mape_itr_backtrace(node)
node_t *node ;
{
	int i ;
	node_t *p, *next = NULL ;
	int pri = -1, new_pri = -1; 

	mape_itr_push_backtrace(node) ;
	if(node_type(node) == PRIMARY_INPUT)
		return node;

	foreach_fanin(node, i, p){
		new_pri = mape_itr_node_priority(p) ;
		if(!next || new_pri > pri){
			next = p ;
			pri = new_pri ;
		}
	}
	return mape_itr_backtrace(next) ;
}

/*
 * mape_itr_clear_change_list. 
 *
 * clear the transistion change list 
 */
void 
mape_itr_clear_change_list(void)
{
	node_t *tmp ;

	if(!change_list)
		return ;

	while(lsLength(change_list))
		lsDelEnd(change_list, (lsGeneric *)&tmp) ;
}

/*
 * mape_itr_pop_untran
 *
 * pop a node from the untran list. 
 */
int
mape_itr_pop_untran(node)
node_t *node ;
{
	lsGen gen ;
	lsHandle h, ht ;
	int find = FALSE ;
	node_t *p ;

	if( (*(untran_pool->f))(node) <= (untran_pool->min_p) )
		return FALSE;

	gen = lsStart(untran_list) ;
	while(lsNext(gen, (lsGeneric *)&p, &h) == LS_OK){
		if(p == node){
			ht = h;
			find = TRUE ;
			break ;
		}
	}
	lsFinish(gen) ;

	if(find)
		LS_ASSERT(lsRemoveItem(ht, (lsGeneric *)&p)) ;

	return find ;
}

/*
 * mape_itr_list_remove. 
 *
 * remove a node from a node's list. 
 */
void 
mape_itr_untran_remove(node)
node_t *node ;
{
	lsGen gen ;
	node_t *p ;
	lsHandle h, ht = 0 ;
	int refresh = FALSE ;

	gen = lsStart(untran_list) ;
	while(lsNext(gen, (lsGeneric *)&p, &h) == LS_OK){
		if(node == p){
			ht = h ;
			break ;
		}
	}
	lsFinish(gen) ;
	
	if(ht){
		if(ht == untran_pool->max_h || ht == untran_pool->min_h)
			refresh = TRUE ;
		LS_ASSERT(lsRemoveItem(ht, (lsGeneric *)&p)) ;
		if(refresh)
			mape_itr_untran_pool_refresh() ;
	}
}

/*
 * mape_itr_update_untran_pool
 *
 * update the untransition list according 
 * to the change list as well as clearing 
 * the change list. 
 */
void 
mape_itr_update_untran_pool(void)
{
	node_t *node, *tmp ;
	int sort_value ;
	
	while(lsLength(change_list)){
		lsDelEnd(change_list, (lsGeneric *)&node) ;
		
		sort_value = (*(untran_pool->f))(node) ;
		if(TRANSITION_VALUE(mape_top(node))){
			if(sort_value > untran_pool->min_p){
				lsNewEnd(untran_list, (lsGeneric)node, LS_NH) ;
				LS_ASSERT(lsRemoveItem(untran_pool->min_h, (lsGeneric*)&tmp)) ;
				mape_itr_untran_pool_refresh() ;
			}
		}else{
			if(sort_value >= untran_pool->min_p){
				mape_itr_untran_remove(node) ;
			}
		}
	}
}

/*
 * mape_itr_refine
 *
 * refine the result of the initilized value. 
 */
void 
mape_itr_refine(network)
network_t *network ;
{
	node_t *node ;
	node_t *pi ;
	node_t *prev_node = node;

	for(;;){
		if((node = mape_itr_fetch_untran(mape_itr_node_sort_value, prev_node)) == NULL) 
			break;
		pi = mape_itr_backtrace(node) ;
		/*
		fprintf(mapeout, "backtrace from node %s(%d) to pi %s\n", node->name, node_num_fanout(node), pi->name) ;
		*/
		mape_itr_change_pi(pi) ;
		/*
		mape_timer_start() ;
		*/
		if(!mape_itr_change_acceptable()){
			mape_itr_backtrack_change(pi) ;
			mape_itr_clear_change_list();
			prev_node = node ;
			mape_itr_clear_backtrace(TRUE) ;
			mape_itr_push_noeffect(node);
		}else{
			mape_itr_update_untran_pool() ;
			prev_node = NULL ;
			mape_itr_clear_backtrace(FALSE) ;
			mape_itr_clear_noeffect() ;
		}
		mape_itr_untran_pool_refresh() ;
		/*
		mape_timer_stop() ;
		*/
		if(mape_itr_refine_stop(network))
			break ;
	}
}

/*
 * mape_itr
 *
 * using iterative method to obtain the maximum
 * switching activity. 
 */
void 
mape_itr(network)
network_t *network ;
{
	lsGen gen ;
	node_t *p ;

	/*
	mape_timer_reset() ;
	*/
	mape_itr_init(network) ;

	/*refine the result*/
	mape_itr_refine(network) ;
	/*
	mape_timer_report("mape timer") ;
	*/

}

/*
 * mape_itr_simulate_node
 *
 * simulate the node according to the value in the podem's stack. 
 * Just using the cude's sharp product operation to finish it. 
 * It is 9 value based. 
 */
mape_node_value_t 
mape_itr_simulate_node(node)
node_t *node ;
{
	int value ;
	pcube input ;
	int i ;
	int oF, oS ; //first and second output. 
	mape_node_value_t o ;
	node_t *p ;


	o = mape_top(node) ;

	/* primary input's value must be set */
	if(node_type(node) == PRIMARY_INPUT)
		return o ; 

	cube.num_vars = node->nin ;
	define_cube_size (node->nin) ;
	input = new_cube() ;

	foreach_fanin(node, i, p){
		value = mape_top_first(p) ;
		PUTINPUT(input, i, value) ;
	}
	oF = mape_set_simulate_node(input, node) ; 
	foreach_fanin(node, i, p){
		value = mape_top_second(p) ;
		PUTINPUT(input, i, value) ;
	}
	oS = mape_set_simulate_node(input, node) ; 
	o = mape_value_convert(oF, oS) ;

	free_cube(input) ;
	return o ;
}

/*
 * mape_itr_clean_untran_pool
 *
 * clean the untran pool.
 */
void 
mape_itr_clean_untran_pool(void)
{
	node_t *node ;

	if(!untran_pool)
		return ;
	untran_pool->max_p	= -999 ;
	untran_pool->max_h	= (lsHandle)0 ;
	untran_pool->min_p	= 999 ;
	untran_pool->min_h	= (lsHandle)0 ; ;
	while(lsLength(untran_list))
		LS_ASSERT(lsDelEnd(untran_list, (lsGeneric*)&node)) ;
}

void 
mape_clean_list(list)
lsList list ;
{
	lsGeneric tmp ;

	if(!list)
		return ;

	while(lsLength(list))
		LS_ASSERT(lsDelEnd(list, &tmp)) ;
		
}

/*
 * mape_clean_node. 
 *
 * clean the node. 
 */
void 
mape_clean_node(node)
node_t *node ;
{
	mape_t *m ;

	m = (mape_t *)node->mape ;
	if(!m)
		return ;

	m->visit 	= 0 ;
	m->C0	= 0 ;
	m->C1	= 0 ;
	m->rdm_v	= 0 ;
	m->rdm_max	= 0 ;
	if(m->values)
		mape_clean_list(m->values) ;
}

/*
 * mape_clean_imp_do
 *
 *
 * clean node. 
 */
int
mape_clean_imp_do(node, arg)
node_t *node ;
char *arg ;
{
	mape_clean_node(node) ;
	return TRUE ;
}


/*
 * mape_clean_nodes. 
 *
 * clean the stack and visit time for each node. 
 */
void 
mape_clean_nodes(network)
network_t *network ;
{
	lsGen gen ;
	node_t *pi ;

	mape_imp_reset(mape_clean_imp_do, NULL) ; 
	foreach_primary_input(network, gen, pi){
		mape_imp_push(pi) ;
		mape_clean_node(pi) ;
	}
	mape_imp_start() ;
}

/*
 * mape_itr_clean
 *
 *
 * clean the global data for the estimation 
 */
void 
mape_itr_clean(network)
network_t *network ;
{
	node_t *node ;
	mape_itr_clean_untran_pool();

	mape_clean_list(change_list) ;
	mape_clean_list(backtrace_list) ;
	mape_clean_list(noeffect_list) ;

	/*clean all the node's information */
	mape_clean_nodes(network) ; 
	gain = 0 ;
	stop_times = 0;
}

/*
 * mape_itr_fmt
 *
 * return the total fanout multiply transistion 
 */
int 
mape_itr_fmt(network)
network_t *network ;
{
	node_t *node ;
	lsGen gen ;
	int rtv = 0 ;
	int count = 0 ;

	foreach_node(network, gen, node){
		if(!mape_is_given_node(node))
			continue ;
		if(node_type(node) == PRIMARY_INPUT || node_type(node) == PRIMARY_OUTPUT) 
			continue ;
		count += LSLENGTH(node->fanout) ;
		rtv += (TRANSITION_VALUE(mape_top(node))?1:0) * LSLENGTH(node->fanout) ;
	}
	/*fprintf(mapeout, "rtv=%d, count=%d\n", rtv, count) ;*/
	return rtv ;
}
