/*
 * mape_sa.c 
 *
 * simulated annealing template. 
 * Changbo Long. (March. 1, 2002).  
 */
#include "mape_int.h"

typedef struct mape_sa_param_struct {
	double 	init_tempt ; 	/* the initial temperature rate */
	double 	tempt_rate ;	/* the temperature reduce rate */
	int 	set ;		/* parameter to indict if the parameters have been set*/

	char * 	(*create)(void) ;  /* create the mem for solution */
	void 	(*destroy)(char *) ;	/* destroy the solution mem*/
	double 	(*cost)(char *) ;  /* evaluate the cost for one solution */
	void 	(*move)(char *, char *) ; 	/* move from one solution to another */
	int 	(*equilibrum)(int, int, int) ;	/* equilibrum function to judge if equilibrum */
	int 	(*stop)(double, int, double) ;	/* function to judge if stop */
	void	(*accept)(int) ;		/* if the last move be accepted */
}mape_sa_param_t  ;

mape_sa_param_t *sa_param = NULL ;

void mape_sa_test_move(char *, char *) ;
double mape_sa_test_cost(char *) ;
char *mape_sa_test_create(void) ;
void mape_sa_test_destroy(char *) ;
int mape_sa_test_equilibrum(int, int, int) ;
int mape_sa_test_stop(double, int, double) ;
void mape_sa_test_accept(int) ;

/*
 * mape_sa_set
 *
 * set all the parameters for the sa algorithm. 
 */
void 
mape_sa_set(init_tempt, tempt_rate, create, destroy, cost, move, equilibrum, stop, accept) 
double init_tempt ;
double tempt_rate ;
char * (*create)(void) ;
void  (*destroy)(char *) ;
double (*cost)(char*) ;
void (*move)(char *, char *) ;
int (*equilibrum)(int, int, int) ;
int (*stop)(double, int, double) ;
void (*accept)(int)  ;
{
	if(sa_param == NULL)
		sa_param = ALLOC(mape_sa_param_t, 1) ;

	if(init_tempt > 0)	
		sa_param->init_tempt = init_tempt ;
	else
		sa_param->init_tempt = 1000 ;

	if(tempt_rate > 0 && tempt_rate < 1)
		sa_param->tempt_rate = tempt_rate ;
	else
		sa_param->tempt_rate = 0.9 ;

	sa_param->create 	= create ;
	sa_param->destroy 	= destroy ;
	sa_param->cost 		= cost ;	
	sa_param->move 		= move ;	
	sa_param->set 		= TRUE ;
	sa_param->equilibrum	= equilibrum ;
	sa_param->stop		= stop ;
	sa_param->accept	= accept ;
}

/*
 * mape_sa_start
 *
 * Start the simulate annealing procedure. 
 * The final solution will be filled in the  
 * initial solution. 
 */
void
mape_sa_start(init_sol)
char **init_sol ;
{
	char *new_sol, *old_sol ;
	double old_cost, new_cost, delt, temp, prob ;
	long r ;
	int t_up_time, t_total_time, total_time, accept_time ;
	double accept_ratio ;
	int i ;

	if(sa_param == NULL){
		fprintf(stderr, "Mape SA: please set the parameters first!") ;
		return ;
	}
	if(!sa_param->set){
		fprintf(stderr, "Mape SA: please set the parameters first!") ;
		return ;
	}

	old_sol = *init_sol;
	old_cost = (*(sa_param->cost))(old_sol) ;
	new_sol = NULL ;
	temp = sa_param->init_tempt ;
	
	srandom(util_cpu_time()) ;
	total_time = 0 ; 
	accept_time = 0 ;

	i = 0 ;
	while( !(*(sa_param->stop))(temp, total_time, accept_ratio) ){
		t_up_time = 0 ;
		t_total_time = 0 ;
		accept_time = 0 ;
		//fprintf(mapeout, "total time is %d\n", total_time) ;
		while( !(*(sa_param->equilibrum))(t_up_time, t_total_time, accept_time) ){
			i ++ ;
			if(new_sol == NULL)
				new_sol = (*(sa_param->create))() ;
			(*(sa_param->move))(old_sol, new_sol) ;
			new_cost = (*(sa_param->cost))(new_sol) ;
			delt = new_cost - old_cost ;
			prob = MIN(1.0, exp(-delt/temp)) ;
			r = random() ;
			if ( ((double)(r%1000))/1000.0 <= prob && prob > 0.0){
				/*
				if(i >= 100){
					fprintf(mapeout, "prob=%lf, find another solution old_cost=%lf, new_cost=%lf, tempt=%.8lf!\n", 
						prob, old_cost, new_cost, temp) ;
					i = 0 ;
				}
				*/
				if(delt < 0) 
					t_up_time ++ ;
				if(delt != 0)
					accept_time ++ ;
				(void)(*(sa_param->destroy))(old_sol) ;
				old_sol = new_sol ;
				old_cost = new_cost ;
				new_sol = NULL ;
				(void)(*(sa_param->accept))(TRUE) ;
			}else 
				(void)(*(sa_param->accept))(FALSE) ;

			total_time ++ ;
			t_total_time ++ ;
		}
		temp *= sa_param->tempt_rate ;
		accept_ratio = ((double)(accept_time))/((double)(t_total_time)) ;
	}
	(void)(*(sa_param->destroy))(new_sol) ;
	sa_param->set = FALSE ;
	*init_sol = old_sol ;
}

/*
 * mape_sa_test_create
 *
 */
char *
mape_sa_test_create(void)
{
	double *r ;

	r = ALLOC(double, 1) ;
	return (char *)r ;
}

/*
 * mape_sa_test_destroy
 *
 */
void 
mape_sa_test_destroy(s) 
char *s ;
{
	free(s) ;
}

double
mape_sa_test_cost(sol)
char *sol ;
{
	double x = *((double *)sol) ;
	return 0.5*x*sin(x) ;
}

void 
mape_sa_test_move(from, to)
char *from ;
char *to ;
{
	long r ;
	double x ;
	double re ;

	r = random() ;
	x = (double)(r%5000) ;
	re = (x/10000-0.25) + *((double *)from) ;
	if(re >= 50)
		re -= 50 ;
	if(re <= 0)
		re += 50 ;

	*((double *)to) = re ;
}

int
mape_sa_test_equilibrum(up, total, accept_time)
int up , total, accept_time ;
{
	if(up >= 500 || total >= 5000)
		return TRUE ;
	else 
		return FALSE;
}

int
mape_sa_test_stop(temp, total, reject_ratio)
double temp, reject_ratio  ; 
int total ;
{
	if(temp <= 0.001 )
		return TRUE ;
	else 
		return 	FALSE ;
}

/*
 * callback function for acception of the 
 * move. 
 *
 * mape_sa_test_accept
 * do nothing. 
 */
void 
mape_sa_test_accept(a)
int a ;
{
}

/* 
 * mape_sagrp_test
 *
 * test for the sa algorithm implementation. 
 */
void 
mape_sag_test(network)
network_t *network ;
{
	
	double x = 20.0 ;
	char *sol ;

	srandom(util_cpu_time()) ;
	mape_sa_set(10000, 0.9, mape_sa_test_create, mape_sa_test_destroy, 
			mape_sa_test_cost, mape_sa_test_move, mape_sa_test_equilibrum, mape_sa_test_stop, 
			mape_sa_test_accept)  ;
	sol = (char *)&x ;
	mape_sa_start(&sol) ;

	fprintf(mapeout, "The optimal value obtained by the sa function is: %lf\n", *((double*)sol) ) ;
}
