/*
 * mape_sascc.c
 *  
 * using simulated annealing to solve the SSC(simultaneous switch current 
 * problem under 3 circuit model: zero delay model, unit delay model and real
 * delay model. 
 *
 * Changbo Long. 	Mar. 19, 2002. 
 */
#include "mape_int.h"

#define OBJECT_SUM 	1
#define OBJECT_MAX 	2

double 	mape_sassc_cost(char *) ;
void 	mape_sassc_accept(int) ;

int group_type ; 		/* current group type */
int group_row, group_col ; 	/* current group row and current group col */
double lg_cost = 0.0 ; 		/*last group's cost */
double b_lg_cost = 0.0 ; 	/*backup of the lg cost */
lsList cost_list = NULL ; 	/*cost list for each group*/
lsList b_cost_list = NULL ;		/*backup the cost list fore each group*/

/*
 * mape_sassc_group_start
 *
 * initialize the group stream. 
 */
void 
mape_sassc_group_start(type)
int type ;
{
	group_type 	= type ;
	group_row	= 1;
	group_col	= 1;
}

/*
 * mape_sassc_push_cost_list. 
 *
 * push a group's cost into cost's list. 
 */
void
mape_sassc_push_cost_list(value)
double value ;
{
	double *cost ;

	cost = ALLOC(double, 1) ;
	*cost = value ;
	
	if(!cost_list)
		cost_list = lsCreate() ;
	lsNewEnd(cost_list, (lsGeneric)cost, LS_NH) ;
}

/*
 * mape_sassc_update_cost_list. 
 *
 * update one value in the cost list
 */
double
mape_sassc_update_cost_list(ind, value)
int ind ;
double value ;
{
	int count = 0 ;
	lsGen gen ;
	double *cost ;
	double rtv ;

	//fprintf(mapeout, "Update cost %d->%.2lf\n", ind, value) ; 
	lsForeachItem(cost_list, gen, cost){
		if(count == ind){
			rtv = value - *cost ;
			*cost = value ;
			lsFinish(gen) ;
			return rtv ;
		}
		count ++ ;
	}
}

/*
 * mape_sassc_backup_cost_list. 
 *
 * backup the cost list. 
 */
void 
mape_sassc_backup_cost_list(void)
{
	double *cost, *n_cost ;
	lsGen gen1, gen2 ;
	lsHandle h1, h2 ;
	int i=0, length  ;

	if(cost_list == NULL)
		return ;

	length = lsLength(cost_list) ;
	if(b_cost_list == NULL){
		b_cost_list = lsCreate() ;
		for(i=0;i<length;i++){
			n_cost = ALLOC(double, 1) ;
			*n_cost = 0.0 ;
			lsNewEnd(b_cost_list, (lsGeneric)n_cost, LS_NH) ;
		}
	}
	gen1 = lsStart(cost_list) ;
	gen2 = lsStart(b_cost_list) ;
	//fprintf(mapeout, "backup cost list:\n") ;

	while(lsNext(gen1, (lsGeneric *)&cost, &h1) == LS_OK && lsNext(gen2, (lsGeneric *)&n_cost, &h2) == LS_OK){
		//fprintf(mapeout, "%.2lf ", *cost) ; 
		*n_cost = *cost ;
	}

	//fprintf(mapeout, "\n") ;

	lsFinish(gen1) ;
	lsFinish(gen2) ;
}

/*
 * mape_sassc_restore_cost_list
 *
 * restore the cost list. 
 */
void 
mape_sassc_restore_cost_list(void)
{
	double *cost, *n_cost ;
	lsGen gen1, gen2 ;
	lsHandle h1, h2 ;

	if(b_cost_list == NULL || cost_list == NULL)
		return ;

	gen1 = lsStart(cost_list) ;
	gen2 = lsStart(b_cost_list) ;
	//fprintf(mapeout, "Restore cost list:\n") ; 

	while(lsNext(gen1, (lsGeneric *)&cost, &h1) == LS_OK && lsNext(gen2, (lsGeneric *)&n_cost, &h2) == LS_OK){
		//fprintf(mapeout, "%.2lf ", *n_cost) ;
		*cost = *n_cost ;
	}
	//fprintf(mapeout, "\n") ;
	lsFinish(gen1) ;
	lsFinish(gen2) ;
}

/*
 * mape_sassc_is_group_changed
 *
 * has the group has been changed
 */
int 
mape_sassc_is_group_changed(group, ex_n1, ex_n2)
lsList group ;
node_t *ex_n1, *ex_n2 ;
{
	int is_latch1, is_latch2 , r , c;
	node_t *node, *latch, *node_r, *node_d, *node_c ;

	is_latch1 = mape_node_is_latch(ex_n1) ;
	is_latch2 = mape_node_is_latch(ex_n2) ;

	if((is_latch1 && is_latch2) || (!is_latch1 && !is_latch2)){
		if(mape_node_in_list(ex_n1, group))
			return TRUE ;
		if(mape_node_in_list(ex_n2, group))
			return TRUE ;
		return FALSE ;
	}

	/* one is latch and another one is not latch */
	if(is_latch1){
		latch = ex_n1 ;
		node  = ex_n2 ;
	}else {
		latch = ex_n2 ;
		node  = ex_n1 ;
	}
	
	r = ((mape_t *)(node->mape))->plt->row ;
	c = ((mape_t *)(node->mape))->plt->col ;
	node_r = mape_saplt_index_node(r, c+1) ;
	node_d = mape_saplt_index_node(r+1, c) ;
	node_c = mape_saplt_index_node(r+1, c+1) ;

	if(mape_node_in_list(latch, group) || mape_node_in_list(node, group) || mape_node_in_list(node_r, group)
		|| mape_node_in_list(node_d, group) || mape_node_in_list(node_c, group))
		return TRUE ;
	return FALSE ;
}

/*
 * mape_sassc_group_next
 *
 * return the next group nodes. 
 */
int
mape_sassc_group_next(group)
lsList group ;
{
	int row_max, col_max ;
	int r, c, i, j ;
	node_t *node ;
	int next_size ;

		
	if(group_row == 0 || group_col == 0) 
		return 	FALSE ;

	while(lsLength(group))
		lsDelEnd(group, (lsGeneric *)&node) ;

	mape_saplt_plt_size(&row_max, &col_max) ;

	for(i=0;i<FIX_GROUP_SIZE;i++){
		r = group_row + i ;
		if(r > row_max)
			break ;
		for(j=0;j<FIX_GROUP_SIZE;j++){
			c = group_col + j ;
			if(c > col_max)
				break ;
			node = mape_saplt_index_node(r, c)  ;
			lsNewEnd(group, (lsGeneric)node,  LS_NH) ;
		}
	}

	if(group_type == FIX_GROUP)
		next_size = FIX_GROUP_SIZE ;
	else if(group_type == MSP_GROUP)
		next_size = 1 ;

	if(group_col + FIX_GROUP_SIZE <= col_max)
		group_col += next_size ;
	else {
		if(group_row + FIX_GROUP_SIZE <= row_max){
			group_row += next_size ;
			group_col = 1 ;
		}else {
			group_row = 0 ;
			group_col = 0 ;
		}
	}
	return TRUE;
}

/*
 * mape_sassc_group_max_ssc
 *
 * return the maximum ssc for a group
 */
double
mape_sassc_group_max_ssc(group, type, time, switch_list)
lsList group, switch_list ;
int type ;
double *time ;
{
	static int ind = 0 ;
	lsGen gen ;
	node_t *node ;
	double gc ;
	int count = 0 ;

	mape_clean_given_gate() ;
	while(lsLength(switch_list))
		lsDelBegin(switch_list, (lsGeneric *)&node) ;

	//fprintf(mapeout, "group: \n") ;
	lsForeachItem(group, gen, node){
		//fprintf(mapeout, "%s ", node?node->short_name:"NULL") ;
		if(mape_mspplt_null_node(node))
			continue ;
		if(node){
			mape_push_given_gatep(node) ;
			count ++ ;
		}
	}
	if(!count)
		return 0.0 ;

	if(type == ZERO_DELAY_MODEL){
		mape_itr_clean(global_network) ;
		mape_itr(global_network) ;
		return (double)mape_itr_fmt(global_network) ;
	}else if(type == UNIT_DELAY_MODEL){ 
		mape_curitr_clear(global_network) ;
		gc = mape_curitr(global_network, time, switch_list) ;
		//gc = mape_cursa(global_network, time, NULL) ;
		return -gc ;
	}else if(type == REAL_DELAY_MODEL){
		mape_curitr_clear(global_network) ;
		gc = mape_curitr(global_network, time, switch_list) ;
		return -gc ;
	}
}

/*
 * mape_sassc_type_cost
 *
 * get the cost. 
 */
double 
mape_sassc_type_cost(g_type, m_type, o_type)
int g_type ; 	/* group type, FIX_GROUP or MSP_GROUP */
int m_type ; 	/* model type, ZERO_DELAY_MODEL, UNIT_DELAY_MODEL or REAL_DELAY_MODEL. */ 
int o_type ;	/* objective type, OBJECT_SUM or OBJECT_MAX */
{
	lsList group ;
	double cost  ;
	double ssc, diff ;
	node_t *ex_n1, *ex_n2 ;
	int ind = 0 ;
	double tmp = 0.0 ;
	double time ;

	mape_saplt_switched_node(&ex_n1, &ex_n2) ;

	group = lsCreate() ;
	cost = lg_cost ;
	/* back up the cost's list */
	b_lg_cost = lg_cost ;
	mape_sassc_backup_cost_list() ;

	mape_sassc_group_start(g_type) ;
	while(mape_sassc_group_next(group)){
		if(ex_n1 == NULL && ex_n2 == NULL){
			ssc = mape_sassc_group_max_ssc(group, m_type, &time, NULL) ;	
			fprintf(mapeout, "ssc=%.2lf\n", ssc) ;
			mape_sassc_push_cost_list(ssc) ;
			if(o_type == OBJECT_SUM)
				cost += ssc ;
			else if(o_type == OBJECT_MAX)
				cost = MAX(cost, ssc) ;
		}else{
			if(mape_sassc_is_group_changed(group, ex_n1, ex_n2)){
				ssc = mape_sassc_group_max_ssc(group, m_type, &time, NULL) ;
				diff = mape_sassc_update_cost_list(ind, ssc) ;
				if(o_type == OBJECT_SUM)
					cost += diff ;
				else if(o_type == OBJECT_MAX)
					cost = MAX(cost, ssc) ;
			}
		}
		ind ++ ;
	}
	lg_cost = cost ; 

	/*
	tmp = 0.0 ;
	mape_sassc_group_start(g_type) ;
	while(mape_sassc_group_next(group)){
		ssc = mape_sassc_group_max_ssc(group, m_type) ;
		if(o_type == OBJECT_SUM)
			tmp += ssc ;
		else if(o_type == OBJECT_MAX)
			tmp = MAX(tmp, ssc) ;
	}
	*/

	/*
	if(tmp != cost){
		fprintf(mapeout, "tmp=%lf, cost=%lf\n", tmp, cost) ;
		mape_error("cost calculation!", __FILE__, __LINE__) ;
	}
	*/

	lsDestroy(group, NULL) ;
	return cost ;
	/*
	return tmp ;
	*/
}

/*
 * mape_sassc_cost. 
 *
 * get the cost function. 
 */
double 
mape_sassc_cost(nouse)
char *nouse ;
{
	/*return cost according to different delay model and objective function in SSC*/
	if(user_option->delay_model == ZERO_DELAY_MODEL){
		if(user_option->ssc_obj == SSCOBJ_GROUP_SUM)
			return mape_sassc_type_cost(FIX_GROUP, ZERO_DELAY_MODEL, OBJECT_SUM) ;
		else if(user_option->ssc_obj == SSCOBJ_GROUP_MAX)
			return mape_sassc_type_cost(FIX_GROUP, ZERO_DELAY_MODEL, OBJECT_MAX) ;
		else if(user_option->ssc_obj == SSCOBJ_MSP_MAX)
			return mape_sassc_type_cost(MSP_GROUP, ZERO_DELAY_MODEL, OBJECT_MAX) ;
		else 
			mape_error("Can't handle objective function type\n", __FILE__, __LINE__) ;
	}else if(user_option->delay_model == UNIT_DELAY_MODEL){
		mape_curlib_set_unit_mode(TRUE) ;
		if(user_option->ssc_obj == SSCOBJ_GROUP_SUM)
			return mape_sassc_type_cost(FIX_GROUP, UNIT_DELAY_MODEL, OBJECT_SUM) ;
		else if(user_option->ssc_obj == SSCOBJ_GROUP_MAX)
			return mape_sassc_type_cost(FIX_GROUP, UNIT_DELAY_MODEL, OBJECT_MAX) ;
		else if(user_option->ssc_obj == SSCOBJ_MSP_MAX)
			return mape_sassc_type_cost(MSP_GROUP, UNIT_DELAY_MODEL, OBJECT_MAX) ;
		else 
			mape_error("Can't handle objective function type\n", __FILE__, __LINE__) ;
	}else if(user_option->delay_model == REAL_DELAY_MODEL){
		if(user_option->ssc_obj == SSCOBJ_GROUP_SUM)
			return mape_sassc_type_cost(FIX_GROUP, REAL_DELAY_MODEL, OBJECT_SUM) ;
		else if(user_option->ssc_obj == SSCOBJ_GROUP_MAX)
			return mape_sassc_type_cost(FIX_GROUP, REAL_DELAY_MODEL, OBJECT_MAX) ;
		else if(user_option->ssc_obj == SSCOBJ_MSP_MAX)
			return mape_sassc_type_cost(MSP_GROUP, REAL_DELAY_MODEL, OBJECT_MAX) ;
		else 
			mape_error("Can't handle objective function type\n", __FILE__, __LINE__) ;
	}else 
		mape_error("Can't handle delay model\n", __FILE__, __LINE__) ;
}

/*
 * mape_sassc_accept
 *
 * be notified that the move is accepted. 
 */
void 
mape_sassc_accept(ac)
int ac ;
{
	if(!ac){
		mape_sassc_restore_cost_list() ;
		//fprintf(mapeout, "restore the cost from %lf to %lf\n", lg_cost, b_lg_cost) ;
		lg_cost = b_lg_cost ;
	}
}

/*
 * mape_sassc
 *
 * using simulated annealing to solve the SSC problem under
 * 3 circuit delay model: zero delay mode, unit delay mode and real delay model. 
 */
void 
mape_sassc(network)
network_t *network ;
{
	double cost ;
	char *sol ;
	int delay_model ;

	srandom(util_cpu_time()) ;
	mape_saplt_init_plt(network) ;
	mape_saplt_dump() ;

	fprintf(mapeout, "Initially the total cost is %lf\n", mape_sassc_cost((char *)network)) ;
	lg_cost = b_lg_cost = 0.0 ;
	mape_sa_set(10, 0.9, mape_saplt_sa_create, mape_saplt_sa_destroy, 
		mape_sassc_cost, mape_saplt_move, mape_saplt_sa_equilibrum, 
		mape_saplt_sa_stop, mape_sassc_accept) ;

	sol = (char *)network ;
	mape_sa_start(&sol) ; 
	mape_saplt_dump() ;
	fprintf(mapeout, "Finally the cost is %lf\n", mape_sassc_cost((char *)network)) ; 
}
