/*
 * mape_stn.c
 *
 *
 * voltage evaluation for each transistor in the 
 * sleep transistor network(STN). 
 *
 * Changbo Long. (clong@cae.wisc.edu) 
 *	Nov. 28. 2002 Happy Thanksgiving! . 
 */
#include "mape_int.h"

#define  EPSILON_0	8.854E-14	// F/cm
#define  EPSILON_SIO2	3.9		// 
#define  T_OX		1.5E-9		// nm
#define  MU_N		415		// cm^2/(volt*sec)
#define  VTH		0.5		// vol
#define  VTL		0.3		// vol			
#define  VDD		1.3			// vol
#define  RI_1		((double)(10.0))	// ohm
#define  RI_5		((double)(50.0))	// ohm
#define  RI_10		((double)(100.0))	// ohm
#define  STN_MINCUR	300

/* all the data for information for the stn. */
typedef struct mape_stn_struct{
	int 	c_number; 	// cluster number ; 
	double 	Ri ;		// interconnect resistance. 
	double 	*Rst ;		// sleep transistor resistances. 	
	double  *CMSSC ;	// cluster MSSCs.  
	double  *vol ;		// obtained voltage for each cluster. 
	char    *matrix ;	// matrix A, where Ax=b ;
}mape_stn_t ;

mape_stn_t  stn = {0, 0.0, NULL, NULL, NULL, NULL} ;

double mape_stn_cal_resistance(double) ;
double mape_stn_cal_wl(double) ;
void mape_stn_isolator_figure(network_t *) ;
char *mape_stn_get_matrix(int, double, double *) ;

extern lsList mape_depsa_get_clusters(void) ;

lsList stn_groups = NULL ;
double *allocate_factor = NULL ;

/*
 * mape_stn_clear
 *
 * clear all the previous stn datas. 
 */
void
mape_stn_clear(void)
{
	stn.c_number 	= 0 ;
	stn.Ri	 	= 0.0 ;
	if(stn.Rst){
		free(stn.Rst) ;
		stn.Rst	= NULL ;
	}
	if(stn.CMSSC){
		free(stn.CMSSC) ;
		stn.CMSSC = NULL ;
	}
	if(stn.vol){
		free(stn.vol) ;
		stn.vol	= NULL ;
	}
}

/*
 * mape_stn_set_c_number
 *
 * set the cluster number. 
 */
void 
mape_stn_set_c_number(n)
int n ;
{
	stn.c_number 	= n ;
}

/*
 * mape_stn_set_Ri
 *
 * set the interconnect resistance value. 
 */
void 
mape_stn_set_Ri(r)
double r ;
{
	stn.Ri 	= r ;
}

/*
 * mape_stn_set_Rst
 *
 * set the sleep transistor resistances. 
 */
void
mape_stn_set_Rst(r, n)
double *r ;
int n ;
{
	int i ;
	if(n != stn.c_number)
		mape_error("Inequal cluster number !", __FILE__, __LINE__) ;

	if(stn.Rst == NULL)
		stn.Rst = ALLOC(double, n) ;
	for(i=0;i<n;i++)
		(stn.Rst)[i] = r[i] ;
}

/*
 * mape_stn_set_CMSSC 
 *
 *
 * set cluster current. 
 */
void 
mape_stn_set_CMSSC(c, n) 
double *c ;
int n ;
{
	int i ;
	if(n != stn.c_number)
		mape_error("Inequal cluster number !", __FILE__, __LINE__) ;

	if(stn.CMSSC == NULL)
		stn.CMSSC = ALLOC(double, n) ;
	for(i=0;i<n;i++)
		(stn.CMSSC)[i] = c[i] ;
}

/*
 * mape_stn_set
 *
 * set up all the datas. 
 */
void 
mape_stn_set(n, Ri, Rst, CMSSC) 
int n ;
double Ri ;
double *Rst, *CMSSC ;
{
	mape_stn_set_c_number(n) ;
	mape_stn_set_Ri(Ri) ;
	mape_stn_set_Rst(Rst, n) ;
	mape_stn_set_CMSSC(CMSSC, n) ;
}

/*
 * mape_stn_create_matrix
 *
 * clear the maxtri A, where Ax=b. 
 */
int
mape_stn_create_matrix(void)
{
	int spError ;
	double *matrixEntry ;
	int i, col, c ;

	if(stn.matrix)
		spClear(stn.matrix) ;
	else {
		stn.matrix = spCreate(stn.c_number, 0, &spError) ;
		assert(spError == spOKAY) ;
	}

	col = (int)(sqrt((double)(stn.c_number*1.0))) ;

	for(i=1;i<=stn.c_number;i++){		// one cluster, one row. 
		c = 0 ;
		if(i+col <= stn.c_number){
			c ++ ;
			matrixEntry = spGetElement(stn.matrix, i, i+col) ;
			spADD_REAL_ELEMENT(matrixEntry, -1/(stn.Ri));
		}
		if(i-col >= 1){
			c ++ ;
			matrixEntry = spGetElement(stn.matrix, i, i-col) ;
			spADD_REAL_ELEMENT(matrixEntry, -1/(stn.Ri));
		}
		if(i%col != 0 && i!=stn.c_number){
			c ++ ;
			matrixEntry = spGetElement(stn.matrix, i, i+1) ;
			spADD_REAL_ELEMENT(matrixEntry, -1/(stn.Ri));
		}
		if((i-1)%col != 0){
			c ++ ;
			matrixEntry = spGetElement(stn.matrix, i, i-1) ;
			spADD_REAL_ELEMENT(matrixEntry, -1/(stn.Ri));
		}
		matrixEntry = spGetElement(stn.matrix, i, i) ;
		spADD_REAL_ELEMENT(matrixEntry, 1/((stn.Rst)[i-1]) + c/stn.Ri);
	}
	spError = spFactor(stn.matrix) ;
	assert(spError == spOKAY) ;
	return TRUE ;
}

/*
 * mape_stn_solve
 *
 *
 * use spSolve to solve the STN equations.  
 */
int
mape_stn_solve(void)
{
	if(stn.vol)
		free(stn.vol);
	stn.vol = ALLOC(double, stn.c_number) ;
	spSolve(stn.matrix, stn.CMSSC, stn.vol) ;
	return TRUE ;
}

/*
 * Calculatin of the W/L ratio for any input current.  
 *
 * mape_stn_cal_wl
 */
double 
mape_stn_cal_wl(I)
double I ;	 // uA
{
	static double k = -99999.0 ;
	double c_ox ;

	if( fabs(k + 99999.0) <= 0.05 ){ // calculate the ration k. 
		c_ox = 100*(EPSILON_0*EPSILON_SIO2)/(T_OX) ;
		//fprintf(mapeout, "c_ox = %lf, ", c_ox ) ;
		k = 0.01/(0.05*MU_N*c_ox*(VDD-VTL)*(VDD-VTH)) ;	
		//fprintf(mapeout, "k = %lf\n", k) ; 
	}

	//fprintf(mapeout, "I=%.2lf, W/L = %.2lf\n", I, k*I);
	return (k*I) ; 
}

/*
 * calculation the Resistance by the L/W. 
 *
 * mape_stn_cal_resistance
 *
 */
double
mape_stn_cal_resistance(wl)
double wl ;
{
	static double k = -99999.0 ;
	double c_ox ;

	if( fabs(k + 99999.0) <= 0.05 ){ // calculate the ration k. 
		c_ox = 100*(EPSILON_0*EPSILON_SIO2)/(T_OX) ;
		//fprintf(mapeout, "c_ox = %lf, ", c_ox ) ;
		k = 10000/(MU_N*c_ox*(VDD-VTH)) ;
		//fprintf(mapeout, "k = %lf\n", k) ; 
	}

	//fprintf(mapeout, "W/L=%.2lf, R = %.2lf\n", wl, k/wl);
	return (k/wl) ; 

}

/*
 * mape_stn_pl_2
 *
 * get the maximum pl. 
 */
double
mape_stn_pl2(n, matrix, cmssc) 
int n ;
char *matrix ;
double *cmssc ;
{
	double rtv = 0.0 ;
	double pl ;
	double *vol ;
	int i ;

	vol = ALLOC(double, n) ; 
	spSolve(matrix, cmssc, vol) ;
	for(i=0;i<n;i++){
			pl = ((vol)[i]/(10000*(VDD-VTL))) ;
			rtv = MAX(pl, rtv) ;
	}
	free(vol) ;
	return rtv ;
}


/*
 * mape_stn_pl
 *
 *
 * performance loss calculation
 */
double
mape_stn_pl(n, Ri, Rst, cmssc)
int n ; 
double Ri, *Rst, *cmssc ;
{
	double rtv = 0.0 ;
	double pl ;
	int i ;

	mape_stn_set(n, Ri, Rst, cmssc) ;
	mape_stn_create_matrix() ;
	mape_stn_solve() ;
	for(i=0;i<n;i++){
			pl = ((stn.vol)[i]/(10000*(VDD-VTL))) ;
			rtv = MAX(pl, rtv) ;
	}

	return rtv ;
}

/*
 * mape_stn_depsa 
 *
 * the following part is to caculate the area and performance lost using depth way!
 */
lsList
mape_stn_dep_groups(r, network)
double *r ;
network_t *network ;
{
	double total_cur ; // MSSC for the whole circuit ; 
	double total_wl ;  // the total wl ; 
	lsList groups, g ; 
	lsGen gen ;
	double c, cmssc[1024], Rst[1024] ;
	double total_cmssc, wl, mpl, pl, ncur, t;
	int n,i,j ;
	lsList vpl, vol_points ;
	lsGeneric tmp ;
	lsGen gen1 ; 
	node_t *node ;

	mape_depsa(network) ;
	groups = mape_depsa_get_clusters() ;

	i = 0 ;
	lsForeachItem(groups, gen, g){
		c = mape_pbssce_itr_estimate_current(network, g) ;
		cmssc[i] = -c ;
		cmssc[i] = MAX(STN_MINCUR, cmssc[i]) ;
		i ++ ;
	}

	total_wl = 0.0 ;
	n = lsLength(groups) ;
	for(i=0;i<n;i++){
		c = cmssc[i] ;
		wl = mape_stn_cal_wl(c) ;
		total_wl += wl ;
		r[i] = mape_stn_cal_resistance(wl) ;
	}
	fprintf(mapeout, "DEPWL = %d\n", (int)(total_wl)) ;
	return groups ;
}


/*
 * mape_stn_stn_groups 
 *
 * return the groups and fill in the Rst, Rst_e vector.  
 */
lsList 
mape_stn_stn_groups(r, r_e, network)
double *r, *r_e ;
network_t *network ;
{
	double total_cur ; // MSSC for the whole circuit ; 
	double total_wl ;  // the total wl ; 
	lsList g, groups1 ; 
	lsGen gen ;
	double c, cmssc[1024], Rst[1024], Rst_e[1024] ;
	double total_cmssc, wl, mpl[6], pl, ncur, t;
	int n,i,j ;
	lsList vpl, vol_points ;
	lsGeneric tmp ;
	lsGen gen1 ; 
	node_t *node ;
	
	stn_groups = mape_group_get_groups(network) ;
	n = lsLength(stn_groups) ;
	//total_cur = mape_curga(network) ;
	total_cur = 77288.78 ; // for C6288.

	mape_pbssce_build_coswitch_db(network) ;

	total_cmssc = 0.0 ;
	i = 0 ;
	lsForeachItem(stn_groups, gen, g){
		c = mape_pbssce_itr_estimate_current(network, g) ;
		cmssc[i] = -c ;
		cmssc[i] = MAX(STN_MINCUR, cmssc[i]) ;
		total_cmssc += -c ;
		i ++ ;
	}

	/*
	total_cmssc = 0.0 ;
	for(i=0;i<n;i++){
		cmssc[i] = allocate_factor[i] ;
		total_cmssc += allocate_factor[i] ;
	}
	*/

	total_wl = 0.0 ;
	/* get the Rst */
	for(i=0;i<n;i++){
		c = (1+ (n*1.00)/500.00)*cmssc[i]*total_cur/total_cmssc ; 
		wl = mape_stn_cal_wl(c) ;
		wl = MAX(2, wl) ;
		total_wl += wl ;
		r[i] = mape_stn_cal_resistance(wl) ;
	}

	/* set the Rst_e */
	wl = total_wl/n ;
	for(i=0;i<n;i++)
		r_e[i] = mape_stn_cal_resistance(wl) ;

	/* the total area */
	fprintf(mapeout, "STNWL = %d\n", (int)total_wl) ;
	return stn_groups ;
}


/*
 * mape_stn_get_matrix
 *
 *
 * get a particular matrix
 */
char *
mape_stn_get_matrix(n, Ri, Rst)
int n ;
double Ri, *Rst; 
{
	int spError ;
	double *matrixEntry ;
	int i, col, c ;
	char *matrix = NULL ;

	matrix = spCreate(n, 0, &spError) ;
	assert(spError == spOKAY) ;

	col = (int)(sqrt((double)(n*1.0))) ;

	for(i=1;i<=n;i++){		// one cluster, one row. 
		c = 0 ;
		if(i+col <= n){
			c ++ ;
			matrixEntry = spGetElement(matrix, i, i+col) ;
			spADD_REAL_ELEMENT(matrixEntry, -1/(Ri));
		}
		if(i-col >= 1){
			c ++ ;
			matrixEntry = spGetElement(matrix, i, i-col) ;
			spADD_REAL_ELEMENT(matrixEntry, -1/(Ri));
		}
		if(i%col != 0 && i!=n){
			c ++ ;
			matrixEntry = spGetElement(matrix, i, i+1) ;
			spADD_REAL_ELEMENT(matrixEntry, -1/(Ri));
		}
		if((i-1)%col != 0){
			c ++ ;
			matrixEntry = spGetElement(matrix, i, i-1) ;
			spADD_REAL_ELEMENT(matrixEntry, -1/(Ri));
		}
		matrixEntry = spGetElement(matrix, i, i) ;
		spADD_REAL_ELEMENT(matrixEntry, 1/((Rst)[i-1]) + c/Ri);
	}
	spError = spFactor(matrix) ;
	assert(spError == spOKAY) ;
	return matrix ;
}


/*
 * mape_stn_stn
 *
 * report all the datas. 
 */
void 
mape_stn(network)
network_t *network ;
{
	lsList group_stn, g, group_dep ; 
	double c, cmssc[1024], Rst[1024], Rst_e[1024], Rst_dep[1024] ;
	double total_cmssc, wl, mpl[7], pl, ncur, t;
	int n,i,j ;
	lsList vpl, vol_points ;
	lsGeneric tmp ;
	lsGen gen, gen1 ; 
	node_t *node ;
	char *matrix[7] ;

	//mape_stn_isolator_figure(network) ;

	memset(cmssc, 0, sizeof(double)*1024) ;
	memset(Rst, 0, sizeof(double)*1024) ;
	memset(Rst_e, 0, sizeof(double)*1024) ;
	memset(Rst_dep, 0, sizeof(double)*1024) ;
	memset(mpl, 0, sizeof(double)*7) ;
	memset(matrix, 0, sizeof(char *)*7) ;

	group_stn = mape_stn_stn_groups(Rst, Rst_e, network) ;	
	group_dep = mape_stn_dep_groups(Rst_dep, network) ;  
	if(lsLength(group_stn) != lsLength(group_dep))
		mape_error("Length is expected to be equal!", __FILE__, __LINE__) ;
	n = lsLength(group_stn) ;

	for(i=0;i<n;i++){
		fprintf(mapeout, "Rst[%d]=%.2lf, Rst_e[%d]=%.2lf, Rst_dep[%d]=%.2lf\n", 
			i, Rst[i], i, Rst_e[i], i, Rst_dep[i]) ;
	}

	vpl = lsCreate() ;
	memset(mpl, 0, sizeof(double)*7) ;
	j = 0 ;

	matrix[0] = mape_stn_get_matrix(n, RI_1, Rst) ;	
	matrix[1] = mape_stn_get_matrix(n, RI_5, Rst) ;	
	matrix[2] = mape_stn_get_matrix(n, RI_10, Rst) ;	
	matrix[3] = mape_stn_get_matrix(n, RI_1, Rst_e) ;	
	matrix[4] = mape_stn_get_matrix(n, RI_5, Rst_e) ;	
	matrix[5] = mape_stn_get_matrix(n, RI_10, Rst_e) ;	
	matrix[6] = mape_stn_get_matrix(n, 99999999.99, Rst_dep) ;	

	for(;;){
		j ++ ;
		mape_cur_clear(network) ;
		if(!mape_disksim_read(network))
			break ;
		i = 0 ;
		lsForeachItem(group_stn, gen, g){
			while(lsLength(vpl))
				lsDelEnd(vpl, (lsGeneric *)&tmp) ;
			lsForeachItem(g, gen1, node){
				if((vol_points = mape_cur_top_vol_points(node)) != NULL)
					lsNewEnd(vpl, (lsGeneric)vol_points, LS_NH) ;
			}
			mape_pbssce_lists_max_current(vpl, &ncur, &t) ;
			cmssc[i] = -ncur ;
			cmssc[i] = MAX(cmssc[i], 0.0) ;
			i++ ;
		}
		pl = mape_stn_pl2(n, matrix[0], cmssc) ;
		mpl[0] = MAX(mpl[0], pl) ;
		pl = mape_stn_pl2(n, matrix[1], cmssc) ;
		mpl[1] = MAX(mpl[1], pl) ;
		pl = mape_stn_pl2(n, matrix[2], cmssc) ;
		mpl[2] = MAX(mpl[2], pl) ;
		pl = mape_stn_pl2(n, matrix[3], cmssc) ;
		mpl[3] = MAX(mpl[3], pl) ;
		pl = mape_stn_pl2(n, matrix[4], cmssc) ;
		mpl[4] = MAX(mpl[4], pl) ;
		pl = mape_stn_pl2(n, matrix[5], cmssc) ;
		mpl[5] = MAX(mpl[5], pl) ;

		i = 0 ;
		lsForeachItem(group_dep, gen, g){
			while(lsLength(vpl))
				lsDelEnd(vpl, (lsGeneric *)&tmp) ;
			lsForeachItem(g, gen1, node){
				if((vol_points = mape_cur_top_vol_points(node)) != NULL)
					lsNewEnd(vpl, (lsGeneric)vol_points, LS_NH) ;
			}
			mape_pbssce_lists_max_current(vpl, &ncur, &t) ;
			cmssc[i] = -ncur ;
			cmssc[i] = MAX(cmssc[i], 0.0) ;
			i++ ;
		}
		pl = mape_stn_pl2(n, matrix[6], cmssc) ;
		mpl[6] = MAX(mpl[6], pl) ;
		if(j % 1000 == 0 && j != 0)
			fprintf(mapeout, "%dth Random simulation\n", j) ;
	}
	lsDestroy(vpl, NULL) ;
	fprintf(mapeout, "Performance loss report: \n") ; 
	fprintf(mapeout, "\t Ri/Rst = 0.1, STNMPL = %.2f\n", mpl[0]) ;  
	fprintf(mapeout, "\t Ri/Rst = 0.5, STNMPL = %.2f\n", mpl[1]) ;  
	fprintf(mapeout, "\t Ri/Rst = 1.0, STNMPL = %.2f\n", mpl[2]) ;  
	fprintf(mapeout, "\t Ri/Rst = 0.1, STNEMPL = %.2f\n", mpl[3]) ;  
	fprintf(mapeout, "\t Ri/Rst = 0.5, STNEMPL = %.2f\n", mpl[4]) ;  
	fprintf(mapeout, "\t Ri/Rst = 1.0, STNEMPL = %.2f\n", mpl[5]) ;  
	fprintf(mapeout, "\t Ri/Rst = any, DEPMPL = %.2f\n", mpl[6]) ;  
}

/*
 * mape_stn_isolator_figure
 *
 * print out the figure for the effect of isolator
 */
void 
mape_stn_isolator_figure(network)
network_t *network ;
{
	int i, n, j ;
	double Ri, wl, pl ;
	double Rst[16], Rst_e[16] ;
	double CMSSC[16]={200.0, 600.0, 500.0, 800.0, 300.0, 700.0, 400.0, 800.0, 400.0, 300.0, 700.0, 400.0,
			  500.0, 800.0, 300.0, 200.0} ;
	double total_r = 0.0 ;
	double mpl1[200], mpl2[200] ;
	
	n = 16 ;
	for(i=0;i<n;i++){
		wl = mape_stn_cal_wl(CMSSC[i]) ;	
		Rst[i] = mape_stn_cal_resistance(wl) ;
		total_r += Rst[i] ;
	}
	for(i=0;i<n;i++)
		Rst_e[i] = total_r/n ;
	
	i = 0 ;
	memset(mpl1, 0, sizeof(double)*200) ;
	memset(mpl2, 0, sizeof(double)*200) ;
	for(Ri=2.0;Ri<=300.0;Ri+=2.0){
		mape_stn_set(n, Ri, Rst, CMSSC) ;
		mape_stn_create_matrix() ;
		mape_stn_solve() ;
		mpl1[i] = 0.0 ;
		for(j=0;j<n;j++){
			pl = ((stn.vol)[j]/(10000*(VDD-VTL))) ;
			mpl1[i] = MAX(mpl1[i], pl) ;
		}

		mape_stn_set(n, Ri, Rst_e, CMSSC) ;
		mape_stn_create_matrix() ;
		mape_stn_solve() ;
		mpl2[i] = 0.0 ;
		for(j=0;j<n;j++){
			pl = ((stn.vol)[j]/(10000*(VDD-VTL))) ;
			mpl2[i] = MAX(mpl2[i], pl) ;
		}
		i++ ;
	}
	i = 0 ;
	fprintf(mapeout, "Ri=[") ;
	for(Ri=2.0;Ri<=300.0;Ri+=2.0){
		fprintf(mapeout, "%.2f", Ri) ; 
		if(Ri < 300.0)
			fprintf(mapeout, ",") ;
		else 
			fprintf(mapeout, "]\n") ;
	}
	i = 0 ;
	fprintf(mapeout, "PL_U=[") ;
	for(Ri=2.0;Ri<=300.0;Ri+=2.0){
		fprintf(mapeout, "%.2f", mpl1[i]) ; 
		if(Ri < 300.0)
			fprintf(mapeout, ",") ;
		else 
			fprintf(mapeout, "]\n") ;
		i++ ;
	}
	i = 0 ;
	fprintf(mapeout, "PL_E=[") ;
	for(Ri=2.0;Ri<=300.0;Ri+=2.0){
		fprintf(mapeout, "%.2f", mpl2[i]) ; 
		if(Ri < 300.0)
			fprintf(mapeout, ",") ;
		else 
			fprintf(mapeout, "]\n") ;
		i++ ;
	}
	fprintf(mapeout, "plot (Ri, PL_U, '-', Ri, PL_E, '-')\n") ;
}




/*
 * mape_stn_test
 *
 * test stn package.  
 */
void 
mape_stn_test(void)
{
	int i, n, j ;
	double Ri ;
	double Rst[10] = {100.0, 80.0, 80.0, 150.0, 140.0, 160.0, 60.0, 160.0, 120.0, 170.0} ; 
	double CMSSC[10] = {0.1, 0.8, 0.8, 0.8, 0.8, 0.8, 0.8, 0.8, 0.8, 0.8} ;
	int wl ;

	n = 10 ;
	Ri = 22 ;

	//wl = (int)mape_stn_cal_wl(400.0) ;
	mape_stn_cal_resistance(2) ;
	return ;

	mape_stn_set(n, Ri, Rst, CMSSC) ;
	mape_stn_create_matrix() ;
	mape_stn_solve() ;

	for(i=0;i<10;i++)
		fprintf(mapeout, "\tvol[%d] = %.2f\n", i, (stn.vol)[i]) ;
	/*
	for(j=0;j<10;j++){
		n = 10 ;
		Ri = 5.0 ;
		for(i=0;i<10;i++){
			Rst[i] = 50.0 ;
			CMSSC[i] = 0.80 + j*0.1;
		}

		mape_stn_set(n, Ri, Rst, CMSSC) ;
		mape_stn_create_matrix() ;
		mape_stn_solve() ;

		fprintf(mapeout, "Uniform CMSSC current: %.2f\n", 0.8 + j*0.1) ;
		for(i=0;i<10;i++)
			fprintf(mapeout, "\tvol[%d] = %.2f\n", i, (stn.vol)[i]) ;
	}
	*/
}

/*
 * mape_stn_curga_update
 *
 *
 * callback function when doing curga. 
 */
void
mape_stn_curga_update(total)
double total ;
{
	static double max_cur = 10.0 ;
	lsList g ; 
	lsGen gen, gen1 ;
	node_t *node ;
	double cur, time ; 
	int i ;


	if(fabs(total/max_cur) < 0.98)
		return ;

	if(allocate_factor == NULL){
		allocate_factor = ALLOC(double, 1024) ;
		memset(allocate_factor, 0, 1024*sizeof(double)) ;
	}

	i = 0 ;
	lsForeachItem(stn_groups, gen, g){
		mape_clean_given_gate() ;
		lsForeachItem(g, gen1, node)
			mape_push_given_gatep(node) ;
		cur = mape_cur_max_current(global_network, &time, NULL) ;
		allocate_factor[i] += fabs(cur * total);
		//fprintf(mapeout, "update allocate_factor[%d] = %.2lf, cur = %.2lf\n", i, allocate_factor[i], cur) ;
		i ++ ;
	}
	mape_clean_given_gate() ;
	if(fabs(max_cur) < fabs(total))
		max_cur = total ;
}
