/*
 * mape_tatpg.c 
 *
 *
 * Using timed ATPG to estimation the maximum current 
 * for the circuit under unit delay model. 
 */
#include "mape_int.h"

#define MAX_NONTRANS 1.0

static lsList trans_frontier = NULL ;
static lsList pi_list ; /* pi value's list */
static lsList backtrace_list = NULL ;	/* a list used to record the last backtrace nodes*/
static int backtrack_limit = 0 ;
static int tStep = 0 ;
static int backtrace_num = 0 ;
static lsList target_nodes = NULL ;
static lsList b_target_nodes = NULL ;
static int pnt = 0 ;
static int pt = 0 ;

int mape_tatpg_imp_do(node_t *node, char *arg) ;
int mape_tatpg_clear_simulation_do(node_t *node, char *arg) ;
int mape_tatpg_backtrack_do(node_t *node, char *arg) ;
void mape_tatpg_simulation(void) ;


/*
 * mape_tatpg_clear_pi_list
 *
 * clear the pi_list. 
 */
void 
mape_tatpg_clear_pi_list(void)
{
	node_t *node ;

	if(pi_list == NULL)
		return ;

	while(lsLength(pi_list))
		lsDelEnd(pi_list, (lsGeneric *)&node) ; 
}

/*
 * mape_tatpg_push_pi
 *
 * push a pi into pi_list ; 
 */
void 
mape_tatpg_push_pi(pi)
node_t *pi;
{
	lsGen gen ;
	node_t *p ;

	if(pi_list == NULL)
		pi_list = lsCreate() ;
	
	if(!mape_node_in_list(pi, pi_list))
		lsNewEnd(pi_list, (lsGeneric)pi, LS_NH) ;
}

/*
 * mape_tatpg_pop_pi
 *
 * pop a pi from the pi_list ; 
 */
void 
mape_tatpg_pop_pi(pi)
node_t *pi;
{
	lsGen gen ;
	node_t *p ;
	node_t *tmp ;

	if(pi_list == NULL)
		return ;

	lsForeachItem(pi_list, gen, p){
		if(p == pi){
			lsDelBefore(gen, (lsGeneric *)&tmp) ;
			lsFinish(gen) ;
			return ;
		}
	}
	mape_error("pop pi which is not pushed!\n", __FILE__, __LINE__) ;
}

/*
 * mape_tatpg_pop
 *
 * pop the node's top's simulation results out.  
 */
int
mape_tatpg_pop(node, pi, pi_value)
node_t *node , *pi;
mape_node_value_t pi_value ;
{
	lsList values ;
	mape_cur_node_value_t *top ;

	if(node->mape == NULL)
		return FALSE ;

	if(((mape_t *)(node->mape))->cur_values == NULL)
		return FALSE ;

	values = ((mape_t *)(node->mape))->cur_values ;
	if(lsLength(values) == 0)
		return FALSE;

	if(mape_cur_top_vol_points(node) == NULL)
		return FALSE ;

	lsLastItem(values, (lsGeneric *)&top, LS_NH) ;
	if(top->pi == pi && top->pi_value == pi_value){
		lsDelEnd(values, (lsGeneric *)&top) ;
		if(top->vol_points)
			lsDestroy(top->vol_points, free) ;
		free(top) ;
		return TRUE ;
	}
	return FALSE ;
}

/*
 * mape_tatpg_top
 *
 * the current node's value
 */
mape_node_value_t
mape_tatpg_top(node)
node_t *node ;
{
	double time ;
	double v1, v2, v3 ;

	if(node->mape == NULL)
		return MAPE_U_U ;

	if(mape_cur_top_vol_points(node) == NULL)
		return MAPE_U_U ;


	time = (double)(HALF_SLOP_TIME*(tStep+1)) ;
	v3 = mape_cur_node_voltage(node, time) ;

	if( fabs(v3 - MAX_VOLTAGE) <= 0.001 ) 
		return MAPE_1_1 ;
	else if(v3 == UNKNOWN_VOL)
		return MAPE_U_U ;
	else if(fabs(v3 - 0.0) <= 0.001)
		return MAPE_0_0 ;
	else{
		v1 = mape_cur_node_voltage(node, time-SLOP_TIME/10) ;
		//fprintf(mapeout, "time=%.2lf, v1=%.2lf\n", time-SLOP_TIME/10, v1) ;
		v2 = mape_cur_node_voltage(node, time+SLOP_TIME/10) ;
		//fprintf(mapeout, "time=%.2lf, v2=%.2lf\n", time+SLOP_TIME/10, v2) ;


		if(v1 >= (double)(MAX_VOLTAGE/2)){
			if(v2 >= (double)(MAX_VOLTAGE/2))
				return MAPE_1_1 ;
			else 
				return MAPE_FALL ;
		}else{
			if(v2 >= (double)(MAX_VOLTAGE/2))
				return MAPE_RISE ;
			else 
				return MAPE_0_0 ;
		}
	}
}

/*
 * mape_tatpg_node_priority
 *
 * return node's priority.  
 */
float
mape_tatpg_node_priority(node)
node_t *node ;
{
	float rtv ;

	rtv = user_option->podem_option.level_priority*mape_get_node_level(node) +
		user_option->podem_option.fanout_priority*mape_node_loading_cap(node) ;

	return rtv ;
}

/*
 * mape_tatpg_pop_frontier
 *
 * pop a node out of the frontier list. 
 */
void 
mape_tatpg_pop_frontier(node)
node_t *node ;
{
	lsGen gen ;
	node_t *p, *tmp ;

	if(node == NULL || trans_frontier == NULL) 
		mape_error("Can not find the node to be removed\n", __FILE__, __LINE__);

	lsForeachItem(trans_frontier, gen, p)
		if(node == p){
			LS_ASSERT(lsDelBefore(gen, (lsGeneric *)&tmp)) ;
			lsFinish(gen) ;
			return ;
		}
	fprintf(mapeout, "try to pop node: %s\n", node->name) ;
	//mape_error("Can not find the node to be removed\n", __FILE__, __LINE__);
}


/*
 * mape_tatpg_push_frontier
 *
 * push a frontier into the frontier list. 
 */
void 
mape_tatpg_push_frontier(node)
node_t *node ;
{
	lsGen gen ;
	node_t *p ;

	if(node == NULL)
		return ;
	if(trans_frontier == NULL)
		trans_frontier = lsCreate() ;

	lsNewEnd(trans_frontier, (lsGeneric)node, NULL) ;
}

/*
 * mape_tatpg_clear_frontier
 *
 * clear the frontier list. 
 */
void 
mape_tatpg_clear_frontier(void)
{
	node_t *node ;

	if(trans_frontier == NULL)
		return ;

	while(lsLength(trans_frontier))
		lsDelBegin(trans_frontier, (lsGeneric *)&node) ;
}


/*
 * mape_tatpg_pick_frontier
 *
 * 
 * pick a frontier from the frontier list. 
 */
int 
mape_tatpg_pick_frontier(frontier, fanin, fanin_value)
node_t **frontier ;		/*transistion frontier*/
node_t **fanin ;		/*selected fanin to make the transistion propogation*/
mape_node_value_t *fanin_value ; /*fanin's value*/
{
	lsGen gen ;
	node_t *p, *tmp ;
	node_t *backup = NULL ;
	float priority, new_prio;
	int i = 0 ;
	int cv ;
	int backup_cv ;
	char text[128] ;

	if(trans_frontier == NULL)
		return FALSE ;

	if(lsLength(trans_frontier) == 0)
		return FALSE ;

	*frontier = NULL ;

	/* select the frontier which has the largest priority */
	lsForeachItem(trans_frontier, gen, p){
		if(mape_value_determined(mape_tatpg_top(p)))
			continue ;
		if(*frontier == NULL){
			*frontier = p ;
			priority = mape_tatpg_node_priority(p);
		}else if((new_prio = mape_tatpg_node_priority(p)) > priority){
			*frontier = p ;
			priority = new_prio ;
		}
		i++ ;
	}

	if(*frontier == NULL)
		return FALSE ;

	lsForeachItem(trans_frontier, gen, p){
		if(p == *frontier){
			lsDelBefore(gen, (lsGeneric *)&tmp) ;
			lsFinish(gen) ;
			break ;
		}
	}

	/*
	 * if this gate has a controlling value, 
	 * fanin is determined by the controlling value. 
	 */
	cube.num_vars = (*frontier)->nin;
	define_cube_size((*frontier)->nin) ;
	foreach_fanin((*frontier), i, p){
		if(!mape_value_determined(mape_tatpg_top(p))){
			if(mape_node_ctrl_input((*frontier)->F, i, &cv) != NO_CTRL_VALUE){ 
				*fanin_value = mape_value_convert(mape_set_value_not(cv), cv) ;
				*fanin = p ;
				return TRUE;
			}
			if(!backup){
				backup = p ;
				backup_cv = cv ;
			}
		}
	}

	/* no ctrl value in any input position, pick the last fanin tried above*/
	*fanin = backup;
	*fanin_value = MAPE_RISE ;
	/*
	fprintf(mapeout, "pick up frontier %s \n", (*frontier)->name) ;
	fflush(mapeout) ;
	*/
	return TRUE ;
}

/*
 * mape_tatpg_all_pi_assigned
 *
 * all the pi have been assigned. 
 */
int
mape_tatpg_all_pi_assigned(network)
network_t *network ;
{
	lsGen gen ;
	node_t *node ;

	foreach_primary_input(network, gen, node) {
		if( !mape_value_determined(mape_tatpg_top(node)) )
			return FALSE ;
	}
	 
	return TRUE ;
}

/*
 * mape_tatpg_backtrace
 *
 *
 * backtrace from a node to a primary input. 
 */
int 
mape_tatpg_backtrace(node, value, pi, pi_v)
node_t *node; 
mape_node_value_t value;
node_t **pi ;
mape_node_value_t *pi_v ;
{
	int i ;
	node_t *p ;
	node_t *next = NULL;
	float priority = -1 ;
	float new_prio = -1 ;
	char text[128] ;

	if(node == NULL)
		return FALSE;

	/*
	fprintf(mapeout, "backtrace: %s\t(%s)\n", node->name, mape_print_node_value(value)) ;
	fflush(mapeout) ;
	*/

	if(backtrace_num > user_option->podem_option.backtrace_limit_a){
		*pi = NULL ;
		return FALSE ;
	}

	if(node_type(node) == PRIMARY_INPUT){
		*pi = node ;
		*pi_v = value ;
		backtrace_num ++ ;
		return TRUE ;
	}
	
	foreach_fanin(node, i, p){
		if(mape_value_determined(mape_tatpg_top(p)))
			continue ;
		new_prio = mape_tatpg_node_priority(p) ;
		if(!next || new_prio > priority){
			next = p ;	
			priority = new_prio ;
		}
	}

	if(next == NULL){
		*pi = NULL ;
		return FALSE ;
	}

	if(mape_node_not_family(node))
		value = mape_value_not(value) ;
	return mape_tatpg_backtrace(next, value, pi, pi_v) ;
}

/*
 * set the target time step ;
 *
 */
void 
mape_tatpg_set_target_time(step)
int step ;
{
	tStep = step ;
}

/*
 * get the target time. 
 *
 * mape_tatpg_get_target_time
 */
double 
mape_tatpg_get_target_time(void)
{
	return (double)(HALF_SLOP_TIME*(tStep+1)) ;
}

/*
 * mape_tatpg_build_target_nodes
 *
 * build the target node list.  
 */
void 
mape_tatpg_build_target_nodes(network, step, list)
network_t *network ;
int step ;
lsList list ;
{
	node_t *node ;	
	lsGen gen, gen1 ;
	lsList st_list ;
	mape_mspplt_st_t *st ;
	int find ;

	foreach_node(network, gen, node){
		if(node_type(node) == PRIMARY_INPUT || node_type(node) == PRIMARY_OUTPUT ||
			!mape_is_given_node(node))
			continue ;
		lsNewEnd(list, (lsGeneric)node, LS_NH) ;
		st_list = ((mape_t *)(node->mape))->plt->st_list ;
		((mape_t *)(node->mape))->plt->st_value = 0 ;
		lsForeachItem(st_list, gen1, st) {
			if(st->step == step)
				((mape_t *)(node->mape))->plt->st_value = (int)(1000*(st->rp + st->fp)) ;
		}
	}
	lsSort(list, mape_mspplt_st_value_cmp) ;

	/*
	lsForeachItem(list, gen, node){
		find = FALSE ;
		fprintf(mapeout, "node %s, st@t%d=", node->name, step) ; 
		st_list = ((mape_t *)(node->mape))->plt->st_list ;
		lsForeachItem(st_list, gen1, st) 
			if(st->step == step){
				find = TRUE ;
				fprintf(mapeout, "%lf\n", st->rp + st->fp) ;
			}
		if(!find)
			fprintf(mapeout, "0.0\n") ;
	}
	*/
}

/*
 * mape_tatpg_imax_step
 *
 * return the time step which has the maxmum sum of the st*current. 
 */
int 
mape_tatpg_imax_step(network)
network_t *network ;
{
	double x[256] ;
	lsGen gen, gen1;
	node_t *node ;
	lsList st_list ;
	mape_mspplt_st_t *st ;
	int i, max_time, rtv=0 ;

	max_time = 0 ;
	memset(x, 0, sizeof(double)*256) ;

	for(i=0;i<256;i++){
		if(max_time && i > max_time)
			break ;
		foreach_node(network, gen, node){
			if(node_type(node) == PRIMARY_INPUT || node_type(node) == PRIMARY_OUTPUT
				|| !mape_is_given_node(node))
				continue ;
			st_list = ((mape_t *)(node->mape))->plt->st_list ;
			lsForeachItem(st_list, gen1, st){
				if(i == 0 && st->step > max_time)
					max_time = st->step ;
				if(st->step == i)
					x[i] += (st->rp + st->fp) ;
			}
		}
		if(x[i] > x[rtv])
			rtv = i ;
		fprintf(mapeout, "Possibility for step %d: %lf\n", i, x[i]) ;
	}
	return rtv ;
}

/*
 * mape_tatpg_imp_do
 *
 * for the mape_imp utility.
 */
int
mape_tatpg_imp_do(node, arg)
node_t *node ;
char *arg ;
{
	int rtv ;
	mape_nv_pair_t *pi_pair = (mape_nv_pair_t *)arg ;

	rtv = mape_cur_simulate_node_vx(node, pi_pair->node, pi_pair->value) ;

	if(!rtv){
		//mape_cur_set_change_pi(node, NULL, MAPE_FALL) ;
		mape_tatpg_push_frontier(node) ;
	}

	return rtv ;
}

/*
 * mape_tatpg_pi_init
 *
 * initialize pi's value to start simulation. 
 */
void 
mape_tatpg_pi_init(pi, pi_value)
node_t *pi ;
mape_node_value_t pi_value ;
{
	double fv, sv ; //first value and the second value. 
	char text[128] ;

	switch(pi_value){
		case MAPE_FALL:
			fv = MAX_VOLTAGE ;
			sv = 0.0 ;
			break ;
		case MAPE_RISE:
			fv = 0.0 ;
			sv = MAX_VOLTAGE ;
			break ;
		case MAPE_0_0:
			fv = 0.0 ;
			sv = 0.0 ;
			break ;
		case MAPE_1_1:
			fv = MAX_VOLTAGE ;
			sv = MAX_VOLTAGE ;
			break ;
		default: 
			mape_error("unexpected pi's value\n", __FILE__, __LINE__) ;
	}
	mape_cur_push_vol_point(pi, fv, (double)0.0, 1) ;
	mape_cur_push_vol_point(pi, sv, (double)SLOP_TIME, 1) ;
	mape_cur_set_change_pi(pi, pi, pi_value) ;
}

/*
 * mape_tatpg_set_pi
 *
 * set pi's value to be pi_value
 */
void 
mape_tatpg_set_pi(pi, pi_value)
node_t *pi ;
mape_node_value_t pi_value ;
{
	if(pi == NULL)
		mape_error("Passing NULL pointer to mape_tatpg_set_pi function\n", __FILE__, __LINE__) ;
	mape_curitr_node_clear(pi) ;
	mape_tatpg_pi_init(pi, pi_value) ;
}

/*
 * mape_tatpg_unset_pi
 *
 * set pi's value to be pi_value
 */
void 
mape_tatpg_unset_pi(pi)
node_t *pi ;
{
	mape_curitr_node_clear(pi) ;
}

/*
 * mape_tatpg_implication
 *
 *
 * implication for the pi's value. 
 */
void
mape_tatpg_implication(pi, pi_value)
node_t *pi ;
mape_node_value_t pi_value ;
{
	mape_nv_pair_t pi_pair ;

	/*can not implicate any more gates */
	if(pi_value == MAPE_U_U)
		return ;

	pi_pair.node = pi ;
	pi_pair.value = pi_value ;

	// initialize pi's value to start simulation. 
	mape_tatpg_pi_init(pi, pi_value) ;
	mape_cur_dump_node(pi) ;
	mape_imp_reset(mape_tatpg_imp_do ,(char *)&pi_pair) ;
	mape_imp_push(pi) ;
	mape_imp_start() ;
}

/*
 * mape_tatpg_simulation
 *
 * simulation for the pi_list. 
 */
void 
mape_tatpg_simulation(void)
{
	mape_nv_pair_t pi_pair ;	
	lsGen gen ;
	node_t *pi ;

	if(pi_list == NULL)
		return ;
	if(lsLength(pi_list) == 0)
		return ;

	mape_imp_reset(mape_tatpg_imp_do, (char *)&pi_pair) ;
	lsForeachItem(pi_list, gen, pi){
		mape_imp_push(pi) ;
		//mape_cur_dump_node(pi) ;
	}

	pi_pair.node = pi ;
	pi_pair.value = MAPE_FALL ;
	mape_imp_start() ;
}

/*
 * mape_tatpg_clear_simulation_do
 *
 * do the clear simulation
 */
int
mape_tatpg_clear_simulation_do(node, arg)
node_t *node ;
char *arg ;
{
	int rtv ;
	mape_nv_pair_t *pi_pair = (mape_nv_pair_t *)arg ;

	if(node_type(node) == PRIMARY_INPUT)
		return TRUE ;

	if(node->mape == NULL)
		return FALSE ;
	if( ((mape_t*)(node->mape))->cur_values == NULL )
		return FALSE ;

	mape_curitr_node_clear(node) ;
	return TRUE ;
}

/*
 * mape_tatpg_clear_simulation
 *
 * clear the simulation for the pi_list. 
 */
void 
mape_tatpg_clear_simulation()
{
	mape_nv_pair_t pi_pair ;	
	lsGen gen ;
	node_t *pi ;

	mape_imp_reset(mape_tatpg_clear_simulation_do, (char *)&pi_pair) ;
	foreach_primary_input(global_network, gen, pi){
		mape_imp_push(pi) ;
	}

	pi_pair.node = pi ;
	pi_pair.value = MAPE_FALL ;
	mape_imp_start() ;
}

/*
 * mape_tatpg_subpodem_backtrack_do
 *
 * backtrack implication funciton. 
 */
int 
mape_tatpg_subpodem_backtrack_do(node, arg)
node_t *node ;
char *arg ;
{
	node_t *pi = ((mape_nv_pair_t *)arg)->node ;
	mape_node_value_t pi_value = ((mape_nv_pair_t *)arg)->value ;

	if(mape_tatpg_pop(node, pi, pi_value) == TRUE){
		return TRUE ;
	}else{
		mape_tatpg_pop_frontier(node) ;
		return FALSE ;
	}
}

/*
 * mape_tatpg_subpodem_backtrack
 *
 * backtrack the last PI's implication. 
 */
void 
mape_tatpg_subpodem_backtrack(pi, pi_value)
node_t *pi ;
mape_node_value_t pi_value ;
{
	mape_nv_pair_t pi_pair ;
	if(pi_value == MAPE_U_U)
		return  ;
	
	pi_pair.node = pi ;
	pi_pair.value = pi_value ;

	backtrack_limit ++ ;
	//fprintf(mapeout, "Start backtrack:%s, %s \n", pi->name, mape_print_node_value(pi_value)) ;
	mape_imp_reset(mape_tatpg_subpodem_backtrack_do, (char *)&pi_pair) ;
	mape_imp_push(pi) ;
	mape_imp_start() ;
}

/*
 * mape_tatpg_num_nontrans
 *
 * the number of nontrans gates 
 */
void
mape_tatpg_num_nontrans(nt, t)
int *nt, *t ;
{
	lsGen gen ;
	node_t *node ;

	*nt = 0 ;
	*t = 0 ;
	foreach_node(global_network, gen, node){
		if(node_type(node) == PRIMARY_INPUT || node_type(node) == PRIMARY_OUTPUT)
			continue ;
		if(!mape_node_in_list(node, b_target_nodes))
			continue ;
		if(mape_tatpg_top(node) == MAPE_1_1 || mape_tatpg_top(node) == MAPE_0_0)
			(*nt) ++ ;
		if(mape_tatpg_top(node) == MAPE_RISE || mape_tatpg_top(node) == MAPE_FALL)
			(*t) ++ ;
	}
}

/*
 * mape_tatpg_subpodem_try_pi
 *
 * try pi's value for podem. 
 */
int 
mape_tatpg_subpodem_try_pi(node, value, pi, pi_v)
node_t *node; 
mape_node_value_t value ; 
node_t *pi ;
mape_node_value_t pi_v ;
{
	node_t *frontier, *fanin, *next_pi ; 
	mape_node_value_t  next_pi_v, fanin_value ;
	int cnt, ct, bpnt, bpt ;

	/* clear all the previous simulation result */
	mape_tatpg_clear_simulation() ;
	mape_tatpg_clear_frontier() ;
	bpnt = pnt ;
	bpt = pt ;

	//fprintf(mapeout, "\tSubpodem try pi: %s(%s) \n", pi->name, mape_print_node_value(pi_v)) ; 
	//mape_tatpg_implication(pi, pi_v) ;

	mape_tatpg_set_pi(pi, pi_v);
	mape_tatpg_push_pi(pi) ;
	mape_tatpg_simulation() ;
	mape_tatpg_num_nontrans(&cnt, &ct) ;
	//fprintf(mapeout, "\tpnt=%d, pt=%d, cnt=%d, ct=%d\n", pnt, pt, cnt, ct) ;

	/*
	if(user_option->version >= 2){
		if( ((double)(cnt) - (double)(pnt))/((double)(ct) - (double)(pt) + 0.1) >= MAX_NONTRANS){
			backtrack_limit ++ ;
			mape_tatpg_pop_pi(pi) ;
			mape_tatpg_unset_pi(pi) ;
			return FALSE ;
		}
	}
	*/

	pnt = cnt ;
	pt = ct ;
	if(mape_tatpg_top(node) == value){
		return TRUE ;
	}
	if(backtrack_limit < user_option->podem_option.backtrack_limit 
		&& !mape_conflict(value, mape_tatpg_top(node))){
		while(mape_tatpg_pick_frontier(&frontier, &fanin, &fanin_value)){
			if(mape_tatpg_backtrace(fanin, fanin_value, &next_pi, &next_pi_v)){
				if(mape_node_in_list(next_pi, pi_list)){
					//fprintf(mapeout, "node in list: %s, ready to continue\n", next_pi) ;
					continue ;
				}
				if(mape_tatpg_subpodem(node, value, next_pi, next_pi_v))
					return TRUE ;
			}
		}
	}

	/*
	fprintf(mapeout, "Conflict happens: %s, expect %s, ", pi->name, mape_print_node_value(value)) ;  
	fprintf(mapeout, "obtain %s\n", mape_print_node_value(mape_tatpg_top(node))) ;  
	*/

	pnt = bpnt;	
	pt = bpt ;
	backtrack_limit ++ ;
	mape_tatpg_pop_pi(pi) ;
	mape_tatpg_unset_pi(pi) ;
	return FALSE ;
}

/*
 * mape_tatpg_subpodem
 *
 * subpodem. 
 */
int 
mape_tatpg_subpodem(node, value, pi, pi_v)
node_t *node ;
mape_node_value_t value ;
node_t *pi ;
mape_node_value_t pi_v ;
{
	if(mape_tatpg_subpodem_try_pi(node, value, pi, pi_v))
		return TRUE ;
	if(mape_tatpg_subpodem_try_pi(node, value, pi, mape_value_first_not(pi_v)))
		return TRUE ;
	if(mape_tatpg_subpodem_try_pi(node, value, pi, mape_value_second_not(pi_v)))
		return TRUE ;
	if(mape_tatpg_subpodem_try_pi(node, value, pi, mape_value_not(pi_v)))
		return TRUE ;
	return FALSE ;
}

/*
 * mape_tatpg_podem  
 *
 * using podem algorithm to justify the node's value. 
 * must under the current stack status.
 */
int 
mape_tatpg_podem(node, value)
node_t *node ;
mape_node_value_t value ;
{
	node_t *pi ;
	mape_node_value_t pi_v ;
	lsGen gen ;
	node_t *np ;

	backtrack_limit = 0 ;
	mape_tatpg_backtrace(node, value, &pi, &pi_v) ;

	if(!mape_node_in_list(pi, pi_list) && pi != NULL)
		return mape_tatpg_subpodem(node, value, pi, pi_v) ;
	else{ 
		lsForeachItem( ((mape_t*)(node->mape))->pi_fanin, gen, np){
			if(mape_node_in_list(np, pi_list)) ;
				continue ;
			return mape_tatpg_subpodem(node, value, np, MAPE_RISE) ;
		}

		/*
		while(mape_tatpg_alternate_backtrace(&pi, &pi_v)){
			if(mape_node_in_list(pi, pi_list))
				continue ;
			return mape_tatpg_subpodem(node, value, pi, pi_v) ;
		}
		*/
		return FALSE;
	}
}

/*
 * mape_justify
 *
 * justify the node's value. when success, all the
 * value should be kept in the node. when fail, all
 * the values should be undo.
 */
int 
mape_tatpg_justify(node, value)
node_t *node ;
mape_node_value_t value ;
{
	int rtv ;
	fprintf(mapeout, "Justify node: %s, ctrl value=%d ... ", 
		node->name, ((mape_t*)node->mape)->C0 + ((mape_t*)node->mape)->C1) ;
	rtv = mape_tatpg_podem(node, value) ;

	if(rtv)
		fprintf(mapeout, "success!\n") ;
	else
		fprintf(mapeout, "fail!\n") ;
	return rtv ;
}

/*
 * mape_tatpg_ctrl_value_cmp
 *
 * compare node according to the control value. 
 */
int
mape_tatpg_ctrl_value_cmp(obj1, obj2)
lsGeneric obj1, obj2 ;
{
	int c1, c2 ;

	c1 = mape_get_node_C0((node_t *)obj1) + mape_get_node_C1((node_t *)obj1) ;
	c2 = mape_get_node_C0((node_t *)obj2) + mape_get_node_C1((node_t *)obj2) ;

	return c1 - c2 ;
}

/*
 * mape_tatpg_satisfy_target_nodes
 *
 * satisfy all the target nodes. 
 */
void 
mape_tatpg_satisfy_target_nodes(network, list)
network_t *network ;
lsList list ;
{
	lsGen gen ;
	node_t *node ;

	if(user_option->version >= 3){
		lsSort(list, mape_tatpg_ctrl_value_cmp) ;
	}

	while(lsLength(list)){
		lsDelBegin(list, (lsGeneric *)&node) ;
		mape_node_set_visit(node) ;
		backtrace_num = 0 ;

		/* assign rising to the node */
		if(!mape_conflict(MAPE_RISE, mape_tatpg_top(node))){
			if(mape_tatpg_justify(node, MAPE_RISE)){
				if(mape_tatpg_all_pi_assigned(network))
					break ;
				else 
					continue ;
			}
		}

		/* assign falling to the node */
		if(!mape_conflict(MAPE_FALL, mape_tatpg_top(node))){
			if(mape_tatpg_justify(node, MAPE_FALL)){
				if(mape_tatpg_all_pi_assigned(network))
					break ;
				else 
					continue ;
			}
		}
	}
}

/*
 * TATPG report
 *
 * report the result. 
 */
int
mape_tatpg_transition_count()
{
	lsGen gen ;
	node_t *node ;
	int count ;

	count = 0 ;
	foreach_node(global_network, gen, node){
		if(node_type(node) == PRIMARY_INPUT || node_type(node) == PRIMARY_OUTPUT)
			continue ;
		if(mape_tatpg_top(node) == MAPE_RISE || mape_tatpg_top(node) == MAPE_FALL)
			count ++ ;
	}
	return count ;
}

/*
 * TATPG report
 *
 * report the result. 
 */
void 
mape_tatpg_report(network)
network_t *network ;
{
	lsGen gen ;
	node_t *node ;
	int count ;

	fprintf(mapeout, "Target time: %d\n", tStep) ;

	foreach_node(network, gen, node){
		if(node_type(node) == PRIMARY_INPUT || node_type(node) == PRIMARY_OUTPUT)
			continue ;
		if(mape_tatpg_top(node) == MAPE_RISE || mape_tatpg_top(node) == MAPE_FALL)
			fprintf(mapeout, "\tnode %s is transitioning\n", node->name) ; 
	}
}

/*
 * mape_tatpg_assign_remaining_pi
 *
 * assign remaining pi's value. 
 */
void 
mape_tatpg_assign_remaining_pi(network, mcur)
network_t *network ;
double *mcur ;
{
	lsGen gen ;
	node_t *node, *pi ;
	double mcur1, time ;
	lsList b_list = NULL ;

	if(pi_list)
		b_list = lsCopy(pi_list, NULL) ;

	foreach_primary_input(network, gen, pi){
		if(mape_node_in_list(pi, pi_list))
			continue ;
		mape_tatpg_set_pi(pi, MAPE_RISE) ;
		mape_tatpg_push_pi(pi) ;
	}
	mape_tatpg_clear_simulation() ;
	mape_tatpg_simulation() ;
	*mcur = mape_cur_max_current(network, &time, NULL) ;
	fprintf(mapeout, "Get rising maximum current: %lf\n", *mcur) ;

	
	foreach_primary_input(network, gen, pi){
		if(mape_node_in_list(pi, b_list))
			continue ;
		mape_tatpg_set_pi(pi, MAPE_FALL) ;
	}
	mape_tatpg_clear_simulation() ;
	mape_tatpg_simulation() ;
	mcur1 = mape_cur_max_current(network, &time, NULL) ;
	fprintf(mapeout, "Get Falling maximum current: %lf\n", mcur1) ;

	if(fabs(*mcur) < fabs(mcur1))
		*mcur = mcur1 ;
	else {
		foreach_primary_input(network, gen, pi){
			if(mape_node_in_list(pi, pi_list))
				continue ;
			mape_tatpg_set_pi(pi, MAPE_FALL) ;
			mape_tatpg_push_pi(pi) ;
		}
		mape_tatpg_clear_simulation() ;
		mape_tatpg_simulation() ;
	}
	if(b_list)
		lsDestroy(b_list, NULL) ;
}

/*
 * mape_tatpg_assign_pi_priority
 *
 * assign the pi's priority. 
 */
void 
mape_tatpg_assign_pi_priority(target_nodes) 
lsList target_nodes ;
{
	lsGen gen, gen1;
	node_t *np, *pi ;
	lsList l ;

	lsForeachItem(target_nodes, gen, np){
		l = ((mape_t *)(np->mape))->pi_fanin ;
		lsForeachItem(l, gen1, pi)
			((mape_t*)(pi->mape))->visit ++ ;
	}
}

/*
 * mape_tatpg_pi_priority_comp
 *
 * compare two pi according to the pi's priority. 
 */
int 
mape_tatpg_pi_priority_cmp(obj1, obj2)
lsGeneric obj1, obj2 ;
{
	return ((mape_t*)(((node_t *)obj1)->mape))->visit  - 
	       ((mape_t*)(((node_t *)obj2)->mape))->visit ;
}

/*
 * mape_tatpg_assign_pi_value
 *
 * assign pi's value. 
 */
void 
mape_tatpg_assign_pi_value(pi)
node_t *pi ;
{
	int cnt1, ct1, cnt2, ct2;
	double p1, p2, t1, t2;

	fprintf(mapeout, "Try pi: %s ... pnt=%d, ct=%d ", pi->name, pnt, pt) ;
	mape_tatpg_clear_simulation() ;
	mape_tatpg_set_pi(pi, MAPE_RISE);
	mape_tatpg_push_pi(pi) ;
	mape_tatpg_simulation() ;

	mape_tatpg_num_nontrans(&cnt1, &ct1) ;

	mape_tatpg_clear_simulation() ;
	mape_tatpg_set_pi(pi, MAPE_FALL) ;
	mape_tatpg_simulation() ;
	mape_tatpg_num_nontrans(&cnt2, &ct2) ;

	t1 = ((double)(cnt1 - pnt))/((double)(ct1-pt+0.1)) ;
	t2 = ((double)(cnt2 - pnt))/((double)(ct2-pt+0.1)) ;

	if(user_option->version >= 2){
		if(t1 >= MAX_NONTRANS && t2 >= MAX_NONTRANS){
			mape_tatpg_pop_pi(pi) ;
			mape_tatpg_unset_pi(pi) ;
			fprintf(mapeout, " fail!\n") ;
			return ;
		}
	}
	if(t2 > t1){
		mape_tatpg_set_pi(pi, MAPE_RISE);
		mape_tatpg_push_pi(pi) ;
		pnt = cnt1 ;
		pt  = ct1 ;
	}else {
		pnt = cnt2 ;
		pt  = ct2 ;
	}
	fprintf(mapeout, " Success!\n") ;
}

/*
 * mape_tatpg_satisfy_target_nodes_g
 *
 * satisfy target nodes in a global way. 
 */
void 
mape_tatpg_satisfy_target_nodes_g(network, target_nodes)
network_t *network ;
lsList target_nodes ;
{
	lsList pis ;
	lsGen gen ;
	node_t *pi ;

	pis = lsCreate() ;
	foreach_primary_input(network, gen, pi)
		lsNewEnd(pis, (lsGeneric)pi, LS_NH) ;

	/* set all the visit to be 0 */
	lsForeachItem(pis, gen, pi) 
		((mape_t *)(pi->mape))->visit = 0 ;

	/* assign the priority */
	mape_tatpg_assign_pi_priority(target_nodes) ;


	/*sorting the pis according to the priority*/
	lsSort(pis, mape_tatpg_pi_priority_cmp) ;

	lsForeachItem(pis, gen, pi)
		mape_tatpg_assign_pi_value(pi) ;

	lsDestroy(pis, NULL) ;
}

/*
 * tATPG algorithm. 
 */
void
mape_tatpg(network)
network_t *network ;
{
	int step ;
	double mcur, time ;

	mape_curlib_set_unit_mode(TRUE) ;
	target_nodes = lsCreate() ;
	mape_curlib_set_unit_mode(TRUE) ; 

	mape_timer_reset() ;
	mape_timer_start() ;
	/*Get the possibility for each time point */
	mape_mspplt_st_possibility(network) ;	

	/*Get all the control value for each node */
	mape_ctrl_get_all(network) ;

	/*Get all pi fanin */
	mape_ctrl_get_all_pi_fanin(network) ;
	/*
	mape_ctrl_dump_pi_fanin(network) ;
	*/

	/*return the time step which has the maximum current*/
	step = mape_tatpg_imax_step(network) ;

	/*get the sorted list for the node to be eabled.*/
	mape_tatpg_build_target_nodes(network, step, target_nodes) ;

	/*set the target time for ATPG*/
	mape_tatpg_set_target_time(step) ;

	/*
	mape_tatpg_assign_remaining_pi(network, &mcur);
	fprintf(mapeout, "Maximum current obtained : %lf\n", mcur) ;
	return ;
	*/
	b_target_nodes = lsCopy(target_nodes, NULL) ;
	/* satisfying the target nodes */
	mape_tatpg_satisfy_target_nodes(network, target_nodes) ;

	/* satisfying the target nodes in a different way */
	/*
	mape_tatpg_satisfy_target_nodes_g(network, target_nodes) ;
	*/

	/* random assignment to remaining PI */
	mape_tatpg_assign_remaining_pi(network, &mcur) ;
	mape_timer_stop() ;
	mape_timer_report("TATPG time:") ;

	/* report the result */
	mape_tatpg_report(network) ;
	fprintf(mapeout, "Maximum current : %lf\n", mcur) ;

	lsDestroy(target_nodes, NULL) ;
}

/*
 * mape_tatpg_grp
 *
 * T-ATPG for cluster. 
 */
void 
mape_tatpg_grp(network)
network_t *network ;
{
	int step ;
	double mcur, time ;
	lsList groups, group ;
	lsGen gen, gen1, gen2 ;
	node_t *node, *pi ;
	int i = 0 ;

	/* get the initial grup method */
	groups = mape_group_get_groups(network) ;

	/*
	fprintf(mapeout, "Dumping groups: \n\n") ;
	i = 0 ;
	lsForeachItem(groups, gen1, group){
		i++ ;
		fprintf(mapeout, "Group #%d\n\t", i) ;
		lsForeachItem(group, gen2, node){
			fprintf(mapeout, "%s ", node->name) ;
		}
		fprintf(mapeout, "\n") ;
	}
	*/

	mape_curlib_set_unit_mode(TRUE) ;
	target_nodes = lsCreate() ;

	/*Get the possibility for each time point */
	mape_mspplt_st_possibility(network) ;	

	/*Get all the control value for each node */
	mape_ctrl_get_all(network) ;

	/*Get all pi fanin */
	mape_ctrl_get_all_pi_fanin(network) ;

	
	lsForeachItem(groups, gen1, group){
		mape_timer_reset() ;
		mape_timer_start() ;
		i++ ;
		fprintf(mapeout, "Group #%d\n", i) ;

		/* renew the initial value */ 
		mape_clean_given_gate() ;
		lsForeachItem(group, gen2, node){
			mape_push_given_gatep(node) ;
			fprintf(mapeout, "%s ", node->name) ;
		}
		fprintf(mapeout, "\n") ;
		mape_tatpg_clear_simulation() ;
		mape_tatpg_clear_pi_list() ;
		foreach_primary_input(network, gen, pi)
			mape_curitr_node_clear(pi) ;
		backtrack_limit = 0 ;
		pnt = pt = 0 ;

		/*return the time step which has the maximum current*/
		step = mape_tatpg_imax_step(network) ;

		/*get the sorted list for the node to be eabled.*/
		mape_tatpg_build_target_nodes(network, step, target_nodes) ;

		/*set the target time for ATPG*/
		mape_tatpg_set_target_time(step) ;
		b_target_nodes = lsCopy(target_nodes, NULL) ;

		/* satisfying the target nodes */
		mape_tatpg_satisfy_target_nodes(network, target_nodes) ;

		/* random assignment to remaining PI */
		mape_tatpg_assign_remaining_pi(network, &mcur) ;
		mape_timer_stop() ;
		mape_timer_report("TATPG time:") ;

		/* report the result */
		/*
		mape_tatpg_report(network) ;
		*/
		fprintf(mapeout, "Maximum current : %lf\n", mcur) ;
	}

	lsDestroy(target_nodes, NULL) ;
}
